/************************************************************************/
/*									*/
/*	File:	ao1.c							*/
/*									*/
/*	Assembly language peephole optimizer for use with Z80 versions	*/
/*	of the Q/C "C" compiler.					*/
/*									*/
/*		Written	by Kirk	Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1984-1986	Logical	Systems.		*/
/*									*/
/*				04/20/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file	*/
#include	<string.h>		/* String functions		*/
#include	<ctype.h>
#include	<unistd.h>
#ifdef MSDOS
#include	<process.h>
#include	<alloc.h>
#endif
#include	"aodef.h"		/* Optimizer definitions	*/
#include	"aotyp.h"		/* Optimizer data declarations	*/
#include	"aoext.h"		/* Optimizer externals		*/

/************************************************************************/
/*									*/
/*			Global Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	I/O definitions.
 */
char	ifnbuf[FNSIZE +	1];		/* Input filename buffer */
STREAM	in_fp		=	NULL;	/* Input file pointer */
char	ofnbuf[FNSIZE +	1];		/* Output filename buffer */
STREAM	out_fp		=	NULL;	/* Output file pointer */
char	rfnbuf[FNSIZE +	1];		/* Restart linkage filename buffer */
STREAM	rst_fp		=	NULL;	/* Restart linkage file	pointer	*/

/*
 *	General	definitions.
 */
UINT	b_added		=	0;	/* # of	bytes of code added */
UINT	b_read		=	0;	/* # of	bytes of code read */
UINT	b_saved		=	0;	/* # of	bytes saved by optimizer */
BOOL	dseg		=	FALSE;	/* Assume in code segment */
char	entrypt[SYMNMESIZE + 1]	= { 0 };     /*	Entry point if linkrst */
BOOL	freq		=	FALSE;	/* TRUE	if checking "call" freq. */
UINT	jqcsize		=	0;	/* Bytes of code in jump queue */
BOOL	jqfull		=	FALSE;	/* TRUE	if jump	queue ever full	*/
BOOL	linkrst		=	FALSE;	/* TRUE	if init. restart linkage */
struct	queue	jq;			/* JP->JR conversion queue */
struct	pbuf	pb	=	{ NULL };	/* Parsing/sizing buffers */
struct	rstslot	restart[8] = { {"",0}, };
				/* Restart linkage info	*/

/* MegaMax will	es so :
struct	rstslot	restart[8] =	{ ""  };	 Restart linkage info */

BOOL	rom		=	FALSE;	/* TRUE	if reading rom code */
BOOL	rst		=	FALSE;	/* TRUE	if linking restart vectors */
struct	symtab	*st;			/* Beginning of	symbol table */
struct	symtab	*stend;			/* Last	entry of symbol	table */
BOOL	stfull		=	FALSE;	/* TRUE	if symbol table	ever full */
BOOL	verbose		=	TRUE;	/* Default: make output	verbose	*/





/************************************************************************/
/*									*/
/*	Routine:	main						*/
/*									*/
/*	Opens the input, restart and output files, deciphers any	*/
/*	command	line switches, allocates free memory and starts	the	*/
/*	optimization going.						*/
/*									*/
/*	Parameters:							*/
/*		"argc" An integer count	of the command line arguments.	*/
/*		"argv" A pointer to an array of	pointers which point	*/
/*			to individual command line arguments.		*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	main(argc, argv)
	int argc;
	char *argv[];
	{
	register BOOL rename;
	register char *ptr;
	register UINT stsize,i;

/*
 *	Check validity of input	filename and open it.
 */
	if(argc	== 1) banner();		/* Tell	user who we are	*/
	if((argc -= 2) < 0) usagerr();	/* Needs at least an input file	*/
	makefn(*++argv,ifnbuf,INPUT_EXT);	/* Expand into full filename */
	if(! (in_fp = (STREAM) fopen(ifnbuf,"r")))
		{
		fatal("Unable to open input file: ",ifnbuf);
		}
/*
 *	Given the input	filename, create a default output filename in
 *	case the user leaves it	to us.
 */
	strcpy(ofnbuf,ifnbuf);		/* Create default output filename */
	strcpy(strchr(ofnbuf,'.'),OUTPUT_EXT);
	rename = FALSE;			/* And assume it will be used */
/*
 *	See if the user	has specified a	file to	read the restart linkage
 *	symbols	from.  If so open the file and process the restart info.
 */
	if(argc	&& (*(*++argv) != '-'))		/* If restart file is given */
		{
		rst = TRUE;		/* Doing restart optimizations */
		makefn(*argv++,rfnbuf,RESTART_EXT);	/* Process filename */
		argc--;			/* One less item on line */
		if(! (rst_fp = fopen(rfnbuf,"r")))
			{
			fatal("Unable to open restart linkage file: ",
				rfnbuf);
			}
		}
/*
 *	Process	any command line switches.
 */
	while(argc--)			/* Process rest	of command line	*/
		{
		ptr = *argv++;
		if (*ptr == '-')
			{
			while (*++ptr)
				{
				switch (toupper(*ptr))
					{
/* Output to console (debug) */		case 'D':
						out_fp = stdout;
						break;
/* Examine "call" symbol frequency? */	case 'F':
						freq = TRUE;
						break;
/* Rename output file */		case 'O':
						rename = TRUE;
						break;
/* Turn	OFF verbose */			case 'Q':
						verbose	= FALSE;
						break;
/* Reading rom code */			case 'S':
						rom = TRUE;
						break;
					default:
						usagerr();
					}
				}
			}
		else	if (rename)		/* New output filename?	*/
				{
/* Default name, new drive? */	if (ptr[1] == ':' && (!	ptr[2]))
					{
					strcpy(ofnbuf,ptr);
					strcat(ofnbuf,ifnbuf[1]==':'
						? ifnbuf + 2 : ifnbuf);
					strcpy(strchr(ofnbuf,'.'),OUTPUT_EXT);
					}
/* New filename	*/		else
					{
/* Output going	to printer? */		if(! strcmp(ptr,"LST:"))
						{
/* Yep */					strcpy(ofnbuf,"lst:");
						}
					else
						{
/* Otherwise expand filename */			makefn(ptr,ofnbuf,OUTPUT_EXT);
						}
					}
				rename = FALSE;
				}
		else	usagerr();
		}
/*
 *	Unless the output is to	go to the screen open the output file.
 */
	if(! out_fp)
		{
/*
 *	Make sure input	and output files are distinct.
 */
		if (! strcmp(ifnbuf,ofnbuf))
			{
			fatal("Output file name	same as	input","");
			}
		if(! (out_fp = fopen(ofnbuf,"w")))
			{
			fatal("Unable to open output file: ",ofnbuf);
			}
		}
/*
 *	Allocate/initialize memory for the symbol table.
 */
	stsize = (SYMTABSIZE * sizeof(struct symtab));	/* #1 entry in ST */
	if((int) (st = (struct symtab *) sbrk(stsize)) == -1)
		{
		fatal("Insufficient memory available for symbol	table","");
		}
	stend =	st + (SYMTABSIZE - 1);
	ptr = (char *) st;
	for(i =	0; i < stsize; i++)
		{
		*ptr++ = '\0';		/* Zero	symbol table out */
		}
/*
 *	Print sign-on messages if appropriate.
 */
	if(verbose) banner();
/*
 *	Verify that the	various	portability configuration options have been
 *	set correctly.
 */

	if(UINT4 ? (sizeof(UINT) != 4) :
		 (sizeof(UINT) != 2))
		 fatal("The size of UINT is not	correctly configured","");
	if(PTR4	? (sizeof(struct symtab	*) != 4) :
		(sizeof(struct symtab *) != 2))
		fatal("The size	of pointers is not correctly configured","");
	if(sizeof(PTRUINT) < sizeof(struct symtab *))
		fatal("The PTRUINT integral type is not	large enough","");

/*lint +e506 lint isn't	very fond of constant valued booleans */

/*
 *	If we are processing call/restart substitutions, read linkage file.
 */
	if(rst)	r_rst();		/* Read	restart	linkage	info */
/*
 *	Do an initial read of the input	file to	see if it has anything.
 */
	if(parse())
		{
		fprintf(stderr,"WARNING: Nothing in input file:	'%s'\n",ifnbuf);
		exit(NOERRORS);
		}
	if(! freq)			/* Otherwise we	won't need queue */
		{
/*
 *	Allocate memory	to the JP->JR conversion queue.
 */
		if((int) (jq.strt = sbrk((UINT)	JQSIZE)) == -1)
			{
			fatal("Insufficient memory for jump queue","");
			}
		jq.end = jq.strt + JQSIZE - 1;	/* Physical end	*/
		jq.in =	jq.out = jq.strt;	/* Logical start and end */
/*
 *	Everything is ready to go, now is the time to:
 */
		optimize();
		}
/*
 *	Otherwise, we are going	to count the static frequency of routines
 *	that are called.
 */
	chkfreq();
	}





/************************************************************************/
/*									*/
/*	Routine:	add_jmp						*/
/*									*/
/*	Add a jump to the symbol table.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the symbol	table entry, or	NULL if	*/
/*			no symbol table	space was available.		*/
/*									*/
/************************************************************************/

struct symtab
	*add_jmp()
	{
	struct symtab *ptr;

	if(! (ptr = lookup())) return ptr;	/* No space here */
	ptr->info.jinfo.jmpcnt++;		/* One more in the queue */
	return ptr;
	}





/************************************************************************/
/*									*/
/*	Routine:	add_lab						*/
/*									*/
/*	Add a label to the symbol table.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the symbol	table entry, or	NULL if	*/
/*			no symbol table	space was available.		*/
/*									*/
/************************************************************************/

struct symtab
	*add_lab()
	{
	struct symtab *ptr;

	if(! (ptr = lookup())) return ptr;	/* No space here */
	if(ptr->info.jinfo.islabel)
		{
		fatal("Duplicate definition of symbol: ",pb.symbuf);
		}
	ptr->info.jinfo.islabel	= TRUE;		/* Label is defined */
	return ptr;
	}





/************************************************************************/
/*									*/
/*	Routine:	banner						*/
/*									*/
/*	Display	the sign-on banner to the user.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	banner()
 {
  fprintf(stderr,"NORD><LINK Assembly Code Optimizer Version 1.1a\n");
  fprintf(stderr,"ported to Linux by dl4sdc\n\n");
  fprintf(stderr,"This version is NOT public domain.\n");
  fprintf(stderr,"It was given to YOU as a radio amateur and ONLY for noncommercial use.\n");
  fprintf(stderr,"(By special permission of Quality Computer Systems)\n\n");
 }





/************************************************************************/
/*									*/
/*	Routine:	chkfreq						*/
/*									*/
/*	Read the input file and	count the static frequency of the	*/
/*	symbols	appearing in "call" instructions.  The symbols used are	*/
/*	then written to	the output file	along with a count of how many	*/
/*	times they appeared.  The output is sorted with	the most	*/
/*	popular	symbols	appearing first.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	chkfreq()
	{
	struct symtab *heap[SYMTABSIZE];
	register UINT i,j,k,l;
	register struct	symtab *ptr;

	if(verbose)
		fprintf(stderr,"Accumulating counts of routines	called\n");
	do
		{
		if(pb.type == CALLTYPE)	/* Was it a "call" instruction?	*/
			{
			if((ptr	= lookup()) != NULL)	/* If space in ST */
				{
				ptr->info.callcnt++;	/* Add one more	*/
				}
			}
		} while(! parse());	/* While some input to read */
/*
 *	We've accumulated the frequencies, figure out which symbols
 *	were used, condense the	pointers to used entries into an array
 *	of pointers and	sort the array (keyed off the frequency).
 */
	if(verbose) fprintf(stderr,"Sorting called routine frequencies\n");
	for(i =	0, ptr = st; ptr <= stend; ptr++)	/* Condense */
		{
		if(*ptr->sname)		/* If something	in this	slot */
			{
			j = ptr->info.callcnt;
			for(k =	0; k < i; k++)
				{
				if(j > heap[k]->info.callcnt) break;
				}
			for(l =	i++; l > k; l--)
				{
				heap[l]	= heap[l - 1];
				}
			heap[k]	= ptr;
			}
		}
/*
 *	Now output them.
 */
	if(verbose) fprintf(stderr,"Writing called routine frequencies\n");
	for(j =	0; j < i; j++)
		{
		strncpy(pb.symbuf,heap[j]->sname,SYMNMESIZE);
		pb.symbuf[SYMNMESIZE] =	'\0';
		fprintf(out_fp,"%s\t%u\n",pb.symbuf,heap[j]->info.callcnt);
		}
	if(verbose) fprintf(stderr,"\n%u different routines called\n",i);
	exit(NOERRORS);			/* Close files */
	}





/************************************************************************/
/*									*/
/*	Routine:	fatal						*/
/*									*/
/*	Print out a "FATAL" error message and quit.			*/
/*									*/
/*	Parameters:							*/
/*		"sptr1"	A pointer to the first of the error message.	*/
/*		"sptr2"	A pointer to the rest of the error message.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	fatal(sptr1,sptr2)
	char *sptr1,*sptr2;
	{
	fprintf(stderr,"FATAL: %s%s\n",sptr1,sptr2);   /* Print	message	*/
	exit(ERRORS);			/* Bye... */
	}





/************************************************************************/
/*									*/
/*	Routine:	get_jq						*/
/*									*/
/*	Get a character	from the jump queue.				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The character read.				*/
/*									*/
/************************************************************************/

char	get_jq()
	{
	char c;

	c = *jq.out++;			/* Get next character */
	if(jq.out > jq.end) jq.out = jq.strt;	/* Handle wraparound */
	return c;
	}





/************************************************************************/
/*									*/
/*	Routine:	lookup						*/
/*									*/
/*	Check if a symbol is in	the symbol table, if not add it.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	A pointer to the resulting symbol table	entry,	*/
/*			or NULL	if the symbol table was	full.		*/
/*									*/
/************************************************************************/

struct symtab
*lookup()
	{
	register char *tptr;
	register UINT num;
	register struct	symtab *ptr;

	if(! stfull)			/* Symbol table	was full! */
		{
		tptr = pb.symbuf;	/* Point to symbol */
		num = 0;		/* Zero	hash sum */
		while(*tptr) num = (num	* 16) +	*tptr++; /* Compute hash */
		num &= SYMTABMASK;	/* Mask	to size	of table */
/*
 *	We now have the	symbol table entry number to start scanning at,
 *	but we need to convert that to the correct pointer, therefore:
 */
		ptr = st + num;
/*
 *	Scan table for exact match or empty slot.
 */
		for(num	= 0; num < SYMTABSIZE; num++ , ptr++)
			{
			if(ptr > stend)	ptr = st;	/* Wrap	around?	*/
			if(*ptr->sname)		/* Entry filled? */
				{
				if(! strncmp(pb.symbuf,ptr->sname,
							SYMNMESIZE))
					{
					return ptr;	/* Exact match */
					}
				continue;	/* Filled, try next */
				}
			strncpy(ptr->sname,pb.symbuf,SYMNMESIZE);
			return ptr;
			}
		fprintf(stderr,"WARNING: Symbol	table full");
		stfull = TRUE;		/* Tell	user once only */
		}
	pb.type	= NORMTYPE;		/* Since not going in ST */
	return NULL;
	}





/************************************************************************/
/*									*/
/*	Routine:	makefn						*/
/*									*/
/*	Expand a skeleton filename by adding a default extension if	*/
/*	none is	supplied.						*/
/*									*/
/*	Parameters:							*/
/*		"sptr"	A pointer to the skeleton filename.		*/
/*		"bptr"	A pointer to the buffer	the result goes	in.	*/
/*		"ext"	A pointer to the default extension to use.	*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	makefn(sptr,bptr,ext)
	char *sptr,*bptr,*ext;
	{
	if(strlen(sptr)	> FNSIZE) usagerr();
	strcpy(bptr,sptr);
	if(! strchr(bptr,'.'))
		{
		if(strlen(bptr)	> (FNSIZE - 4))	usagerr();
		strcat(bptr,ext);
		}
	}





/************************************************************************/
/*									*/
/*	Routine:	optimize					*/
/*									*/
/*	Top level controlling routine for the assembly language		*/
/*	peephole optimizer.  Calls the routines	that read the source	*/
/*	file, perform the various optimizations, and write the output	*/
/*	file.  Also prints the trailing	execution messages to the user.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	optimize()
	{
	while(read_jq())		/* While code to read */
		{
		write_jq();		/* Write a line	from queue */
		}
/*
 *	The input file is now empty, but the jump queue	still contains
 *	code, thus continue to optimize	until the jump queue itself
 *	is empty.
 */
	while(write_jq()) ;		/* While code to write */
	if(linkrst) lnk_rst();		/* Generating restart linkages?	*/
	if(verbose)
		{
		fprintf(stderr,"%u byte(s) of code saved ",b_saved);
		fprintf(stderr,"out of %u byte(s) read",b_read);
		if(linkrst)
			{
			fprintf(stderr,
				"\nAdded %u byte(s) for	restart	linkages",
				b_added);
			}
		}
	fprintf(stderr,"\n");
	exit(NOERRORS);
	}





/************************************************************************/
/*									*/
/*	Routine:	put_jq						*/
/*									*/
/*	Place a	character in the jump queue and	update the queue	*/
/*	pointers as appropriate.					*/
/*									*/
/*	Parameters:							*/
/*		"c"	The character to add to	the queue.		*/
/*									*/
/*	Returns:	TRUE if	the jump queue was full	and the		*/
/*			character would	not fit.  FALSE	otherwise.	*/
/*									*/
/************************************************************************/

BOOL	put_jq(c)
	char c;
	{
	*jq.in++ = c;			/* Put character in queue */
	if(jq.in > jq.end)		/* Past	physical end of	queue? */
		{
		jq.in =	jq.strt;	/* If so, wraparound */
		if(jq.in == jq.out)	/* Was queue full? */
			{
			jq.in =	jq.end;	/* Put back to original	*/
			return TRUE;	/* Indicate full */
			}
		}
	else if(jq.in == jq.out)	/* Was queue full? */
		{
		--jq.in;		/* Put back to original	*/
		return TRUE;		/* Indicate full */
		}
	return FALSE;
	}





/************************************************************************/
/*									*/
/*	Routine:	read_jq						*/
/*									*/
/*	Read input lines from the line buffers and stuff them into the	*/
/*	jump queue (up to 129 bytes worth of code to check for jump	*/
/*	relative instructions).						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if	something was available	from the input	*/
/*			file, FALSE if a hard EOF.			*/
/*									*/
/************************************************************************/

BOOL	read_jq()
	{
	register char c,*cptr,*oldjqptr;
	register PTRUINT ptr;

	FOREVER
		{
		if(jqcsize)		/* If anything in jump queue */
			{
			if(jqcsize + pb.size > 129) return TRUE;
			}
		cptr = pb.sb;		/* Point to statement buffer */
		oldjqptr = jq.in;	/* Remember where we started from */
		put_jq(pb.type);	/* Enter statement type	*/
		switch(pb.type)
			{
			case NORMTYPE:	/* Instruction size */
				put_jq((char) pb.size);
				put_jq((char) (pb.size >> 8));
#if	UINT4
				put_jq((char) (pb.size >> 16));
				put_jq((char) (pb.size >> 24));
#endif
				break;

			case LABTYPE:
				ptr = (PTRUINT)	add_lab();
				put_jq((char) ptr);
				put_jq((char) (ptr >> 8));
#if	PTR4
				put_jq((char) (ptr >> 16));
				put_jq((char) (ptr >> 24));
#endif
				break;

			case JMPTYPE:
				ptr = (PTRUINT)	add_jmp();
				put_jq((char) ptr);
				put_jq((char) (ptr >> 8));
#if	PTR4
				put_jq((char) (ptr >> 16));
				put_jq((char) (ptr >> 24));
#endif
				break;
#if	DEBUG
			default:
				fatal("Bad parse type","");
#endif
			}
/*
 *	Queue overhead information filled in, now write	the statement itself.
 */
		while((c = *cptr++) != '\0')
			{
			if(put_jq(c))	/* Was match queue full? */
				{
				jq.in =	oldjqptr;	/* Restore */
				switch(pb.type)
					{
					case NORMTYPE:
						break;
					case LABTYPE:
						((struct symtab	*) ptr)->
							info.jinfo.islabel =
								FALSE;
						break;
					case JMPTYPE:
						((struct symtab	*) ptr)->
							info.jinfo.jmpcnt--;
						break;
					}
				if(! jqfull)
					{
					fprintf(stderr,
						"WARNING: Jump queue full");
					jqfull = TRUE;
					}
				return TRUE;
				}
			}
		b_read += pb.size;	/* Update # of code bytes read */
		jqcsize	+= pb.size;	/* One more stmt in queue */
		if(pb.type == JMPTYPE) do_jopts(oldjqptr);
		if(parse()) return FALSE;	/* Hard	EOF */
		}
	return FALSE;			/* Satisfy lint... */
	}





/************************************************************************/
/*									*/
/*	Routine:	scan_jq						*/
/*									*/
/*	Scans the jump queue and returns the character pointed at by	*/
/*	the "queue.scan" pointer.  The pointer is then updated.	 This	*/
/*	routine	is similar to "get_jq",	except the actual queue		*/
/*	pointers are NOT modified.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The next character from	the jump queue.		*/
/*									*/
/************************************************************************/

char	scan_jq()
	{
	char c;

	c = *jq.scan++;			/* Read	character */
	if(jq.scan > jq.end) jq.scan = jq.strt;	/* Handle wraparound */
	return c;
	}





/************************************************************************/
/*									*/
/*	Routine:	usagerr						*/
/*									*/
/*	Report incorrect optimizer invocation to user with a suggestion	*/
/*	of how it should be done.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	usagerr()
	{
	fatal("Usage is	'ao infile [rstfile] [-dfoqs] [outfile]'","");
	}





/************************************************************************/
/*									*/
/*	Routine:	write_jq					*/
/*									*/
/*	Write a	line from the jump queue to the	output file.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if	anything was left in the jump queue to	*/
/*			be written onto	the output file.  FALSE		*/
/*			otherwise.					*/
/*									*/
/************************************************************************/

BOOL	write_jq()
	{
	register BOOL addtab;
	register char c;
	register UINT csize;
	register PTRUINT ptr;

	do
		{
		if(jq.in == jq.out) return FALSE;	/* Queue was empty */
		if(*jq.out == JMPTYPE) do_jopts(jq.out);
		switch(get_jq())	/* Was it a special type? */
			{
			case NORMTYPE:	/* Non jump/label */
				csize =	get_jq() & (UINT) 0xff;
				csize |= (get_jq() & (UINT) 0xff) << 8;
#if	UINT4
				csize |= (get_jq() & (UINT) 0xff) << 16;
				csize |= (get_jq() & (UINT) 0xff) << 24;
#endif
				if(*jq.out == ';') addtab = FALSE;
				else addtab = TRUE;
				break;

			case LABTYPE:
				csize =	0;	/* Generates no	code */
				ptr = get_jq() & (PTRUINT) 0x00ff;
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	8;
#if	PTR4
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	16;
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	24;
#endif
				((struct symtab	*) ptr)->info.jinfo.islabel =
					FALSE;
				addtab = FALSE;
				break;

			case JMPTYPE:
				csize =	3;	/* Jmp's are 3 bytes long */
				ptr = get_jq() & (PTRUINT) 0x00ff;
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	8;
#if	PTR4
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	16;
				ptr |= (get_jq() & (PTRUINT) 0x00ff) <<	24;
#endif
				((struct symtab	*) ptr)->info.jinfo.jmpcnt--;
				addtab = TRUE;
				break;

			case EXJPTYPE:	/* Non jump/label */
				csize =	get_jq() & (UINT) 0xff;
				csize |= (get_jq() & (UINT) 0xff) << 8;
#if	PTR4
				get_jq();	/* Always zero */
				get_jq();	/* Always zero */
#endif
				if(*jq.out == ';') addtab = FALSE;
				else addtab = TRUE;
				break;
#if	DEBUG
			default:
				fatal("Bad jump	queue type","");
#endif
			}
		jqcsize	-= csize;
		do				/* Write statement out */
			{
			if(addtab)
				{
				c = '\t';	/* Prepend tab */
				addtab = FALSE;
				}
			else
				{
				c = get_jq();	/* A char to write out */
				}
			if((putc(c,out_fp)) == EOF)
				{
				fatal("Error in	writing	output file: ",
					ofnbuf);
				}
			} while(c != '\n');
		} while	(! csize);
	return TRUE;
	}

