/************************************************************************/
/*									*/
/*	File:	ao2.c							*/
/*									*/
/*	Assembly language peephole optimizer for use with Z80 versions	*/
/*	of the Q/C "C" compiler.					*/
/*									*/
/*		Written	by Kirk	Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1984-1986	Logical	Systems.		*/
/*									*/
/*				04/20/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<stdlib.h>
#include	<string.h>		/* String functions		*/
#include	<ctype.h>
#ifdef MSDOS
#include	<process.h>
#include	<alloc.h>
#endif
#include	"aodef.h"		/* Optimizer definitions */
#include	"aotyp.h"		/* Optimizer data declarations */
#include	"aoext.h"		/* Optimizer externals */

/************************************************************************/
/*									*/
/*			Mnemonic State Definitions			*/
/*									*/
/************************************************************************/

/*
 *	For "state" variable in	"nextc".
 */
#define	BOL		0		/* Beginning-of-line */
#define	COMMENT		1		/* Within comment */
#define	LABEL		2		/* Within label	*/
#define	LABEL2		3		/* Remember to send extra linefeed */
#define	STMT		4		/* Within statement/label */

/************************************************************************/
/*									*/
/*			Local Data Definitions				*/
/*									*/
/************************************************************************/

/*
 *	Instruction length table.
 */

static	struct	opcode	isize[]	=
		{
			{	".8080"	,	ii	},
			{	".Z80"	,	zero	},
			{	"ADC"	,	size1	},
			{	"ADD"	,	size1	},
			{	"AND"	,	size2	},
			{	"BIT"	,	size3	},
			{	"CALL"	,	size12	},
			{	"CCF"	,	one	},
			{	"CP"	,	size4	},
			{	"CPD"	,	two	},
			{	"CPDR"	,	two	},
			{	"CPIR"	,	two	},
			{	"CPI"	,	two	},
			{	"CPL"	,	one	},
			{	"CSEG"	,	scase3	},
			{	"DAA"	,	one	},
			{	"DEC"	,	size5	},
			{	"DEFB"	,	scase5	},
			{	"DEFS"	,	scase2	},
			{	"DEFW"	,	scase4	},
			{	"DI"	,	one	},
			{	"DJNZ"	,	two	},
			{	"DSEG"	,	scase1	},
			{	"EI"	,	one	},
			{	"END"	,	scase8	},
			{	"ENDM"	,	scase6	},
			{	"EX"	,	size6	},
			{	"EXTRN"	,	zero	},
			{	"EXX"	,	one	},
			{	"HALT"	,	one	},
			{	"IM"	,	two	},
			{	"IN"	,	two	},
			{	"INC"	,	size5	},
			{	"IND"	,	two	},
			{	"INDR"	,	two	},
			{	"INI"	,	two	},
			{	"INIR"	,	two	},
			{	"JP"	,	size7	},
			{	"JR"	,	two	},
			{	"LD"	,	size8	},
			{	"LDD"	,	two	},
			{	"LDDR"	,	two	},
			{	"LDI"	,	two	},
			{	"LDIR"	,	two	},
			{	"NEG"	,	two	},
			{	"NOP"	,	one	},
			{	"OR"	,	size2	},
			{	"OTDR"	,	two	},
			{	"OTIR"	,	two	},
			{	"OUT"	,	two	},
			{	"OUTD"	,	two	},
			{	"OUTI"	,	two	},
			{	"POP"	,	size9	},
			{	"PUBLI"	,	zero	},
			{	"PUSH"	,	size9	},
			{	"REPT"	,	scase7	},
			{	"RES"	,	size3	},
			{	"RET"	,	one	},
			{	"RETI"	,	two	},
			{	"RETN"	,	two	},
			{	"RL"	,	size10	},
			{	"RLA"	,	one	},
			{	"RLC"	,	size10	},
			{	"RLCA"	,	one	},
			{	"RLD"	,	two	},
			{	"RR"	,	size10	},
			{	"RRA"	,	one	},
			{	"RRC"	,	size10	},
			{	"RRCA"	,	one	},
			{	"RRD"	,	two	},
			{	"RST"	,	one	},
			{	"SBC"	,	size11	},
			{	"SCF"	,	one	},
			{	"SET"	,	size3	},
			{	"SLA"	,	size10	},
			{	"SRA"	,	size10	},
			{	"SRL"	,	size10	},
			{	"SUB"	,	size2	},
			{	"XOR"	,	size2	}
		};





/************************************************************************/
/*									*/
/*	Routine:	do_jopts					*/
/*									*/
/*	This routine checks to see if a	jump instruction can be		*/
/*	converted into a jump relative instruction.  Does this by	*/
/*	checking the current contents of the symbol table to see if the	*/
/*	matching label is present, if so convert the jump into a jump	*/
/*	relative.  Note	that the jump queue by definition will never	*/
/*	contain	more bytes worth of instructions than a	relative jump	*/
/*	can reach.							*/
/*									*/
/*	Parameters:							*/
/*		"iptr"	A pointer to the jump instruction in the jump	*/
/*			queue that is being examined.			*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	do_jopts(iptr)
	char	*iptr;
	{
	char	*cptr1,*cptr2;
#if	PTR4
	char	*cptr3,*cptr4;
#endif
	PTRUINT	ptr;

/* fuer	den Fehler im MegaMax-Preprozessor : */
	int dummy;

	jq.scan	= iptr;			/* Start of instruction	to scan	*/
	scan_jq();			/* Skip	type (was JMPTYPE) */
	cptr1 =	jq.scan;		/* Address of low byte of pointer */
	ptr = scan_jq()	& (PTRUINT) 0x00ff;	/* Get pointer to ST */
	cptr2 =	jq.scan;
	ptr |= (scan_jq() & (PTRUINT) 0x00ff) << 8;
#if	PTR4
	cptr3 =	jq.scan;
	ptr |= (scan_jq() & (PTRUINT) 0x00ff) << 16;
	cptr4 =	jq.scan;
	ptr |= (scan_jq() & (PTRUINT) 0x00ff) << 24;
#endif
	if(((struct symtab *) ptr)->info.jinfo.islabel)	/* Can we convert? */
		{
		((struct symtab	*)ptr)->info.jinfo.jmpcnt--;	/* Yes */
		*iptr =	EXJPTYPE;	/* Not convertable now */
		*cptr1 = 2;		/* Instruction is two bytes long */
		*cptr2 = 0;		/* Zero	high byte of length */
#if	PTR4

/* fuer	den Fehler im MegaMax-Preprozessor : */
		dummy=0;

		*cptr3 = 0;
		*cptr4 = 0;
#endif
		b_saved++;		/* We've saved a byte! */
		jqcsize--;		/* One less byte in jump queue */
		scan_jq();		/* Skip	"j" in instruction */
		if(*jq.scan == 'P')
			{
			*jq.scan = 'R';	/* Make	"JP" ->	"JR" */
			}
		else
			{
			*jq.scan = 'r';		/* Make	"jp" ->	"jr" */
			}
		}
	}





/************************************************************************/
/*									*/
/*	Routine:	ii						*/
/*									*/
/*	This routine prints an "illegal	instruction/directive" error	*/
/*	message	and quits.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

UINT	ii()
	{
	fatal("Illegal instruction/directive:\n\t",pb.sb);
	return 0;		     /*	Satisfy	lint */
	}





/************************************************************************/
/*									*/
/*	Routine:	isreg						*/
/*									*/
/*	Used by	the various code size determining routines called from	*/
/*	"size" to check	and see	if the next characters pointed at by	*/
/*	"pb.mptr" satisfy the definition of a Z80 register name.	*/
/*	If so "pb.mptr"	is updated to point beyond the matched item.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if	successfully matched.			*/
/*									*/
/************************************************************************/

BOOL	isreg()
	{
	char	*temp;

	temp = pb.mptr;		/* Remember in case match fails	*/
	switch(toupper(*pb.mptr++))
		{
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'H':
		case 'L':
			break;
		default:
			pb.mptr	= temp;
			return FALSE;
		}
	switch(*pb.mptr)
		{
		case ',':
		case '\n':
		case '\t':
		case ')':
			return TRUE;
		default:
			pb.mptr	= temp;
		}
	return FALSE;
	}





/************************************************************************/
/*									*/
/*	Routine:	isrp						*/
/*									*/
/*	Used by	the various code size determining routines called from	*/
/*	"size" to check	and see	if the next characters pointed at by	*/
/*	"pb.mptr" satisfy the definition of a Z80 register pair	name.	*/
/*	If so "pb.mptr"	is updated to point beyond the matched item.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if	successfully matched.			*/
/*									*/
/************************************************************************/

BOOL	isrp()
	{
	char	*temp;

	temp = pb.mptr;			/* Remember in case match fails	*/
	switch(toupper(*pb.mptr++))
		{
		case 'B':
			if(toupper(*pb.mptr++) != 'C')
				{
				pb.mptr	= temp;
				return FALSE;
				}
			break;
		case 'D':
			if(toupper(*pb.mptr++) != 'E')
				{
				pb.mptr	= temp;
				return FALSE;
				}
			break;
		case 'H':
			if(toupper(*pb.mptr++) != 'L')
				{
				pb.mptr	= temp;
				return FALSE;
				}
			break;
		case 'S':
			if(toupper(*pb.mptr++) != 'P')
				{
				pb.mptr	= temp;
				return FALSE;
				}
			break;
		default:
			pb.mptr	= temp;
			return FALSE;
		}
	switch(*pb.mptr)
		{
		case ',':
		case '\n':
		case '\t':
		case ')':
			return TRUE;
		default:
			pb.mptr	= temp;
		}
	return FALSE;
	}





/************************************************************************/
/*									*/
/*	Routine:	isxrp						*/
/*									*/
/*	Used by	the various code size determining routines called from	*/
/*	"size" to check	and see	if the next characters pointed at by	*/
/*	"pb.mptr" satisfy the definition of a Z80 extended register	*/
/*	pair name.  If so "pb.mptr" is updated to point	beyond the	*/
/*	matched	item.							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	TRUE if	successfully matched.			*/
/*									*/
/************************************************************************/

BOOL	isxrp()
	{
	char	*temp;

	temp = pb.mptr;			/* Remember in case match fails	*/
	if(toupper(*pb.mptr++) != 'I')
		{
		pb.mptr	= temp;
		return FALSE;
		}
	switch(toupper(*pb.mptr++))
		{
		case 'X':
		case 'Y':
			break;
		default:
			pb.mptr	= temp;
			return FALSE;
		}
	switch(*pb.mptr)
		{
		case ',':
		case '\n':
		case '\t':
		case ')':
		case '+':
			return TRUE;
		default:
			pb.mptr	= temp;
		}
	return FALSE;
	}





/************************************************************************/
/*									*/
/*	Routine:	match						*/
/*									*/
/*	Used by	the various code size determining routines called from	*/
/*	"size" to check	and see	if the next characters pointed at by	*/
/*	"pb.mptr" match	the string passed to "match".  If so "pb.mptr"	*/
/*	is updated to point beyond the matched item.			*/
/*									*/
/*	Parameters:							*/
/*		"str"	A pointer to the string	being compared.		*/
/*									*/
/*	Returns:	TRUE if	successfully matched.			*/
/*									*/
/************************************************************************/

BOOL	match(str)
	char *str;
	{
	char	*temp;

	temp = pb.mptr;			/* Remember in case match fails	*/
	while(*str)
		{
		if(*str++ != toupper(*pb.mptr++))
			{
			pb.mptr	= temp;
			return FALSE;
			}
		}
	return TRUE;
	}





/************************************************************************/
/*									*/
/*	Routine:	nc1						*/
/*									*/
/*	A helping function for "nextc".	 Reads input characters	and	*/
/*	filters	out NULL's.  Converts formfeeds	and carriage returns	*/
/*	into newlines.	Note that the Q/C library routines already	*/
/*	convert	carriage-return, linefeed pairs	into linefeeds on	*/
/*	input.								*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The next character read.			*/
/*									*/
/************************************************************************/

int	nc1()
	{
	int c;

	FOREVER				/* Or at least a long time! */
		{
		c = getc(in_fp);	/* Read	character */
		switch(c)
			{
			case NULL:	break;	/* Try again */
			case '\f':
			case '\r':	return '\n';
			default:	return c;
			}
		}
	return c;			/* Make	lint happy */
	}





/************************************************************************/
/*									*/
/*	Routine:	nextc						*/
/*									*/
/*	Get the	next character from the	input file.  This routine makes	*/
/*	the assembly language input into a "standard" form:		*/
/*									*/
/*	1. Input lines which contain labels are	split into a line	*/
/*	containing the label and one containing	the rest of the		*/
/*	statement.							*/
/*	2. If EOF is detected within a line, a linefeed	is prepended to	*/
/*	make it	appear on its own line.					*/
/*	3. Whitespace before opcodes is	eliminated.			*/
/*	4. Trailing comments and whitespace is eliminated.  But	note	*/
/*	that comments that start at the	beginning of a line are	saved.	*/
/*	5. Internal whitespace is shrunk to 1 tab per span (except in	*/
/*	comments which aren't mucked with).				*/
/*	6. The character coercesions described in "nc1"	also apply.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The next virtual character.			*/
/*									*/
/*	Restrictions:	Note that this routine is simpleminded and	*/
/*			has restrictions the assembler doesn't have, IN	*/
/*			PARTICULAR, NO ':' or ';' MAY APPEAR		*/
/*			ANYWHERE EXCEPT	THE APPROPRIATE	PLACE FOR	*/
/*			LABELS,	STATEMENTS AND COMMENTS.  Character	*/
/*			constants and string literals are NOT		*/
/*			supported.  The	compiler doesn't generate these	*/
/*			and any	"#asm" stuff that is run through here	*/
/*			better not either.				*/
/*									*/
/************************************************************************/

int nextc()
    {
    static UINT	state =	BOL;		/* Start at beginning-of-line */
    register int c;

    FOREVER
	{
	switch(state)			/* What	happened last time? */
	    {
	    case BOL:			/* We are starting a new line */
		c = nc1();		/* Get first character */
		while(c	 == ' '	|| c ==	'\t') c	= nc1();
		switch(c)
		    {
		    case '\n':
			break;		/* Skip	blank lines */
		    case ';':
			state =	COMMENT;	/* Whole line comment */
			return c;
		    default:
			state =	STMT;	/* Either a label or a statement */
			return c;
		    }
		break;			/* Cycle to skip blank lines */
	    case COMMENT:		/* We are within a comment */
		switch(c = nc1())
		    {
		    case '\n':		/* We've found the end */
		    case EOF:		/* Or the bitter end */
			state =	BOL;
			return '\n';
		    default:
			return c;
		    }
	    case STMT:			/* Within a label/statement */
		switch(c = nc1())
		    {
		    case '\n':
			state =	BOL;
			return c;
		    case ':':
			state =	LABEL;	/* Must	be a label */
			return c;
		    case EOF:
			state =	BOL;
			return '\n';
		    case ' ':
		    case '\t':
			while(c	== ' ' || c == '\t')	/* Make	1 tab */
			    {
			    switch(c = nc1())
				{
				case '\n':
				case EOF:
				    state = BOL;	/* Trailing spaces */
				    return '\n';
				}
			    }
			ungetc(c,in_fp);	/* Spaces within stmt. */
			return '\t';
		    case ';':
			state =	BOL;	/* Clobber trailing comments */
			while(c	!= '\n')
			    {
			    if((c = nc1()) == EOF) return '\n';
			    }
			return c;
		    default:
			return c;
		    }
	    case LABEL:			/* Doing a label now */
		switch(c = nc1())
		    {
		    case ':':
			state =	LABEL2;	/* Remember to send extra linefeed */
			return c;	/* Must	have been an external */
		    case EOF:
			state =	BOL;	/* Put on own line */
			return '\n';
		    default:
			state =	BOL;	/* Put on own line */
			ungetc(c,in_fp);	/* Part	of next	line */
			return '\n';
		    }
	    case LABEL2:		/* Do extra linefeed */
		state =	BOL;
		return '\n';
	    }
	}
    return c;				/* Make	lint happy */
    }





/************************************************************************/
/*									*/
/*	Routine:	one						*/
/*									*/
/*	This routine just returns "1".	It is used in the instruction	*/
/*	length table to	indicate an instruction	of length 1 (byte).	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	"1".						*/
/*									*/
/************************************************************************/

UINT	one()
	{
	return 1;
	}





/************************************************************************/
/*									*/
/*	Routine:	parse						*/
/*									*/
/*	Read a line of code from the input, figure out the statement	*/
/*	size and type, and put it in a buffer to be processed by	*/
/*	"read_jq".							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	FALSE if something was available from the input	*/
/*			file, TRUE if a	hard EOF.			*/
/*									*/
/************************************************************************/

BOOL	parse()
	{
	register char *cptr;
	register int c;
	register UINT count;

	cptr = pb.sb;			/* Place to put	statement */
	count =	1;			/* Line	length counter */
	while((c = nextc()) != EOF)	/* Read	while we've got	data */
		{
		switch(c)
			{
			case '\n':		/* End of the line */
				*cptr++	= c;
				*cptr =	'\0';	/* Terminate */
				size();		/* Figure out what it was */
				return FALSE;	/* We got something */

			case ':':		/* Must	be a label */
				if(*pb.sb != ';')	/* Not comment */
					{
					*cptr =	'\0';	/* Terminate */
					strncpy(pb.symbuf,pb.sb,SYMNMESIZE);
					pb.symbuf[SYMNMESIZE] =	'\0';
					pb.type	= LABTYPE;
					pb.size	= 0;	/* Labels 0 length */
					*cptr++	= c;
					if((c =	nextc()) == ':') /* Extrn? */
						{
						if(count++ > (MAXLNSIZE	- 1))
							{
				fatal("Line too	long in	input file: ",ifnbuf);
							}
						*cptr++	= c;
						nextc(); /* Flush '\n' */
						}
					*cptr++	= '\n';
					*cptr =	'\0';
					return FALSE;	/* Read	something */
					}

			default:
				*cptr++	= c;
				if(count++ > (MAXLNSIZE	- 1))
					{
				fatal("Line too	long in	input file: ",ifnbuf);
					}
				break;	/* Get next character */
			}
		}
	if(ferror(in_fp))		/* Check if error or EOF */
		{
		fatal("Error in	reading	input file: ",ifnbuf);
		}
	return TRUE;			/* Hard	EOF */
	}





/************************************************************************/
/*									*/
/*	Routine:	scase1						*/
/*									*/
/*	Helping	function for "size" to handle the "dseg" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase1()
	{
	dseg = TRUE;			/* Now in a data segment */
	return 0;
	}





/************************************************************************/
/*									*/
/*	Routine:	scase2						*/
/*									*/
/*	Helping	function for "size" to handle the "defs" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase2()
	{
	return	(UINT) atoi(pb.mptr);
	}





/************************************************************************/
/*									*/
/*	Routine:	scase3						*/
/*									*/
/*	Helping	function for "size" to handle the "cseg" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase3()
	{
	dseg = FALSE;			/* In code segment now */
	return 0;
	}





/************************************************************************/
/*									*/
/*	Routine:	scase4						*/
/*									*/
/*	Helping	function for "size" to handle the "defw" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase4()
	{
	UINT count;

	count =	1;
	while(*pb.mptr)
		{
		if(*pb.mptr++ == ',') count++;
		}
	return (count << 1);
	}





/************************************************************************/
/*									*/
/*	Routine:	scase5						*/
/*									*/
/*	Helping	function for "size" to handle the "defb" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase5()
	{
	UINT count;

	count =	1;
	while(*pb.mptr)
		{
		if(*pb.mptr++ == ',') count++;
		}
	return count;
	}





/************************************************************************/
/*									*/
/*	Routine:	scase6						*/
/*									*/
/*	Helping	function for "size" to handle the "endm" directive.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase6()
	{
	pb.rept	= FALSE;		/* No longer in	macro */
	return 0;
	}





/************************************************************************/
/*									*/
/*	Routine:	scase7						*/
/*									*/
/*	Helping	function for "size" to handle the "rept" directive.	*/
/*	Note that the "rept" is	handled	correctly as to	the number of	*/
/*	bytes of code it generates, BUT	it is not actually expanded.	*/
/*	The instructions that are being	"rept"ed should	not contain	*/
/*	jump or	call instructions or LABELs (even the assembler	won't	*/
/*	like those!).							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase7()
	{
	pb.rept	= atoi(pb.mptr);	/* Establish repeat count */
	return 0;
	}





/************************************************************************/
/*									*/
/*	Routine:	size						*/
/*									*/
/*	Examine	an assembly language instruction/pseudo-op/directive to	*/
/*	determine whether it is	a jump instruction that	is potentially	*/
/*	convertable to a jump relative.	 Also examine the instruction	*/
/*	to determine the length	of the code it generates.		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	size()
	{
	register char *cptr;
	register int result;
	register UINT lower,middle,upper;

	pb.type	= NORMTYPE;		/* Assume non convertable jump */
	if(*pb.sb == ';')
		{
		pb.size	= 0;		/* Whole line comment */
		return;
		}
	strncpy(pb.symbuf,pb.sb,OPNMESIZE);	/* Copy	opcode */
	pb.symbuf[OPNMESIZE] = '\0';	/* Terminate if	necessary */
	if(! (cptr = strchr(pb.symbuf,'\t')))
		{
		if((cptr = strchr(pb.symbuf,'\n')) != NULL) *cptr++ = '\0';
		else cptr = pb.symbuf +	OPNMESIZE;
		}
	else *cptr++ = '\0';		/* Terminate opcode substring */
	pb.mptr	= pb.sb	+ (cptr	- pb.symbuf);	/* Start match at operand */
/*
 *	We have	a string purporting to be an opcode/directive/psuedo-op,
 *	now see	if we recognize	it.  (Binary search opcode table).  The
 *	string is converted to upper case for the matching operation.
 */
	for(cptr = pb.symbuf; *cptr; cptr++) *cptr = toupper(*cptr);

	lower =	0;			/* Lower bound of key */
	upper =	(sizeof(isize) / sizeof(struct opcode))	- 1;	/* Upper */
	while(lower <= upper)		/* While some table to search */
		{
		middle = (lower	+ upper) >> 1;	/* Middle to check */
		result = strncmp(pb.symbuf,isize[middle].oname,OPNMESIZE);
		if(result > 0)
			{
			lower =	middle + 1;	/* Look	in upper half */
			continue;
			}
		else if(result < 0)
			{
			upper =	middle - 1;	/* Look	in lower half */
			continue;
			}
		else				/* We found it!	*/
			{
			pb.size	= (*isize[middle].length)();
			if(dseg)
				{
				pb.size	= 0;	/* If in data segment */
				pb.type	= NORMTYPE;
				}
			else if(pb.rept) pb.size *= pb.rept;
			return;
			}
		}
	ii();				/* Wasn't in table */
	}





/************************************************************************/
/*									*/
/*	Routine:	size1						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		ADC,ADD							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size1()
	{
	if(match("A,"))
		{
		if(match("("))
			{
			if(match("HL)")) return	1;
			return 3;
			}
		if(isreg()) return 1;
		return 2;
		}
	if(isxrp()) return 2;
	return 1;
	}





/************************************************************************/
/*									*/
/*	Routine:	size2						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		AND,OR,SUB,XOR						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size2()
	{
	if(isreg()) return 1;
	if(match("("))
		{
		if(match("HL)")) return	1;
		return 3;
		}
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	size3						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		BIT,RES,SET						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size3()
	{
	pb.mptr	+= 3;
	if(isxrp()) return 4;
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	size4						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		CP							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size4()
	{
	if(match("("))
		{
		if(match("HL)")) return	1;
		return 3;
		}
	if(isreg()) return 1;
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	size5						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		DEC,INC							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size5()
	{
	if(match("("))
		{
		if(match("HL)")) return	1;
		return 3;
		}
	if(isxrp()) return 2;
	return 1;
	}





/************************************************************************/
/*									*/
/*	Routine:	size6						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		EX							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size6()
	{
	if(match("(SP),"))
		{
		if(isxrp()) return 2;
		}
	return 1;
	}





/************************************************************************/
/*									*/
/*	Routine:	size7						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		JP							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size7()
	{
	char *cptr;

	if(match("("))
		{
		if(isrp()) return 1;
		return 2;
		}
	if(match("M,") || match("P,") || match("PE,") || match("PO,"))
		{
		return 3;
		}
	pb.type	= JMPTYPE;		/* Possibly convertable	jump */
	if((cptr = strchr(pb.mptr,','))	!= NULL)  /* Find symbol name */
		{
		strncpy(pb.symbuf,cptr + 1,SYMNMESIZE);
		}
	else
		{
		strncpy(pb.symbuf,pb.mptr,SYMNMESIZE);
		}
	pb.symbuf[SYMNMESIZE] =	'\0';	/* Terminate */
	if((cptr = strchr(pb.symbuf,'\n')) != NULL) *cptr = '\0'; /* Nuke linefeed */
	return 3;
	}





/************************************************************************/
/*									*/
/*	Routine:	size8						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		LD							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size8()
	{
	if(match("("))
		{
		if(isrp())
			{
			pb.mptr	+= 2;
			if(isreg()) return 1;
			return 2;
			}
		if(isxrp())
			{
			while(*pb.mptr)
				{
				if(*pb.mptr++ == ')')
					{
					pb.mptr++;
					if(isreg()) return 3;
					return 4;
					}
				}
			ii();
			}
		while(*pb.mptr)
			{
			if(*pb.mptr++ == ')')
				{
				pb.mptr++;
				if(match("A") || match("HL")) return 3;
				return 4;
				}
			}
		ii();
		}
	if(isreg())
		{
		pb.mptr++;
		if(match("("))
			{
			if(isrp()) return 1;
			return 3;
			}
		if(isreg()) return 1;
		return 2;
		}
	if(match("HL,")) return	3;
	if(match("SP,"))
		{
		if(match("(")) return 4;
		if(isxrp()) return 2;
		if(isrp()) return 1;
		return 3;
		}
	if(isrp())
		{
		if(match(",("))	return 4;
		return 3;
		}
	if(isxrp()) return 4;
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	size9						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		POP,PUSH						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size9()
	{
	if(isxrp()) return 2;
	return 1;
	}





/************************************************************************/
/*									*/
/*	Routine:	size10						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		RL,RLC,RR,RRC,SLA,SRA,SRL				*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size10()
	{
	if(match("("))
		{
		if(isxrp()) return 4;
		}
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	size11						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the following instructions (in bytes):	*/
/*									*/
/*		SBC							*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size11()
	{
	if(isrp()) return 2;
	pb.mptr	+= 2;
	if(match("("))
		{
		if(match("HL)")) return	1;
		return 3;
		}
	if(isreg()) return 1;
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	two						*/
/*									*/
/*	This routine just returns "2".	It is used in the instruction	*/
/*	length table to	indicate an instruction	of length 2 (bytes).	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	"2".						*/
/*									*/
/************************************************************************/

UINT	two()
	{
	return 2;
	}





/************************************************************************/
/*									*/
/*	Routine:	zero						*/
/*									*/
/*	This routine just returns "0".	It is used in the instruction	*/
/*	length table to	indicate an a directive	that generates no code.	*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	"0".						*/
/*									*/
/************************************************************************/

UINT	zero()
	{
	return 0;
	}

