/************************************************************************/
/*									*/
/*	File:	ao3.c							*/
/*									*/
/*	Assembly language peephole optimizer for use with Z80 versions	*/
/*	of the Q/C "C" compiler.					*/
/*									*/
/*		Written	by Kirk	Bailey of Logical Systems.		*/
/*									*/
/*		Copyright (c) 1984-1986	Logical	Systems.		*/
/*									*/
/*				04/20/86				*/
/*									*/
/************************************************************************/

#include	<stdio.h>		/* Standard include file */
#include	<string.h>		/* String functions		*/
#include	<ctype.h>
#ifdef MSDOS
#include	<process.h>
#include	<alloc.h>
#endif
#include	"aodef.h"		/* Optimizer definitions */
#include	"aotyp.h"		/* Optimizer data declarations */
#include	"aoext.h"		/* Optimizer externals */





/************************************************************************/
/*									*/
/*	Routine:	findit						*/
/*									*/
/*	Searches the symbol table for a	specified string.		*/
/*									*/
/*	Parameters:							*/
/*		"cptr"	A pointer to the string	being searched for.	*/
/*									*/
/*	Returns:	TRUE if	the string (symbol) is in the table,	*/
/*			FALSE otherwise.				*/
/*									*/
/************************************************************************/

BOOL	findit(cptr)
	char *cptr;
	{
	register char *tptr;
	register UINT num;
	register struct	symtab *ptr;

	tptr = cptr;			/* Start of symbol being hunted	*/
	num = 0;			/* Zero	hash accumulater */
	while(*tptr) num = (num	* 16) +	*tptr++;	/* Compute hash	*/
	num &= SYMTABMASK;		/* Make	within table */
	ptr = st + num;
	for(num	= 0; num < SYMTABSIZE; num++ , ptr++)
		{
		if(ptr > stend)	ptr = st;	/* Wrap	around */
		if(*ptr->sname)		/* Something here */
			{
			if(! strncmp(cptr,ptr->sname,SYMNMESIZE))
				{
				return TRUE;	/* Found a match */
				}
			continue;	/* Else	try next slot */
			}
		return FALSE;		/* Found an empty slot */
		}
	return FALSE;			/* Checked in whole table */
	}





/************************************************************************/
/*									*/
/*	Routine:	lnk_rst						*/
/*									*/
/*	Generates the restart linkages specified by user.  Examines the	*/
/*	restart	initialization information supplied by the user	and	*/
/*	generates code which either copies a jump into the restart	*/
/*	vector or copies the eight bytes of code located at the	symbol	*/
/*	into the restart vectors.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	lnk_rst()
	{
	char buf[3];
	register BOOL wascopy;
	register UINT i;

	fprintf(out_fp,"\tPUBLIC\t?RST?\n");
	fprintf(out_fp,"?RST?:\n");
	for(i =	0; i < 8; i++)		/* See if any non-copies wanted	*/
		{
		if(*restart[i].rname)	/* This	one used? */
			{
			if(! restart[i].copy)
				{
				fprintf(out_fp,"\tLD\tA,195\n");
				b_added	+= 2;
				break;
				}
			}
		}
	wascopy	= FALSE;		/* Last	restart	wasn't copied into */
	for(i =	0; i < 8; i++)
		{
		if(*restart[i].rname)	/* This	one used? */
			{
			if(! findit(restart[i].rname))	/* Extern? */
				{
				fprintf(out_fp,
					"\tEXTRN\t%s\n",restart[i].rname);
				}
			fprintf(out_fp,"\tLD\tHL,%s\n",restart[i].rname);
			if(restart[i].copy)	/* Copy	linkage? */
				{
				if(! wascopy)
					{
					sprintf(buf,"%u",i * 8);
					fprintf(out_fp,"\tLD\tDE,%s\n",buf);
					b_added	+= 3;
					}
				fprintf(out_fp,"\tLD\tBC,8\n");
				fprintf(out_fp,"\tLDIR\n");
				b_added	+= 8;
				wascopy	= TRUE;
				}
			else
				{
				sprintf(buf,"%u",i * 8);
				fprintf(out_fp,"\tLD\t(%s),A\n",buf);
				sprintf(buf,"%u",(i * 8) + 1);
				fprintf(out_fp,"\tLD\t(%s),HL\n",buf);
				b_added	+= 9;
				wascopy	= FALSE;
				}
			}
		else wascopy = FALSE;
		}
	fprintf(out_fp,"\tJP\t%s\n",entrypt);	/* REAL	start of program */
	b_added	+= 3;			/* 3 byte jump */
	fprintf(out_fp,"\tEND\t?RST?\n");
	if(ferror(out_fp))		/* Any problems	writing? */
		{
		fatal("Error in	writing	output file: ",ofnbuf);
		}
	}





/************************************************************************/
/*									*/
/*	Routine:	r_rst						*/
/*									*/
/*	Read the file containing the restart linkage information and	*/
/*	process	the contents into the "restart"	array.	Determine	*/
/*	whether	the symbols are	to be linked by	jumps at the restart	*/
/*	locations or if	the actual code	at the symbol address is to be	*/
/*	copied into low	memory (up to 8	bytes at least).		*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	None.						*/
/*									*/
/************************************************************************/

VOID	r_rst()
    {
    register char *cptr;
    register int c;
    register UINT count,i;

    cptr = pb.sb;			/* Buffer to parse file	in */
    count = 1;				/* Line	length counter */
    FOREVER
	{
	switch(c = getc(rst_fp))
	    {
	    case ' ':
	    case '\t':
	    case '\f':
	    case '\r':
		continue;		/* Skip	whitespace */

	    case EOF:
		if(ferror(rst_fp))
		    {
		    fatal("Error in reading restart file: ",rfnbuf);
		    }
		if(cptr	== pb.sb) return;	/* If nothing on line */

	    case '\n':
		*cptr =	'\0';		/* Null	terminate line */
		cptr = pb.sb + 1;	/* Flag	char */
		if(*pb.sb < '0'	|| *pb.sb > '7')
		    {
		    fatal("Restart vector must be 0-7 in: ",rfnbuf);
		    }
		count =	*pb.sb - '0';	/* Restart # */
		if(*cptr != '&'	&& *cptr != '+')
		    {
		    fatal("Illegal restart linkage type	in: ",rfnbuf);
		    }
		if(*cptr == '+') restart[count].copy = TRUE;
		else restart[count].copy = FALSE;
		cptr++;			/* Point to symbol */
		for(i =	0; i < 8; i++)	/* See if unique symbol	*/
		    {
		    if(*restart[i].rname)
			{
			if(! strncmp(cptr,restart[i].rname,SYMNMESIZE))
			    {
			    fatal("Duplicate restart symbol: ",cptr);
			    }
			}
		    }
		strncpy(restart[count].rname,cptr,SYMNMESIZE);
		restart[count].rname[SYMNMESIZE] = '\0';
		cptr = pb.sb;		/* Setup for next line */
		count =	1;
		break;
	    default:
		if(count++ > (MAXLNSIZE	- 1))	/* Long? */
		    {
		    fatal("Line	too long in restart file: ",rfnbuf);
		    }
		*cptr++	= c;		/* Store character away	*/
		break;
	    }
	}
    }





/************************************************************************/
/*									*/
/*	Routine:	scase8						*/
/*									*/
/*	Helping	function for "size" to handle "end" directive.		*/
/*	Figures	out if the restart linkages should be initialized and	*/
/*	sets a flag (linkrst) if so.					*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	scase8()
	{
	char	*cptr;

	if(rom || (! rst) || (!	*pb.mptr)) return 0;	/* No init. needed */
	linkrst	= TRUE;			/* Initialize restart linkages */
	strncpy(entrypt,pb.mptr,SYMNMESIZE);	/* Remember old	entry pt. */
	entrypt[SYMNMESIZE] = '\0';	/* If maximum length */
	if((cptr = strchr(entrypt,'\n')) != NULL) *cptr	= '\0';
	strcpy(pb.sb,"CSEG\n");
	dseg = FALSE;			/* In code segment now */
	return 0;
	}





/************************************************************************/
/*									*/
/*	Routine:	size12						*/
/*									*/
/*	This routine is	a helping function for "size" which is used to	*/
/*	determine the length of	the "call" instructions	(in bytes).	*/
/*	If the optimizer is being used to count	the frequency of	*/
/*	routines being called then this	will handle putting the	symbols	*/
/*	in the symbol table and	returning the special instruction type	*/
/*	of "CALLTYPE".	If the optimizer is handling "call" to		*/
/*	restart	instruction conversions	then this routine checks to see	*/
/*	if the symbol matches one of the specified restarts, and makes	*/
/*	the substitution if so.						*/
/*									*/
/*	Parameters:	None.						*/
/*									*/
/*	Returns:	The length in bytes.				*/
/*									*/
/************************************************************************/

UINT	size12()
	{
	char buf[3],*cptr;
	UINT count;

	if(strchr(pb.mptr,','))	return 3;	/* Not unconditional */
	if(freq)				/* Counting frequencies? */
		{
		pb.type	= CALLTYPE;		/* We've got one */
		strncpy(pb.symbuf,pb.mptr,SYMNMESIZE);	/* Put in buffer */
		pb.symbuf[SYMNMESIZE] =	'\0';	/* In case maximum length */
		if((pb.mptr = strchr(pb.symbuf,'\n')) != NULL) *pb.mptr	= '\0';
		return 3;
		}
	if(rst)				/* Substituting	for "call"'s? */
		{
		strncpy(pb.symbuf,pb.mptr,SYMNMESIZE);
		pb.symbuf[SYMNMESIZE] =	'\0';	/* Terminate string */
		if((cptr = strchr(pb.symbuf,'\n')) != NULL) *cptr = '\0';
		for(count = 0; count < 8; count++)	/* Check all 8 */
			{
			if(*restart[count].rname)	/* Used? */
				{
				if(! strncmp(restart[count].rname,pb.symbuf,
					SYMNMESIZE))
					{
					strcpy(pb.sb,"RST\t");
					sprintf(buf,"%u",count << 3);
					strcat(pb.sb,buf);
					strcat(pb.sb,"\n");
					b_read += 2;	/* Adjust up */
					b_saved	+= 2;	/* Saved a bit */
					return 1;
					}
				}
			}
		}
	return 3;			/* Just	a normal "call"	... */
	}

