/**************************************************************************\
*									   *
*   REL.C   -	unido X-Mac, Lese- und Schreib-Utilities fuer		   *
*		Extended-REL-Format					   *
*									   *
*			     aus :	Michael	Roehner, unido ChiPSy	   *
*		Extended rel von :	Andreas Czechanowski, dl4sdc	   *
*									   *
*  Features von Extended rel :						   *
*	- Symbollaenge bis zu 15 Zeichen (theoretisch beliebig)		   *
*	- keine Bitschiebe-Operationen mehr fuer Zugriff noetig,	   *
*	  dadurch schneller						   *
*	- Symbolnamen sichtbar in Datei, z.B. mit "strings *.rel" lesbar   *
*	- Nachteil : Dateien werden etwas groesser			   *
*									   *
\**************************************************************************/


/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>		/* Standard-I/O	benutzt		  */
#include <string.h>		/* Stringfunktionen benutzt	  */
#include "all.h"		/* allgemeine Festlegungen	  */
#include "rel.h"		/* Aufbau REL			  */


/*								Globals	  */
/**************************************************************************/

FILE	   *inrel;	      /* Eingabe-REL-Datei			  */
FILE	   *outrel;	      /* Ausgabe-REL-Datei			  */

unsigned    item;	      /* aktueller Item-Code			  */
unsigned    type;	      /* Typfeld				  */
unsigned    field;	      /* aktuelles Bitfeld			  */
unsigned    iacc = 0;	      /* Hilfsflag fuer aufeinanderfolgende ABS's */
unsigned    oacc = 0;	      /* Hilfsflag fuer aufeinanderfolgende ABS's */
unsigned    icount = 0;	      /* Anzahl der noch folgenden abs. Eintr.    */
unsigned    ocount = 0;	      /* Anzahl der zu schreibenden abs. EIntr.   */

char	    symbol[17];	      /* aktueller Symbolstring	(0-terminiert)	  */
char	    obuf[128];	      /* buffer fuer absolute Eintraege		  */



/**************************************************************************\
*									   *
* action      :	 Naechstes REL-Item aus	Datei inrel holen.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 inrel	 -  geoeffnete REL-Lesedatei			   *
*									   *
* r/w globals :	 icount  -  Anzahl der noch folgenden abs. Eintraege	   *
*		 iacc	 -  Flag fuer Folge von absoluten Eintraegen	   *
*		 item	 -  Item-Code					   *
*		 type	 -  Feld-Typ					   *
*		 field	 -  Feld-Wert					   *
*		 symbol	 -  Symbolname					   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Item-Code im Erfolgsfall,				   *
*		 ERR sonst.						   *
*									   *
\**************************************************************************/

int getrel()
{
    int c;

    if ((c = getc(inrel)) == EOF) return FALSE;

    if (iacc) {
	field = c;
	if (!(--icount)) iacc = 0;
	return (type = item = ABS);
    } else if ((c & ABSTOK_N) == ABSTOK) {
	icount = (c & 0x7f) - 1;
	if (icount > 0) iacc = 1;
	if ((field = getc(inrel)) == EOF) return (ERR);
	return (type = item = ABS);
    } else if ((c & RELTOK_N) == RELTOK) {
	type = item = (c >> 4) & 0x03;
	if (getword() == FALSE) return (ERR);
	return (item);
    } else if ((c & SPECTOK_N) == SPECTOK) {
      item = (c & 0x0f) + 4;
      type = (c >> 4) & 0x03;
      switch (item)
      {
	case ENAME:			/* Einsprungpunkt		  */
	case CNAME:			/* Commonblock auswaehlen	  */
	case PNAME:			/* Programmname			  */
	case LNAME:			/* Bibliothekssuche erforderlich  */
	case EXT:			/* erweitertes Link-Item	  */
	  if (getsym() == FALSE) return (ERR);
	  return (item);

	case CSIZE:			/* Common-Groesse definieren	  */
	case XCHAIN:			/* Anfang einer	Verweiskette	  */
	case EPOINT:			/* Einsprungspunkt definieren	  */
	  if (getword() == FALSE || getsym() == FALSE) return (ERR);
	  return (item);

	case XMOFF:			/* extern - Offset		  */
	case XPOFF:			/* extern + Offset		  */
	case DSIZE:			/* Groesse des Datenbereichs	  */
	case SETLC:			/* Positionszaehler Laden setzen  */
	case CHAIN:			/* Verkettungsadresse		  */
	case PSIZE:			/* Programmgroesse		  */
	case EPROG:			/* Programmende			  */
	  if (getword() == FALSE) return (ERR);
	  return(item);

	case EFILE:			/* Dateiende			  */
	  return (item);
      }
      return (ERR);
    }
    return (ERR);			/* dieser Punkt wird nie erreicht */
}


/**************************************************************************
* 16-bit-Wert von Eingabedatei in field in holen, low byte first	  *
* return-Wert : TRUE bei fehlerfreier Ausfuehrung, sonst FALSE.		  *
**************************************************************************/
int getword()
{
    int c,d;

    if (((c = getc(inrel)) == EOF) || ((d = getc(inrel)) == EOF))
	return FALSE;
    field = c | (d << 8);
    return TRUE;
}



/**************************************************************************\
*									   *
* action      :	 Symbol	holen, Itemcode	zurueckgeben.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 item	-  Itemcode, der zurueckgegeben	wird		   *
*									   *
* r/w globals :	 symbol	-  zu setzendes	Symbol				   *
*		 inrel	-  (FILE *) fuer input-Datei			   *
*									   *
* locals      :	 i	-  Zaehler Symbollaenge				   *
*		 cp	-  Zeiger in Symbol				   *
*									   *
* returns     :	 Itemcode des geholten Feldes oder ERR,	wenn Fehler	   *
*		 auftraten						   *
*									   *
\**************************************************************************/

int getsym()
  {
    int	    i;
    int	    cnt;
    char   *cp;

    cnt = 0;
    cp = symbol;			/* Symbolanfang			  */
    do { 
	if ((i = getc(inrel)) == EOF)
	    return(FALSE);
	*cp++ = i;			/* alles bis einschl. \0 kopieren */
	if (cnt++ > MAXSYM) return (FALSE);
    } while (i != '\0');
    return (TRUE);			/* Itemcode zurueckgeben	  */
  }


/**************************************************************************\
*									   *
* action      :	 Schreiben eines REL-Eintrages in relokatible		   *
*		 REL-Objektdatei.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 item	    - Item-Code					   *
*		 type	    - Typ des Felds				   *
*		 field	    - Wert des Felds				   *
*		 symbol	    - Name des Symbols				   *
*									   *
* r/w globals :	 outrel	    - Ausgabe-REL-Datei				   *
*		 oacc	    - Flag fuer aufeinanderfolgende abs. Eintraege *
*		 obuf	    - Buffer fuer abs. Eintraege		   *
*		 ocount	    - Anzahl der Elemente in obuf		   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 TRUE im Erfolgsfall, FALSE sonst			   *
*									   *
\**************************************************************************/

BOOLEAN	putrel()
{
    int i;

    if (item == ABS) {			/* einfache Elemente : 0 <8 bit>  */
	if (oacc) {			/* in obuf[] bis zu 127 bytes	  */
	    obuf[ocount++] = field;	/* sammeln, da ABS-items sehr oft */
	    if (ocount >= 127) {	/* vorkommen.			  */
		putbyte(ABSTOK | ocount);
		for (i=0; i<ocount; i++) putbyte(obuf[i]);
		ocount = 0;
	    }
	} else {
		ocount = 1;
		obuf[0] = field;
		oacc = 1;
	}
	return (TRUE);

    } else if (ocount > 0) {
	putbyte(ABSTOK | ocount);
	for (i=0; i<ocount; i++) putbyte(obuf[i]);
    }
    ocount = 0;
    oacc = 0;
    switch (item)
      {
	case PREL:			/*   1 01 <16 bit>		  */
	case DREL:			/*   1 10 <16 bit>		  */
	case CREL:			/*   1 11 <16 bit>		  */
	   if (putbyte(RELTOK | (item << 4)) == FALSE) return (FALSE);
	   return(putword());

	case CSIZE:			/*   ... tt <16	bit> lll <string> */
	case XCHAIN:
	case EPOINT:
	   putbyte(SPECTOK | (type << 4) | (item - 4));
	   if (putword() == FALSE) return (FALSE);
	   if (putsym() == FALSE) return (FALSE);
	   return (TRUE);

	case ENAME:			/*   ... lll <string>		  */
	case CNAME:
	case PNAME:
	case LNAME:
	case EXT:
	   putbyte(SPECTOK | (item - 4));
	   if (putsym() == FALSE) return (FALSE);
	   return (TRUE);

	case XMOFF:			/*   ... tt <16	bit>		  */
	case XPOFF:
	case DSIZE:
	case SETLC:
	case CHAIN:
	case PSIZE:
	case EPROG:
	   putbyte(SPECTOK | (type << 4) | (item - 4));
	   return (putword());

	case EFILE:			/*   Dateiende			  */
	   return (putbyte(SPECTOK | (item - 4)));
      }

   return (FALSE);
}

/**************************************************************************
* 8-bit-Wert in v in Ausgabedatei schreiben				  *
* return-Wert : TRUE bei fehlerfreier Ausfuehrung, sonst FALSE.		  *
**************************************************************************/
int putbyte(v)
int v;
{
    if (putc(v,outrel) == EOF) return FALSE;
    return TRUE;
}

/**************************************************************************
* 16-bit-Wert in field in Ausgabedatei schreiben, low byte first	  *
* return-Wert : TRUE bei fehlerfreier Ausfuehrung, sonst FALSE.		  *
**************************************************************************/
int putword()
{
    if ((putc(field & 0xff, outrel) == EOF) || (putc(field >> 8, outrel) == EOF))
	return FALSE;
    return TRUE;
}





/**************************************************************************\
*									   *
* action      :	 Symbol	symbol auf Ausgabe-REL-Datei outrel ausgeben.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 symbol	    - 0-terminierter String des	zu schreibenden	   *
*			      Symbols					   *
*									   *
* r/w globals :	 outrel	    - Ausgabe-REL-Datei				   *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 TRUE im Erfolgsfall, FALSE sonst			   *
*									   *
\**************************************************************************/

BOOLEAN	putsym()
{
    int	    i;				     /*	Stringlaenge, Index	  */
    char   *cp;				     /*	Zeiger in den String	  */

    i = 0;
    cp = symbol;
    while (i++ < MAXSYM && (*cp))
	if (putc(*cp++, outrel) == EOF) return FALSE;
    if (putc('\0', outrel) == EOF) return FALSE;
    return (TRUE);			     /*	alles ok		  */
}

/**************************************************************************
* Eingabe der inrel-Datei initialisieren				  *
**************************************************************************/
void inrelinit()
{
    iacc = 0;
    icount = 0;
}

/**************************************************************************
* Ausgabe der outrel-Datei initialisieren				  *
**************************************************************************/
void outrelinit()
{
    oacc = 0;
    ocount = 0;
}


/* Ende	von REL.C */
