/**************************************************************************\
*									   *
*   SYSIBMTC.C -   unido CHiPSy, Systemanpassungen fuer	IBM PC		   *
*									   *
*		   Compiler: Turbo C					   *
*									   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* FILE, NULL, EOF, ...		  */
#include <alloc.h>
#include <stdlib.h>
#include <conio.h>


/*						   Modulinterne	Globals	  */
/**************************************************************************/

static char	 *memstart;		/* Beginn initialer Freispeicher  */
static unsigned	  memfree;		/* # Bytes initialer Freispeicher */
static char	 *reststart;		/* Beginn Rest-Freispeicher	  */
static unsigned	  restfree;		/* # Bytes Rest-Freispeicher	  */





/**************************************************************************\
*									   *
* action      :	 Allokieren eines zusammenhaengenden Speicherblocks.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 size	-  Anzahl der zu allokierenden Bytes		   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Zeiger	auf den	allokierten Speicherbereich oder NULL,	   *
*		 wenn das Allokieren fehlschlug				   *
*									   *
\**************************************************************************/

/* farmalloc() in TurboC vorhanden */





/**************************************************************************\
*									   *
* action      :	 Groesse des groessten durch farmalloc() zu allokierenden  *
*		 zusammenhaengenden Speicherbereiches.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Anzahl	Bytes des groessten Speicherbereiches (s.o.)	   *
*									   *
\**************************************************************************/

/* farcoreleft() in TurboC vorhanden */





/**************************************************************************\
*									   *
* action      :	 Initialisieren	des eigenen Memory-Allocators clrsbrk().   *
*									   *
*		 Programmabbruch mit Meldung, wenn nicht genug		   *
*		 Freispeicher vom Betriebssystem angefordert werden kann.  *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 memstart   -  Beginn initialer	Freispeicher		   *
*		 memfree    -  Anzahl Bytes Laenge initialer Freispeicher  *
*		 reststart  -  Zeiger auf Anfang Restfreispeicher	   *
*		 restfree   -  Anzahl Bytes noch frei im Restfreispeicher  *
*									   *
* locals      :	 n	    -  Laufvariable Freispeicher Loeschen	   *
*		 p	    -  Zeiger in Freispeicher fuer Loeschen	   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void initsbrk()
  {
    unsigned	 n;
    char	 *p;

    n =	restfree = memfree = (unsigned)	50000L;
    reststart =	memstart = malloc(memfree);
    if (memstart == NULL)
      {
	printf("FATAL ERROR: cannot allocate sbrk space	");
	printf("(request: %ld, there is: %ld)\n",memfree, coreleft());
	exit(1);
      }
    else
      for (p = memstart; n > 0L; --n); *p++=0;
  }





/**************************************************************************\
*									   *
* action      :	 Zeiger	auf ausgenullten Speicherbereich der Laenge	   *
*		 request zurueckgeben, immer gerade Anzahl Bytes	   *
*		 allokieren.						   *
*									   *
*		 Diese Funktion	wird anstatt der Standardfunktion sbrk()   *
*		 benutzt, die beim Atari aufgrund der voll dynamischen	   *
*		 Speicherverwaltung auf	Betriebssystemebene bei	zu vielen  *
*		 Aufrufen Probleme machen kann (nur begrenzte Anzahl	   *
*		 Aufrufe innerhalb eines Programms moeglich).		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 request    -  Anzahl Bytes angeforderter Freispeicher	   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 reststart  -  Zeiger auf Anfang Restfreispeicher	   *
*		 restfree   -  Anzahl Bytes noch frei im Restfreispeicher  *
*									   *
* locals      :	 ret	    -  Zeiger auf angeforderten	Speicher	   *
*									   *
* returns     :	 ret, wenn die Anforderung erfuellt werden kann, -1 sonst  *
*									   *
\**************************************************************************/

char *clrsbrk(request)

unsigned request;

  {
    char *ret;

    if (request	& 0x0001) ++request;
    if (restfree >= request)
      {
	ret = reststart;
	reststart += request;
	restfree -= request;
	return (ret);
      }
    else
      return ((char *)(-1));
  }





/**************************************************************************\
*									   *
* action      :	 Tastatur abfragen, wenn ein Zeichen eingegeben	wurde,	   *
*		 Rueckgabe dieses Zeichens, sonst Rueckgabe von	EOF.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 EOF, wenn keine Eingabe erfolgt ist, das eingegebene	   *
*		 Zeichen sonst						   *
*									   *
\**************************************************************************/

inkey(void)
  {
    if (kbhit())
      return (getch());
    else
      return (EOF);
  }





/**************************************************************************\
*									   *
* action      :	 Tastatur-Zeichenzwischenspeicher leeren.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void kbclear()
  {
    while(inkey() != EOF) ;
  }



/* Ende	von SYSIBM.C */
