/**************************************************************************\
*									   *
*   UTI.C   -	unido X-Mac, Utilities					   *
*									   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* Standardausgabe benutzt	  */
#include <stdlib.h>
#include <ctype.h>
#include "all.h"			/* allgemeine Festlegungen	  */



BOOLEAN	    lexorder __ARGS((char, char));
BOOLEAN	    atend __ARGS((int));


/**************************************************************************\
*									   *
* action      :	 Unsigned Integer von Adresse a	holen und zurueckgeben,	   *
*		 es wird in high-Byte /	low-Byte Reihenfolge		   *
*		 einzelbyteweise geholt.				   *
*									   *
*		 Unabhaengig vom Prozessor und von der vom Compiler	   *
*		 vorgegebenen Integerlaenge werden die mit getuint() und   *
*		 putuint() behandelten Integerzahlen immer als 16-bit	   *
*		 high Byte / low Byte abgespeichert. So	koennen	Tabellen,  *
*		 die Integers sequentiell beinhalten, portabel gehalten	   *
*		 werden	(ACHTUNG: INTSZ	in all.h immer 2).		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 a - Zeiger auf	die zu holende Integerzahl		   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 die an	*a stehende Integerzahl				   *
*									   *
\**************************************************************************/

unsigned getuint(a)

unsigned char *a;

  {
    return ((unsigned)((*a << 8) | (*(a	+ 1) & 0x00FF)));
  }





/**************************************************************************\
*									   *
* action      :	 Unsigned Integer u an Adresse a speichern, es wird	   *
*		 in high-Byte /	low-Byte Reihenfolge einzelbyteweise	   *
*		 gespeichert.						   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 a - Adresse, an die u geschrieben werden soll		   *
*		 u - abzuspeichernde Integerzahl			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID putuint(a,u)

unsigned char	*a;
unsigned	 u;

  {
    *a++ = u >>	8;
    *a	 = u;
  }





/**************************************************************************\
*									   *
* action      :	 Zahl nbr rechtsbuendig	mit fuehrenden Blanks in String	   *
*		 der Laenge sz nach str	umwandeln.			   *
*									   *
*		 itou()	 -  vorzeichenloser Dezimalstring		   *
*		 itox()	 -  Hexadezimalstring				   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 nbr	-  umzuwandelnde Zahl				   *
*		 str	-  Zeiger auf Stringbuffer fuer	Zahlenstring	   *
*		 sz	-  > 0	-> str hinter Zahl mit Null abschliessen   *
*			   = 0	-> Zahl	ans rechte Ende	von str	schreiben  *
*			   < 0	-> str hinter Zahl nicht mit Null abschl.  *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 itou()	:  lowbit -  Bitzwischenspeicher		   *
*		 itox()	:  digit  -  Ziffernzwischenspeicher		   *
*									   *
* returns     :	 Zeiger	auf str						   *
*									   *
\**************************************************************************/



char *itou(nbr,str,sz)

int    nbr;
char   str[];
int    sz;

  {
    int	lowbit;

    if (sz > 0)	str[--sz] = '\0';
    else
      if (sz < 0) sz = -sz;
      else
	while (str[sz] != '\0')	++sz;

    while (sz)
      {
	lowbit = nbr & 1;
	nbr = (nbr >> 1) & 0xEFFF;
	str[--sz] = ((nbr % 5) << 1) + lowbit +	'0';
	if ((nbr = nbr / 5) == 0) break;
      }
    while (sz) str[--sz] = ' ';
    return (str);
  }



char *itox(nbr,str,sz)

int    nbr;
char   str[];
int    sz;

  {
    int	digit;

    if (sz > 0)	str[--sz] = '\0';
    else
      if (sz < 0) sz = -sz;
      else
	while (str[sz] != '\0')	++sz;

    while (sz)
      {
	digit =	nbr & 0xF;
	nbr = (nbr >> 4) & 0xFFF;
	str[--sz] = "0123456789ABCDEF"[digit];
	if (nbr	== 0) break;
      }
    while (sz) str[--sz] = ' ';
    return (str);
  }





/**************************************************************************\
*									   *
* action      :	 Vorzeichenlosen Ziffernstring ...str in Zahl umwandeln	   *
*		 und in	*nbr speichern.					   *
*									   *
*		 btoi()	 -  Binaerzahl					   *
*		 otoi()	 -  Oktalzahl					   *
*		 utoi()	 -  Dezimalzahl					   *
*		 xtoi()	 -  Hexadezimalzahl				   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 ...str	 -  Zahlenstring				   *
*		 nbr	 -  Zeiger auf Ergebnisvariable			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 d	-  Laenge der Zahl im String			   *
*		 t	-  Ergebniszwischenspeicher			   *
*									   *
* returns     :	 ERR wenn Fehler entdeckt wurden,			   *
*		 sonst die Laenge der Zahl im String			   *
*									   *
\**************************************************************************/



int btoi(binstr,nbr)

char   *binstr;
int    *nbr;

  {
    int	d;
    int	t;

    d =	0;
    *nbr = 0;
    while (*binstr == '0' || *binstr ==	'1')
      {
	t = *nbr;
	t = (t << 1) + (*binstr++ - '0');
	if (t >= 0 && *nbr < 0)	return (ERR);
	++d;
	*nbr = t;
      }
    return (d);
  }



int otoi(octstr,nbr)

char   *octstr;
int    *nbr;

  {
    int	d;
    int	t;

    d =	0;
    *nbr = 0;
    while (*octstr >= '0' && *octstr <=	'7')
      {
	t = *nbr;
	t = (t << 3) + (*octstr++ - '0');
	if (t >= 0 && *nbr < 0)	return (ERR);
	++d;
	*nbr = t;
      }
    return (d);
  }



int utoi(decstr,nbr)

char   *decstr;
int    *nbr;

  {
    int	d;
    int	t;

    d =	0;
    *nbr = 0;
    while (isdigit(*decstr))
      {
	t = *nbr;
	t = (10	* t) + (*decstr++ - '0');
	if (t >= 0 && *nbr < 0)	return (ERR);
	++d;
	*nbr = t;
      }
    return (d);
  }



int xtoi(hexstr,nbr)

char   *hexstr;
int    *nbr;

  {
    int	d;
    int	t;

    d =	0;
    *nbr = 0;
    LOOP
      {
	if (isdigit(*hexstr))
	  t = '0';
	else
	  if ((*hexstr >= 'A') && (*hexstr <= 'F'))
	    t =	'A' - 10;
	  else
	    if ((*hexstr >= 'a') && (*hexstr <=	'f'))
	      t	= 'a' -	10;
	    else
	      break;

/*
	if (d <	4)
	  ++d;
	else
	  return (ERR);
*/
	++d;

	*nbr = *nbr << 4;
	*nbr = *nbr + (*hexstr++) - t;
      }
    return (d);
  }





/**************************************************************************\
*									   *
* action      :	 Im String str linksbuendige Leerzeichen (Spaces)	   *
*		 loeschen (String linksbuendig ausrichten).		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 str	-  Zeiger auf den String			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 str2	-  lokaler Stringzeiger				   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID left(str)

char *str;

  {
    char *str2;

    str2 = str;
    while (*str2 == ' ') ++str2;
    while ((*str++ = *str2++) == TRUE);
  }





/**************************************************************************\
*									   *
* action      :	 Lexikografischer Vergleich der	Zeichen	c1 und c2.	   *
*									   *
*		 Lexikografisch	heisst,	die Ordnung der	Zeichen	nicht wie  *
*		 vom ASCII-Code	vorgegeben, sondern in aufsteigender	   *
*		 Reihenfolge :						   *
*									   *
*				 Controlcodes				   *
*				 Sonderzeichen				   *
*				 Zahlen					   *
*				 Buchstaben (gross gleich klein)	   *
*				 DEL					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 c1	-  1. Zeichen					   *
*		 c2	-  2. Zeichen					   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 lex	-  Array der lexikografischen Zeichenordnung	   *
*									   *
* returns     :	 < 0  wenn  c1 < c2					   *
*		   0  wenn  c1 = c2					   *
*		 > 0  wenn  c1 > c2					   *
*									   *
\**************************************************************************/

int lexorder(c1,c2)

char c1;
char c2;

  {
    static char	lex[128] =
      {
	/****************   NUL	- /   ******************/
	  0,   1,   2,	 3,   4,   5,	6,   7,	  8,   9,
	 10,  11,  12,	13,  14,  15,  16,  17,	 18,  19,
	 20,  21,  22,	23,  24,  25,  26,  27,	 28,  29,
	 30,  31,  32,	33,  34,  35,  36,  37,	 38,  39,
	 40,  41,  42,	43,  44,  45,  46,  47,
	/******************   0	- 9   ******************/
	 65,  66,  67,	68,  69,  70,  71,  72,	 73,  74,
	/**************	  : ; <	= > ? @	  **************/
	 48,  49,  50,	51,  52,  53,  54,
	/******************   A	- Z   ******************/
	 75,  76,  77,	78,  79,  80,  81,  82,	 83,  84,
	 85,  86,  87,	88,  89,  90,  91,  92,	 93,  94,
	 95,  96,  97,	98,  99, 100,
	/***************   [ \ ] ^ _ `	 ***************/
	 55,  56,  57,	58,  59,  60,
	/******************   a	- z   ******************/
	 75,  76,  77,	78,  79,  80,  81,  82,	 83,  84,
	 85,  86,  87,	88,  89,  90,  91,  92,	 93,  94,
	 95,  96,  97,	98,  99, 100,
	/***************   { | } ~ DEL	 ***************/
	 61,  62,  63,	64, 101
      };

    return (lex[(unsigned char)c1] - lex[(unsigned char)c2]);
  }





/**************************************************************************\
*									   *
* action      :	 Ausgabe einer Fehlermeldung auf stdout	und Programm	   *
*		 abbrechen. Ausgegeben wird "FATAL ERROR: ", dann die	   *
*		 beiden	Strings	s1 und s2, dann	Newline.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 s1 - Zeiger auf String	des 1. Teils der Fehlermeldung	   *
*		 s2 - Zeiger auf String	des 2. Teils der Fehlermeldung	   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 (Errorcode 1 bei Programmabbruch)			   *
*									   *
\**************************************************************************/

VOID fatal(s1,s2)

char *s1;
char *s2;

  {
    printf("FATAL ERROR: %s%s\n",s1,s2);	  /* Meldung ausgeben	  */
    exit(1);					  /* Programm abbrechen	  */
  }





/**************************************************************************\
*									   *
* action      :	 Test, ob ch ein Zeilendezeichen (Kommentarzeichen,	   *
*		 Stringende oder Zeilenende) ist.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 ch	-  zu testendes	Zeichen				   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 YES - ch ist Zeilenendezeichen				   *
*		 NO  - sonst						   *
*									   *
\**************************************************************************/

BOOLEAN	atend(ch)

int ch;

  {
    switch (ch)
      {
	case COMMENT :
	case NULL    :
	case '\n'    :	return (YES);
      }
    return (NO);
  }





/**************************************************************************\
*									   *
* action      :	 Stringfelder lexikografisch vergleichen.		   *
*									   *
*		 Feld =	durch String/Zeilenende	oder isspace-Zeichen	   *
*			beendetes Feld in einem	String			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 s	-  Zeiger auf erstes Feld			   *
*		 t	-  Zeiger auf zweites Feld			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 < 0  wenn  s <	t					   *
*		   0  wenn  s =	t					   *
*		 > 0  wenn  s >	t					   *
*									   *
\**************************************************************************/

int fldcmp(s,t)

char *s;
char *t;

  {
    while (lexorder(*s,*t) == 0)
      {
	if (!isgraph(*s)) return (0);
	++s;
	++t;
      }
    if ((isspace(*s) ||	atend(*s)) && (isspace(*t) || atend(*t)))
      return (0);
    return (*s - *t);
  }





/**************************************************************************\
*									   *
* action      :	 Zeiger	auf n-tes nichtleeres Feld in str zurueckgeben.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 n	-  s.o.						   *
*		 str	-  Zeiger auf Zeichenstring			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 cp	-  Zeiger auf String				   *
*									   *
* returns     :	 Zeiger	auf das	n-te nichtleere	Feld in	str.		   *
*									   *
\**************************************************************************/

char *skip(n,str)

int    n;
char   str[];

  {
    char *cp;

    cp = str;
    while (isspace(*cp)) ++cp;
    while (--n)
      {
	while (isgraph(*cp)) ++cp;
	while (isspace(*cp)) ++cp;
      }
    return (cp);
  }



/* Ende	von UTI.C */
