/**************************************************************************\
*									   *
*   MAC3.C   -	  unido	X-Mac, Teil 3 von 5				   *
*									   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* Standardausgabe benutzt	  */
#include <string.h>			/* Stringfunktionen benutzt	  */
#include <ctype.h>
#include "all.h"			/* allgemeine Festlegungen	  */
#include "mac.h"			/* Festlegungen	fuer MAC	  */
#include "mace.h"			/* Externals MAC		  */
#include "rel.h"			/* Aufbau REL			  */
#include "uti.h"
#include "rele.h"			/* Externals REL.C		  */
#include "mite.h"			/* Externals MIT.C / TAB.C	  */

/*------------------------------------------------------------------------*/
/*---  Prototypen der verwendeten nicht-integer	Funktionen	       ---*/
/*------------------------------------------------------------------------*/
void dodat __ARGS((int));
void doend __ARGS((void));
void doext __ARGS((void));
void doent __ARGS((void));
void doloc __ARGS((int));
void docseg __ARGS((void));
void dodseg __ARGS((void));
void domac2 __ARGS((char *));
void doval __ARGS((int));

/**************************************************************************\
*									   *
* action      :	 Das Symbol, auf das lp	zeigt, daraufhin untersuchen	   *
*		 ob es eine Pseudoinstruktion ist. Wenn	ja, dann Token	   *
*		 davon zurueckgeben.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 lp	-  Eingang: Zeiger auf zu untersuchendes Symbol	in *
*				    Source Zeile			   *
*			   Ausgang: Zeiger hinter das Symbol aus das	   *
*				    naechste Nicht-Space-Zeichen	   *
*									   *
* locals      :	 cp	-  Zeiger in Sourcezeile			   *
*									   *
* returns     :	 ERR, wenn nicht gefunden, sonst s.u.			   *
*									   *
\**************************************************************************/

int aifind()
  {
    char *cp;

    cp = lp;				/* Anfang fuer Vergleich	  */
    while (isgraph(*lp)) ++lp;		/* Ausgang: Zeiger auf naechstes  */
    while (isspace(*lp)) ++lp;		/*	    Symbol		  */

	 if (fldcmp(cp,"DW")	    == 0) return (DW);	    /* Pseudoop	  */
    else if (fldcmp(cp,"DEFW")	    == 0) return (DW);	    /* finden und */
    else if (fldcmp(cp,"DB")	    == 0) return (DB);	    /* Token	  */
    else if (fldcmp(cp,"DEFB")	    == 0) return (DB);	    /* zurueck	  */
    else if (fldcmp(cp,"EXTRN")	    == 0) return (EX);
    else if (fldcmp(cp,"EQU")	    == 0) return (EQU);
    else if (fldcmp(cp,"DS")	    == 0) return (DS);
    else if (fldcmp(cp,"DEFS")	    == 0) return (DS);
    else if (fldcmp(cp,"CSEG")	    == 0) return (CSEG);
    else if (fldcmp(cp,"DSEG")	    == 0) return (DSEG);
    else if (fldcmp(cp,"PUBLIC")    == 0) return (GLOBAL);
    else if (fldcmp(cp,"ORG")	    == 0) return (ORG);
    else if (fldcmp(cp,"END")	    == 0) return (END);
    else if (fldcmp(cp,".Z80")	    == 0) return (IGNORE);
    else if (fldcmp(cp,"EXT")	    == 0) return (EX);
    else if (fldcmp(cp,"EXTERNAL")  == 0) return (EX);
    else if (fldcmp(cp,"SET")	    == 0) return (SET);
    else if (fldcmp(cp,"GLOBAL")    == 0) return (GLOBAL);
    return (ERR);					    /* unbekannt  */
  }





/**************************************************************************\
*									   *
* action      :	 Pseudo-Op in der Sourcezeile line ausfindig machen und	   *
*		 ausfuehren.						   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 gotlabel -  TRUE: 1. Feld in Zeile war	Label		   *
*		 line	  -  die aktuelle Sourcezeile			   *
*									   *
* r/w globals :	 stsym	  -  "", wenn kein Symbol im ersten Feld	   *
*									   *
* locals      :	 j	  -  Pseudoop-Token				   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doasm()
  {
    int	j;

    if (    atend(*lp)			/* falls am Sourcezeilenende und  */
	 && (!stsym[0] || gotlabel)	/* kein	Symbol da oder das Symbol */
       ) return;			/* ist Label, nichts tun	  */
    if ((j = aifind()) != ERR)		/* falls Pseudoop im 2.	Feld,  != */
      {					/* SET/EQU, dann muss erstes Feld */
	if (	!gotlabel		/* ein Label gewesen sein, aber	  */
	     &&	j != SET		/* Pseudoop ggf. ok, daher weiter */
	     &&	j != EQU
	   ) laberr();
      }
    else
      {					/* falls Pseudoop nicht	im 2.	  */
	lp = skip(1,line);		/* dann	1. Feld	untersuchen in	  */
	j = aifind();			/* Sourcezeile und		  */
	stsym[0] = '\0';		/* dann	kein Label in der Zeile	  */
      }
    switch (j)
      {
	case EX	    :  doext();	       return;	  /* dem Pseudoop	  */
	case GLOBAL :  doent();	       return;	  /* entsprechende	  */
	case DW	    :  dodat(2);       return;	  /* Funktion ausfuehren  */
	case DB	    :  dodat(1);       return;
	case DS	    :  doloc(YES);     return;
	case ORG    :  doloc(NO);      return;
	case SET    :  doval(SETBIT);  return;
	case EQU    :  doval(0);       return;
	case END    :  doend();	       return;
	case CSEG   :  docseg();       return;
	case DSEG   :  dodseg();       return;
	case IGNORE :		       return;
      }
    oprerr();					  /* voellig unbekannt	  */
  }





/**************************************************************************\
*									   *
* action      :	 Daten definieren (DB &	DW). Uebergeben	wird in	lp ein	   *
*		 Zeiger	auf die	Parameterliste des DB/DW in der	Eingabe-   *
*		 zeile.	Die Parameter werden alle ausgewertet, in Pass 2   *
*		 wird Code erzeugt/gelistet. Bei Fehlern (Ueberlauf,	   *
*		 Fehler	im Ausdruck, Relokationsfehler)	wird das	   *
*		 entsprechende Fehlerflag gesetzt. Zusammengesetzte	   *
*		 Ausdruecke werden ausgewertet.				   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 sz	-  1 ->	DB, 2 -> DW				   *
*									   *
* r/o globals :	 pass	-  wenn	2 Code erzeugen				   *
*									   *
* r/w globals :	 lp	-  Zeiger auf Parameterstring von DB/DW		   *
*		 field	-  zur Coderzeugung/Ausdruckauswertung		   *
*		 type	-    "						   *
*		 item	-  zur Codeerzeugung				   *
*		 prior	-    "						   *
*		 priort	-    "						   *
*		 loc	-  aktueller Code/Datenzaehlerstand, wird bei	   *
*			   jedem DB/DW-Parameter um sz bzw. Stringlaenge   *
*			   * sz	erhoeht					   *
*		 ep	-  Zeiger auf auszuwertenden Ausdruck fuer expr()  *
*									   *
* locals      :	 dlm	-  Stringdelimiter, " oder '			   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID dodat(sz)

int sz;

  {
    int	dlm;

    while (!atend(*lp))				  /* ganze Parameterliste */
      {
	if (isspace(*lp) || *lp	== ',')		  /* zum naechsten Objekt */
	  ++lp;
	else
	  if (	  (*lp == '\"' || *lp == '\'')	  /* String ?		  */
	       && lp[0]	!= lp[2]		  /* "c", 'c' kein String */
	     )
	    {					  /* String :		  */
	      dlm = *lp;			  /* Delimiterzeichen	  */
	      while (!atend(*++lp))		  /* String durchgehen	  */
		{				  /* bis Delimiter	  */
		  if (*lp == dlm && *++lp != dlm) /* erscheint		  */
		    break;
		  if (pass == 2)
		    {				  /* Pass 2 :		  */
		      field = *lp;		  /*   Code generieren	  */
		      genabs(sz);
		    }
		  else				  /* Pass 1 :		  */
		    loc	+= sz;			  /*   nur weiterzaehlen  */
		}
	    }
	  else
	    {
	      ep = lp;				  /* zusammengesetzten	  */
	      expr(&field,&type);		  /* Ausdruck auswerten	  */
	      lp = ep;				  /* (naechster	Ausdruck) */
	      if (pass == 2)
		{				  /* Pass 2 :		  */
		  if ((type & RELBITS) == ABS)	  /* absoluter Ausdruck	: */
		    {
		      if (    sz == 1		  /*   Wert zu gross fuer */
			   && field > 255	  /*   1 Byte		  */
			 ) rngerr();
		      genabs(sz);		  /*   generieren	  */
		    }
		  else
		    {				  /* relokatibel :	  */
		      if (sz ==	1)
			{			  /* 1 Byte relokatibel	: */
			  relerr();		  /*   gibt es nicht !	  */
			  genabs(1);		  /*   Schein wahren	  */
			}
		      else			  /* 2 Byte relokatibel	: */
			if (type & XRBIT)	  /*   externe Referenz	  */
			  {
			    if (field)
			      {			  /* Offset :		  */
				item = XPOFF;	  /*   Typ Offset	  */
				type = ABS;	  /*   immer absolut	  */
				listcode(2,"+ "); /*   anzeigen		  */
				putrel();	  /*   und generieren	  */
			      }
			    field = prior;	  /* Verweis auf voriges  */
			    genrel(priort);	  /* Glied in Kette	  */
			  }
			else			  /* keine ext.	Ref. :	  */
			  genrel(type);		  /*   2 Byte generieren  */
		    }
		}
	      else				  /* Pass 1 :		  */
		loc += sz;			  /*   nur weiterzaehlen  */
	    }
      }	/* end while (!atend(*lp)) */		  /* ... bis Zeile leer	  */
  }





/**************************************************************************\
*									   *
* action      :	 END-Pseudo-Op verarbeiten. Wertet, falls vorhanden, den   *
*		 Ausdruck hinter dem END aus. Wenn er vom Typ PREL ist,	   *
*		 werden	endt/endv entsprechend gesetzt,	sonst Fehler.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 type	-  Typ des Ausdrucks hinter END	 (werden durch	   *
*		 field	-  Wert	"   "	      "	     "	 onexpr() gesetzt) *
*									   *
* r/w globals :	 eom	-  wird	auf YES	gesetzt, Modulende erreicht	   *
*		 endt	-  Typ des Ausdrucks hinter END, falls vorhanden   *
*		 endv	-  Wert	des Ausdrucks hinter END, falls	vorhanden  *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doend()
  {
    eom	= YES;				     /*	Ende des Moduls	erreicht  */
    onexpr();				     /*	Ausdruck auswerten	  */
    if ((type &	RELBITS) == PREL)	     /*	nur annehmen, wenn PREL	  */
      {
	endt = PREL;
	endv = field;
      }
    else
      if (field) relerr();
  }





/**************************************************************************\
*									   *
* action      :	 Externe Referenz definieren (EXT, EXTRN, EXTERNAL).	   *
*		 Uebergeben wird in lp ein Zeiger auf die Externalnamen	   *
*		 des EXT/EXTRN/EXTERNAL	in der Eingabezeile.		   *
*		 Es werden neue	Symbole	in der Symboltabelle angelegt,	   *
*		 wenn noch nicht vorhanden.				   *
*		 Test auf falsch oder doppelt definierte Symbole machen,   *
*		 ggf. entsprechende Fehlerflags	setzen.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 badsym	 -  zum	Test, ob Symbol	in Ordnung		   *
*		 pass	 -  wenn 1 dann	External-Typ ABS		   *
*									   *
* r/w globals :	 lp	 -  Zeiger auf Externalnamen in	Eingabezeile	   *
*		 stptr[] -  Eintrag in Symboltabelle fuer zu definierenden *
*			    External					   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doext()
  {
    while (!atend(*lp))			     /*	alle Externals/Zeile	  */
      {
	while (isspace(*lp) || *lp == ',')   /*	naechstes Objekt	  */
	  ++lp;
	lp = getsy(lp,NO);		     /*	naechstes Symbol holen	  */
	if (badsym)			     /*	Symbol-Fehler		  */
	  {
	    symerr();
	    continue;
	  }
	else
	  if (stfind())			     /*	bereits	in Tabelle ?	  */
	    {
	      if (stptr[STFLAG]	& (LABBIT|EQUBIT|SETBIT))
		{
		  rederr();		     /*	ja, Fehler		  */
		  continue;
		}
	    }
	  else
	    addsym();			     /*	noch nicht definiert, neu */
	if (pass == 1)
	  stptr[STFLAG]	|= XRBIT|ABS;	     /*	erste ext. Referenz ABS	0 */
      }
  }





/**************************************************************************\
*									   *
* action      :	 Entrypoint definieren (PUBLIC,	GLOBAL).		   *
*		 Uebergeben wird in lp ein Zeiger auf die Entrypointnamen  *
*		 des PUBLIC/GLOBAL in der Eingabezeile.			   *
*		 Es werden neue	Symbole	in der Symboltabelle angelegt,	   *
*		 wenn noch nicht vorhanden.				   *
*		 Test auf falsch oder doppelt definierte Symbole machen,   *
*		 ggf. entsprechende Fehlerflags	setzen.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 badsym	 -  zum	Test, ob Symbol	in Ordnung		   *
*		 pass	 -  wenn 1 dann	EPBIT setzen fuer Symbol	   *
*									   *
* r/w globals :	 lp	 -  Zeiger auf Entrypointnamen in Eingabezeile	   *
*		 stptr[] -  Eintrag in Symboltabelle fuer zu definierenden *
*			    Entrypoint					   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doent()
  {
    while (!atend(*lp))			     /*	Zeile ganz abarbeiten	  */
      {
	while (isspace(*lp) || *lp == ',')   /*	zum naechsten Symbol	  */
	  ++lp;
	lp = getsy(lp,NO);		     /*	Symbol holen		  */
	if (badsym)			     /*	Fehler in Symbol :	  */
	  {
	    symerr();			     /*	  Symbolfehler mitteilen  */
	    continue;			     /*	  gleich zum naechsten	  */
	  }
	else
	  if (stfind())			     /*	schon in Symboltabelle,	  */
	    {
	      if (stptr[STFLAG]	& (EQUBIT|SETBIT|XRBIT))
		{
		  rederr();		     /*	wenn als SET/EQU/EXT,	  */
		  continue;		     /*	dann Redefine-Fehler und  */
		}			     /*	zum naechsten		  */
	    }
	  else				     /*	wenn nicht in Symbol-	  */
	    addsym();			     /*	tabelle, hineintun	  */
	if (pass == 1)			     /*	im Pass	1 als Entrypoint  */
	  stptr[STFLAG]	|= EPBIT;	     /*	deklarieren		  */
      }
  }





/**************************************************************************\
*									   *
* action      :	 Label in Sourcezeile line suchen und wenn gefunden,	   *
*		 verarbeiten.						   *
*									   *
*		 Entsprechend den Passes auf Redefinement testen und ggf.  *
*		 melden, nach Durchlauf	Pass-Label-Define-Flags	im	   *
*		 Symboltabelleneintrag setzen.				   *
*									   *
*		 Neue Label in Symboltabelle einfuegen,	SSSSSS:: Entry-	   *
*		 point-Label beachten.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 loc	-  aktuelle Position im	Code/Datensegment	   *
*		 line	-  aktuelle Sourcezeile				   *
*									   *
* r/w globals :	 lp	-  Zeiger auf zu untersuchende Zeile, enthaelt bei *
*			   Rueckgabe Zeiger auf	das naechste Token hinter  *
*			   dem Label					   *
*		 stptr	-  Zeiger auf Label-Eintrag in st		   *
*		 (st)	-  Label-Eintrag wird gesetzt/geaendert, s.u.	   *
*									   *
*		 gotlabel, badsym, gotep, stptr				   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID dolabel()
  {
    lp = skip(1,line);	      /* Label muss 1. Feld der	Sourcezeile sein  */
    lp = getsy(lp,NO);	      /* Symbol	holen, nicht als Referenz	  */
    if (gotlabel)	      /* Label gefunden				  */
      {
	if (badsym)			/* Label fehlerhaft :		  */
	  {
	    laberr();			/*   Fehler mitteilen		  */
	    return;			/*   das war's			  */
	  }
	if (stfind())			/* Label schon in Symboltabelle	: */
	  {
	    if (pass ==	1)		/*   Pass 1, schon definiert ?	  */
	      {
		if (stptr[STFLAG] & (LABBIT|EQUBIT|SETBIT|XRBIT))
		  {
		    rederr();		/*   ja, Redefine-Fehler melden	  */
		    return;		/*   das war's			  */
		  }
	      }				/*   Pass 2, schon definiert ?	  */
	    else
	      if (stptr[STFLAG]	& (LABBIT2|EQUBIT|SETBIT|XRBIT))
		{
		  rederr();		/*   ja, Redefine-Fehler melden	  */
		  return;		/*   das war's			  */
		}
	      else
		stptr[STFLAG] |= LABBIT2;  /* Pass 2, Label definiert !	  */
	  }
	else				/* Label ist noch nicht	in	  */
	  addsym();			/* Symboltabelle, hineintun	  */
	if (pass == 1)
	  {					  /* Pass 1 :		  */
	    putuint(stptr + STVALUE,loc);	  /*   Label-Wert setzen  */
	    if (gotep)				  /*   Flags setzen	  */
	      stptr[STFLAG] |= LABBIT |	seg | EPBIT;
	    else
	      stptr[STFLAG] |= LABBIT |	seg;
	  }
      }	/* end if (gotlabel) */
  }





/**************************************************************************\
*									   *
* action      :	 Positionszaehler setzen (ORG, DS).			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 bump	-  YES:	(DS)  =	 + Locationcounter		   *
*			   NO:	(ORG) =	 = Locationcounter		   *
*									   *
* r/o globals :	 pass	-  aktueller Assemblerpass, 1 oder 2		   *
*		 seg	-  aktuelles Ladesegment			   *
*									   *
* r/w globals :	 field	-  Wert	des Ausdrucks hinter dem Pseudoop	   *
*		 loc	-  aktueller Locationcounter			   *
*		 item	-  REL-Item fuer REL-Codeerzeugung		   *
*		 type	-  Typ	      "	  "    "			   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doloc(bump)

int bump;

  {
    if (onexpr())			/* wenn	Ausdruck hinter	dem	  */
      {					/* Pseudoop ok,			  */
	if (bump)
	  field	= loc += field;		/* Locationcounter + Wert	  */
	else
	  if (loc <= field)
	    loc	= field;		/* Locationcounter = Wert	  */
	  else
	    bakerr();			/* Fehler: "Backward Movement"	  */
	if (pass == 2)			/* im Pass 2 REL-Code erzeugen	  */
	  {
	    item = SETLC;		/* "Setze Locationcounter"	  */
	    type = seg;			/* fuer	entsprechendes Segment	  */
	    putrel();			/* REL-Code schreiben		  */
	  }
      }
  }





/**************************************************************************\
*									   *
* action      :	 In Codesegment	umschalten (CSEG).			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 cloc	-  Sicherung Programmlocationcounter		   *
*		 pass	-  aktueller Assemblerpass, 1 oder 2		   *
*									   *
* r/w globals :	 seg	-  wenn	DREL, dann CREL	neu			   *
*		 dloc	-  Sicherung Datenlocationcounter		   *
*		 loc	-  Locationcounter				   *
*		 field	-  fuer	REL-Codeerzeugung, Wert			   *
*		 item	-   "	 "    "		 , Item			   *
*		 type	-   "	 "    "		 , Typ			   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID docseg()
  {
    if (seg == DREL)			/* Umschalten nur, wenn	im	  */
      {					/* Datensegment			  */
	dloc = loc;			/* Datenzaehler	sichern		  */
	loc = cloc;			/* aktueller Zaehler = Programm	  */
	seg = PREL;			/* jetzt im Programmsegment	  */
	if (pass == 2)			/* Codeerzeugung nur im	Pass 2	  */
	  {
	    field = loc;		/* REL-Wert			  */
	    item = SETLC;		/* REL-Item			  */
	    type = PREL;		/* REL-Typ			  */
	    putrel();			/* REL-Code schreiben		  */
	  }
      }
  }





/**************************************************************************\
*									   *
* action      :	 In Datensegment umschalten (DSEG).			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 dloc	-  Sicherung Datenlocationcounter		   *
*		 pass	-  aktueller Assemblerpass, 1 oder 2		   *
*									   *
* r/w globals :	 seg	-  wenn	PREL, dann DREL	neu			   *
*		 cloc	-  Sicherung Programmlocationcounter		   *
*		 loc	-  Locationcounter				   *
*		 field	-  fuer	REL-Codeerzeugung, Wert			   *
*		 item	-   "	 "    "		 , Item			   *
*		 type	-   "	 "    "		 , Typ			   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID dodseg()
  {
    if (seg == PREL)			/* Umschalten nur, wenn	im	  */
      {					/* Programmsegment		  */
	cloc = loc;			/* Programmzaehler sichern	  */
	loc = dloc;			/* aktueller Zaehler = Daten	  */
	seg = DREL;			/* jetzt im Datensegment	  */
	if (pass == 2)			/* Codeerzeugung nur im	Pass 2	  */
	  {
	    field = loc;		/* REL-Wert			  */
	    item = SETLC;		/* REL-Item			  */
	    type = DREL;		/* REL-Typ			  */
	    putrel();			/* REL-Code schreiben		  */
	  }
      }
  }





/**************************************************************************\
*									   *
* action      :	 Eingabezeile auf Maschineninstruktion testen, wenn eine   *
*		 gueltige gefunden, dann im Pass 2 Code	erzeugen und	   *
*		 Locationcounter entsprechend erhoehen,	im Pass	1 nur	   *
*		 Locationcounter erhoehen.				   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 gotlabel -  YES: lp zeigt auf Instruktion in Eingabezeile *
*		 lp	  -  s.o.					   *
*		 line	  -  aktuelle Eingabezeile			   *
*		 pass	  -  Assemblerpass, 1 oder 2			   *
*		 loc	  -  Locationcounter, wird der Maschineninstr.	   *
*			     entsprechend erhoeht oder bleibt, wenn keine  *
*			     keine gefunden				   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 fmt	-  wenn	Maschineninstruktion gefunden, Zeiger auf  *
*			   Formatbyte, sonst NULL			   *
*		 cp	-  lokaler Zeiger in Eingabezeile auf		   *
*			   Maschineninstruktion				   *
*									   *
* returns     :	 YES:  Maschineninstruktion in Eingabezeile gefunden	   *
*		 NO :  keine Maschineninstruktion in Eingabezeile	   *
*		       gefunden, muss Pseuoop sein			   *
*									   *
\**************************************************************************/

BOOLEAN	domach()
  {
    char *fmt;
    char *cp;

    if (gotlabel)			/* wenn	Label, dann zeigt lp auf  */
      cp = lp;				/* Befehlsmnemonic,		  */
    else				/* sonst steht er im ersten Feld  */
      cp = skip(1,line);		/* der Eingabezeile		  */
    if ((fmt = find(cp)) != 0x00)	/* ist es Maschineninstruktion ?  */
      {					/* ja, Zeiger auf Formatbyte der  */
	fmt += INTSZ;			/* Instruktion berechnen	  */
	if (pass == 2)			/* im Pass 2 Code erzeugen	  */
	  domac2(fmt);
	else				     /*	im Pass	1 nur Location-	  */
	  loc += (getuint(fmt) & 0x07) + 1;  /*	counter	setzen		  */
	return (YES);			     /*	Maschineninstruktion !	  */
      }
    return (NO);			     /*	keine Maschineninstr. !	  */
  }





/**************************************************************************\
*									   *
* action      :	 Maschineninstruktion auswerten	und entsprechenden	   *
*		 Objektcode generieren.					   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 ptr	-  Zeiger auf Formatbyte in MIT			   *
*									   *
* r/o globals :	 expbuf	-  (gesetzter) Ausdrucksbuffer fuer Instruktion	   *
*		 seg	-  aktuelles Segment, PREL oder	DREL		   *
*									   *
* r/w globals :	 ep	-  expbuf, fuer	expr()				   *
*		 opadj	-  Opcodeteil operandenabhaengig 0,1,2,	...	   *
*		 item	-  benutzt fuer	Offset-Generierung, dann XPOFF	   *
*		 type	-  benutzt fuer	Offset-Generierung, dann ABS	   *
*		 field	-  Wert	fuer Codegenerierung			   *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID domac2(ptr)

char *ptr;

  {
    int	      format;			/* Formatbytebuffer		  */
    int	      len;			/* Instruktionslaengenzaehler	  */
    int	      ilen;			/* Instruktionslaenge		  */
    BOOLEAN   pcr;			/* YES:	Adressierung PC-relativ	  */
    int	      t;			/* lokaler Typspeicher		  */
    int	      t1;			/*   "	     "			  */
    int	      v;			/* lokaler Wertspeicher		  */
    int	      opcode = 0;		/* ein Byte Opcode		  */
    BOOLEAN   holding;			/* YES:	noch ein Opcodebyte ueber */



    format = getuint(ptr++);		/* Formatbyte holen		  */
    len	= ilen = (format & 7) +	1;	/* Befehlslaenge aus Formatbyte	  */
    format >>= 3;			/* 1. Bit nun Code/Expr-Flag	  */
    holding = NO;			/* noch	kein Opcodebyte		  */
    ep = expbuf;			/* ep fuer expr() setzen	  */
    while (len-- > 0)			/* fuer	jedes Codebyte ...	  */
      {
	if (format & 1)			/* zusammengesetzter Ausdruck	  */
	  {
	    if (holding)		/* ggf.	vorheriges Opcodebyte	  */
	      {				/* wegschreiben			  */
		holding	= NO;
		field =	opcode + opadj;
		opadj =	0;
		genabs(1);
	      }
	    expr((unsigned *)&v, (unsigned *)&t);
					/* naechsten Ausdruck auswerten	  */
	    format >>= 1;		/* PC-relatives	Bit		  */
	    if (format & 1)		/* Wert	PC-relativ ?		  */
	      {				/* ja, Segment checken,	wenn	  */
		if (t &	RELBITS)	/* Ausdruck relativ		  */
		  {
		    if ((t & RELBITS) != seg)	  /* relativ nur im	  */
		      relerr();			  /* selben Segment	  */
		    v -= ilen +	iloc;		  /* Offset vom	Befehl	  */
		    t =	( t & ~RELBITS)	+ ABS;	  /* Typ ABS		  */
		  }
		else
		  v -= ilen;
		pcr = YES;		/* merken, dass	es PC-relativ war */
	      }
	    else
	      pcr = NO;
	    format >>= 1;		/* Groessenbit			  */
	    if (format & 1)		/* 2 Byte langer Ausdruck ?	  */
	      {
		t1 = t;			/* ja, Typ merken		  */
		if (t &	XRBIT)		/* externe Referenz		  */
		  {
		    if (v)		/* Offset von externer Referenz	  */
		      {
			item = XPOFF;
			type = ABS;
			field =	v;
			listcode(2,"+ ");	  /* Offset ausgeben	  */
			putrel();		  /* 2-Byte-Offset	  */
		      }
		    field = prior;
		    t1 = priort;
		  }
		else
		  field	= v;			  /* Wert des Ausdrucks	  */
		if ((t & RELBITS) == ABS)
		  genabs(2);			  /* 2 absolute	Bytes	  */
		else
		  genrel(t1);			  /* 2 relokatible Bytes  */
		--len;
	      }
	    else				  /* 1 Byte Ausdruck	  */
	      {
		if (t &	RELBITS)
		  relerr();			  /* 1 Byte nie	relokat.  */
		if (pcr	&& (v >	127 || v < -128))
		  rngerr();			  /* Bereichsfehler	  */
		field =	v;			  /* Wert des Ausdrucks	  */
		genabs(1);			  /* 1 absolutes Byte	  */
	      }
	  } /* end if (format &	1) */
	else					  /* Codebyte		  */
	  {
	    if (holding)
	      {
		field =	opcode;		/* noch	nicht das letzte Byte	  */
		genabs(1);		/* voriges Codebyte schreiben	  */
	      }
	    opcode = *++ptr & 0xFF;	/* vielleicht kommt noch was	  */
	    holding = YES;
	  }
	format >>= 1;
      }	/* end while (len-- > 0) */
    if (holding)			/* wenn	noch Codebyte uebrig,	  */
      {					/* letztes Codebyte schreiben	  */
	field =	opcode + opadj;
	genabs(1);
      }
  }





/**************************************************************************\
*									   *
* action      :	 Symbolwert definieren (SET, EQU). Bei falschen	oder	   *
*		 doppelten Definitionen	entsprechende Fehlerflags setzen.  *
*		 Im Pass 2 Wert/Typ listen, im Pass 1 definieren.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 set	-  0 wenn EQU, SETBIT wenn SET			   *
*									   *
* r/o globals :	 pass	  -  Assembler Pass 1 oder 2			   *
*		 stsym[0] -  muss ungleich 0 sein, d.h.	es muss	in der	   *
*			     EQU/SET-Zeile ein Symbol vor EQU/SET	   *
*			     definiert sein				   *
*		 badsym	  -  muss = 0 sein, Symbol (s.o.) muss fehlerfrei  *
*			     sein					   *
*		 gotlabel -  muss = 0 sein, Symbol (s.o.) darf kein Label  *
*			     sein					   *
*									   *
* r/w globals :	 type	  -  EQU/SET-Typ ohne LABBIT			   *
*		 (field)  -  EQU/SET-Wert, wird	durch onexpr() gesetzt	   *
*									   *
* locals      :	 ptr	  -  Zeiger auf	Eintrag	in Symboltabelle	   *
*		 found	  -  0 wenn Symbol nicht gefunden		   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

VOID doval(set)

int set;

  {
    char   *ptr;
    int	    found;

    if (!stsym[0] || badsym || gotlabel)	  /* vor dem EQU/SET muss */
      {						  /* ein gueltiges Nicht- */
	symerr();				  /* Labelsymbol stehen,  */
	return;					  /* sonst Fehler	  */
      }
    if ((found = stfind()) == 0) addsym();	  /* nicht definiert, neu */
    ptr	= stptr;				  /* Symboltab.-Eintrag	  */
    onexpr();					  /* Ausdruck auswerten	  */
    type &= ~LABBIT;				  /* Pass-2-Fehler ...	  */
    if (pass ==	1 || set)			  /* Wert + Typ	zuweisen  */
      {
	if (found == 0 || ptr[STFLAG] &	set || ptr[STFLAG] & EPBIT)
	  {
	    putuint(ptr	+ STVALUE,field);
	    ptr[STFLAG]	= (ptr[STFLAG] & EPBIT)	| set |	type;
	  }
	else					       /* Doppeldefinit.  */
	  rederr();				       /* abfangen	  */
      }
    else
      if (ptr[STFLAG] &	(LABBIT|EQUBIT|SETBIT|XRBIT))  /* Doppeldefinit.  */
	rederr();				       /* abfangen	  */
      else
	ptr[STFLAG] |= EQUBIT;			       /* ist EQU	  */
    if (pass ==	2)				       /* Pass 2 :	  */
      switch (ptr[STFLAG] & RELBITS)		       /* Wert/Typ listen */
	{
	  case PREL :  listcode(2,"' =");    break;
	  case DREL :  listcode(2,"\" =");   break;
	  default   :  listcode(2,"  =");    break;
	}
  }



/* Ende	von MAC3.C */
