/**************************************************************************\
*									   *
*   MIT.C   -	unido X-Mac, Funktionen	fuer die Tabelle der		   *
*		Maschineninstruktionen (MIT)				   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/



/*
 * Aufbau der MIT :
 *
 * Die eigentliche MIT steht sequentiell in mitbuf, zunaechst alle
 * jeweils paarweise verschiedenen Operandenfelder im Klartext,	jedes
 * Feld	abgeschlossen mit einer	0. Dann	folgen alle Mnemonics mit
 * zugehoerigen	Objektcodes in allen Varianten.
 *
 *
 *    +-- mitbuf
 *    |
 *    v
 *   +---------+---+---------+---+--	     ----------+---+
 *   | opfield | 0 | opfield | 0 |     ...     opfield | 0 |
 *   +---------+---+---------+---+--	     ----------+---+
 *		    A
 *		    |  (Index)
 *		    +--------------+				    |
 *				   |				    |
 *   +---+----------+---+------+-------+--------+--------+------+-------+--
 *   | 0 | mnemonic | 0	| vlen | opidx | format	| object | vlen	| opidx	|
 *   +---+----------+---+------+-------+--------+--------+------+-------+--
 *
 * --------+--------+------+-------+--------+--------+---+----------+--
 *  format | object | vlen | opidx | format | object | 0 | mnemonic |  ...
 * --------+--------+------+-------+--------+--------+---+----------+--
 *							  A
 *   ---------+---+					  |
 *     object |	0 |					  |
 *   ---------+---+			 +-----+	  |
 *			      mitidx --> |     |	  |
 *					 +-----+  (Index) |
 *					 |     |----------+
 *					 +-----+
 *					 |     |
 *					    .
 *					    .
 *					 |     |
 *					 +-----+
 *					 |     |
 *					 +-----+
 *
 *
 *
 *  opfield   -	Operandenfeld, das dritte Feld in im MIT-Text,
 *		je eine	Zeile aus dem MIT-Text ergibt ein Operandenfeld,
 *		jedoch werden gleiche Operandenfelder unterdrueckt
 *
 * mnemonic   -	mnemonischer Befehl, das 2. Feld einer Zeile im	MIT-Text,
 *		gleiche	Mnemonic-Felder	werden unterdrueckt, je	eine
 *		Zeile aus dem MIT-Text ergibt eine Variante
 *
 *     vlen   -	Laenge in Bytes	der Variante (opidx + format + object Bytes)
 *
 *    opidx   -	2 Byte Index des zur Variante gehoerenden Operandenfeldes,
 *		der Zeiger auf das Operandenfeld ergibt	sich durch
 *		mitbuf + opidx,	EOF (-1) bedeutet "kein	Operand"
 *
 *   object   -	die Objektcodebytes zur	Variante
 *
 *   format   -	2 Byte Format, dies ist	ein 16-Bit "Bitstream",	rechtsbuendig
 *		ausgerichtet. Die ersten drei Bits rechts geben	die
 *		Instruktionslaenge an (000 = 1,	001 = 2, 010 = 3, ...).
 *		Die naechsten Bits von rechts nach links geben die weiteren
 *		Codebytes an, ist das Bit eine 0, dann folgt im	Code ein
 *		ein Byte aus object, ist das Bit eine 1, dann folgt im Code
 *		ein Ausdruck, den die naechsten	beiden Bits angeben :
 *
 *		  00 1	 =  8-bit Ausdruck
 *		  01 1	 =  8-bit PC relativer Ausdruck
 *		  10 1	 =  16-bit Ausdruck
 *		  11 1	 =  16-bit PC relativer	Ausdruck
 *
 *
 * In mitidx sind die Indizes (Zeiger durch mitbuf + mitidx[n])	auf die
 * Mnemonics zusammengefasst. Ein EOF in mitidx	zeigt einen leeren Eintrag
 * an (hier an sich nicht notwendig, da	alle Indizes hintereinander
 * eingetragen sind).
 *
 * In mitndx stehen alle Hash-Indize auf mitidx. Ein EOF in mitndx zeigt
 * einen leeren	Eintrag	an. D.h. es wird mit dem Instruktionsmnemonic ein
 * Hashwert errechnet, dieser ist Index	in mitndx. An dieser Stelle in
 * mitndx wird der Index auf mitidx eingetragen, unter dem man in mitidx
 * den Index in	den MIT-Buffer auf die Stelle findet, wo das
 * Instruktionsmnemonic	steht.
 *
 * Zunaechst werden alle Hashwerte in mitndx eingetragen, bei mehrfach
 * auftretenden	Hashwerten wird	jeweils	zunaechst nur der erste	mitidx-Index
 * eingetragen.	Nachdem	so mit allen Mnemonics verfahren wurde,	werden
 * die noch nicht eingetragenen	Mehrfachwerte sequentiell hintereinander
 * in noch nicht besetzte mitndx-Stellen eingetragen. Der Zugriff auf diese
 * Werte geschieht ueber mitnxt.
 *
 * In mitnxt wird genauso gehasht wie in mitndx, eingetragen wird nun aber
 * der Index auf mitndx. Ist hier eine Stelle schon besetzt, wird die
 * naechste freie mitnxt-Stelle	ueber den an der besetzten Stelle stehenden
 * Index gesucht, bis eine Stelle frei wird.
 *
 * Auf eine Instruktion	gematcht wird nun, indem zunaechst der Hashwert
 * des Mnemonics berechnet wird. Dann wird ueber diesen	Hashwert als Index
 * in mitndx der Index in mitidx geholt. Dort wird der Index in	den
 * MIT-Buffer geholt und der dort stehende Mnemonic mit	dem der	Instruktion
 * verglichen. Sind sie	gleich,	ist der	Mnemonic gefunden.
 * Sind	sie ungleich, dann wird	mit dem	Hashwert als Index aus mitnxt ein
 * Index in mitndx geholt und wie oben weiterverfahren.	Ist der
 * Mnemonic immer noch nicht gleich, wird der in mitnxt	stehende Wert als
 * neuer Index in mitnxt genommen und solange weiterverfahren, bis entweder
 * der Mnemonic	gefunden, oder in mitnxt ein EOF (freier Eintrag) steht.
 *
 */





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* Standardausgabe benutzt	  */
#include <string.h>			/* Stringfunktionen benutzt	  */
#include <ctype.h>
#include "all.h"			/* allgemeine Festlegungen	  */
#include "mac.h"			/* MIT-Groessen			  */
#include "mace.h"


BOOLEAN	 match __ARGS((char *, char *));


/*	      externe nicht-int-Funktionen (nicht aus stdio.h/string.h)	  */
/**************************************************************************/

extern char	  *skip();
extern unsigned	   getuint();


/*							      Externals	  */
/**************************************************************************/

extern int    mitndx[];			/* MIT-Indizes (hash ->	which)	  */
extern int    mitnxt[];			/* Kette der MIT-Synonyme	  */
extern int    mitidx[];			/* Indizes fuer	Mnemonic-Syntax	  */
extern char   mitbuf[];			/* Puffer Instruktions-Syntax	  */
extern int    looks;			/* Anzahl der Suchversuche	  */





/*								Globals	  */
/**************************************************************************/

int  opadj;		      /* Ausrichtung fuer Opcode		  */
char expbuf[MAXLINE];	      /* Puffer	fuer die Operanden-Ausdruecke	  */





/**************************************************************************\
*									   *
* action      :	 Hash-Wert (0 .. cnt - 1) aus dem String, auf den ptr	   *
*		 zeigt,	 berechnen.					   *
*									   *
*		 Die Berechnung	erfolgt	zeichenweise durch Linksschieben   *
*		 des Hashwertes	und Addieren des Zeichens. Lowercase-	   *
*		 Zeichen werden	nach Uppercase gewandelt.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 ptr - Zeiger auf den String, von dem ein Hashwert	   *
*		       berechnet werden	soll				   *
*		 cnt - maximal moeglicher Hashwert - 1			   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 errechneter Hashwert (% cnt)				   *
*									   *
\**************************************************************************/

int hash(ptr,cnt)

register char *ptr;
int cnt;

  {
    register int hashval;

    hashval = 0;
    while (*ptr	> ' ' && atend(*ptr) ==	NO)
      hashval =	(hashval << 1) + toupper(*ptr++);
    return (hashval % cnt);
  }





/**************************************************************************\
*									   *
* action      :	 Suchen	der Instruktion, auf die instr zeigt, in der MIT.  *
*									   *
 **************************************************************************
*									   *
* parameter   :	 inst -	Zeiger auf den String der in der MIT zu	findenden  *
*			Instruktion					   *
*									   *
* r/o globals :	 mitbuf, mitidx, mitndx, mitnxt,(MICOUNT)		   *
*									   *
* r/w globals :	 looks - zeigt die Anzahl der Suchversuche an		   *
*			 (fuer jeden neuen Mnemonic  u n d  jede Variante) *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 NULL, wenn Instruktion	nicht gefunden,			   *
*		 sonst Zeiger auf den Operandenfeldindex der Variante in   *
*		 der MIT, der die Instruktion entspricht		   *
*									   *
\**************************************************************************/

char *find(inst)

char *inst;

  {
    char   *mit;	      /* Zeiger	in mitbuf			  */
    int	    h;		      /* Hashwert				  */
    int	    ndx;	      /* Wert aus mitndx			  */
    int	    opidx;	      /* Buffer	Operandenindex aus Variante	  */


    looks = 0;					  /* noch kein Versuch	  */
    ndx	= mitndx[h = hash(inst,MICOUNT)];	  /* Hashwert Mnemonic	  */
    while (ndx != EOF)				  /* Index suchen	  */
      {
	++looks;				  /* fuer jeden	Mnemonic  */
	if (fldcmp(inst,mit = mitbuf + mitidx[ndx]) == 0)
	  {				/* Mnemonic gefunden :		  */
	    inst = skip(2,inst);	/*   Operandenfeld		  */
	    mit	+= strlen(mit) + 1;	/*   erste Variante		  */
	    while (*mit++)		/*   alle Varianten durchgehen	  */
	      {
		++looks;		/*   fuer jede Variante		  */
		opidx =	getuint(mit);
		if (match(inst,	opidx == 0xffff	? NULL : mitbuf	+ opidx))
		  return (mit);		/*   Variante gefunden !	  */
		mit += *(mit - 1);	/*   nicht gefunden, naechste	  */
	      }
	    return (NULL);		/*   keine Variante passt	  */
	  }
	if ((h = mitnxt[h]) == EOF)	/* noch	Mnemonics selber Hash ?	  */
	  return (NULL);		/* nein	- Mnemonic gibt	es nicht  */
	ndx = mitndx[h];		/* ja	- naechsten probieren	  */
      }	/* end while (ndx != EOF) */
    return (NULL);			/* keinen passender Mnemonic da	  */
  }





/**************************************************************************\
*									   *
* action      :	 Suchend des Operanden in MIT-Buffer-Operandenfeld.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 inst  -  Zeiger auf Operandenfeld im Instruktionsstring   *
*		 mit   -  Zeiger auf Operandenfeld im MIT-Buffer	   *
*			  oder NULL wenn kein Operand zu testen		   *
*									   *
* r/o globals :	 (ANOTHER)						   *
*									   *
*									   *
* r/w globals :	 opadj	 -  gibt den Operanden im Operandenfeld	der MIT	   *
*			    an,	falls Operand gefunden wurde (0	= erster   *
*			    Operand im Feld, 1 = zweiter ...)		   *
*		 expbuf	 -  wenn der Operand gefunden wurde und	ein oder   *
*			    mehrere Ausdruecke sich in diesem befinden,	   *
*			    stehen hier	die kopierten Ausdruecke durch ',' *
*			    separiert und mit ",\0" beendet		   *
*									   *
* locals      :	 s.u.							   *
*									   *
* returns     :	 YES - Operand im Operandenfeld	(MIT) der Variante	   *
*		       gefunden						   *
*		 NO  - Operand nicht gefunden				   *
*									   *
\**************************************************************************/

BOOLEAN	match(inst,mit)

char *inst;
char *mit;

  {
    char   *backup;	      /* Merker	fuer Zeiger auf	Instruktionstring */
    char   *exp;	      /* Zeiger	in expbuf			  */
    int	    nest;	      /* Zaehler Klammertiefe			  */

    opadj = 0;				/* Merker welcher Op. neu	  */
    backup = inst;			/* Zeiger auf Instruktion merken  */
    if (mit == NULL)			/* kein	Op.feld	in Variante,	  */
      {
	if (atend(*inst))		/* Instruktion auch am Ende ?	  */
	  return (YES);			/* ja, gefunden	!		  */
	return (NO);			/* nein, Variante stimmt nicht	  */
      }
    exp	= expbuf;			/* Ausdruckspuffer initialisieren */
    LOOP				/* Inst. zeichenweise abarbeiten  */
      {
	while (isspace(*inst)) ++inst;	/* Leerzeichen in Inst.	skippen	  */
	while (isspace(*mit)) ++mit;	/* Leerzeichen in Op.feld skippen */
	if (atend(*inst))		/* Instruktionsende erreicht :	  */
	  {
	    if (atend(*mit) || *mit == ANOTHER)	/*   auch Ende in MIT ?	  */
	      return (YES);			/*   ja, gefunden !	  */
	    goto next;				/*   nein, naechster Op.  */
	  }					/*	   im Op.feld	  */
	if (atend(*mit))			/* Operandenfeld zuende	? */
	  return (NO);				 /* ja,	nicht gefunden	   */

	if (islower(*mit))			/* Ausdruck im Op.feld	  */
	  {					/* (x1,	x2, p1,	p2) :	  */
	    ++mit;				/* x oder p uebergehen	  */
	    nest = 0;				/* Klammerungstiefe neu	  */
	    while (!atend(*inst))		/* Ausdruck durchgehen	  */
	      {
		if (*inst == ',')		/* bis zum naechsten ','  */
		  break;
		if (*inst == ')' && nest == 0)	/* oder	bis unbalancierte */
		  break;			/* ')' gefunden		  */
		switch (*inst)			/* Klammerungstiefe	  */
		  {				/* abtesten / setzen	  */
		    case '(':  ++nest;	break;
		    case ')':  --nest;	break;
		  }
		*exp++ = *inst++;		/* Ausruck kopieren	  */
	      }
	    *exp++ = ','; *exp = '\0';		/* kop.	Ausdruck beenden  */
	    continue;				/* weiter im Operanden	  */
	  }

	if (lexorder(*inst++,*mit++))	/* Operanden mit MIT vergleichen  */
	  {
	    next :
	    while (*mit)		/* ungleich, naechsten Operand	  */
	      {				/* im MIT-Op.feld anwaehlen	  */
		if (*mit == ANOTHER)	/* Ende	gerader	versuchter	  */
		  {			/* Operand im MIT-Operandenfeld	: */
		    ++opadj;		/*   merken, welcher Op. im Feld  */
		    ++mit;		/*   '|' uebergehen		  */
		    inst = backup;	/*   ab	Operand	in Inst. von vorn */
		    exp	= expbuf;	/*   Ausdrucksbuffer resetten	  */
		    break;
		  }
		++mit;			/* bis zum Ende	des Operanden	  */
	      }
	    if (atend(*mit))		/* naechster Op. in MIT	Op.feld	? */
	      return (NO);		/* keiner mehr,	Vergleich negativ */
	  }

      }	/* end LOOP */
  }



/* Ende	von MIT.C */
