/********************************************************/
/*							*/
/*		 Q/C Compiler (Part 5)			*/
/*							*/
/*     Copyright (c) 1985 Quality Computer Systems	*/
/*		 All Rights Reserved			*/
/*							*/
/*			07/24/85			*/
/********************************************************/

#include <stdio.h>
#include <ctype.h>
#include "qc.h"
#include "ctp.h"
#include "qcdecl.c"


/* This	section	identifies tokens in declarations
 *	and manages all	of the compiler	tables
 */

/* Check next token to see if it is a keyword */
int chkeyword()
	{
	return (  chks("break")	    || chks("case")	|| chks("continue")
	       || chks("default")   || chks("do")	|| chks("else")
	       || chks("entry")	    || chks("enum")	|| chks("for")
	       || chks("goto")	    || chks("if")	|| chks("return")
	       || chks("sizeof")    || chks("switch")	|| chks("while")
	       || chkdecl());
	}
/* Check to see	if we have a declaration */
int chkdecl()
	{
	return (chktype() || chksc());
	}
/* Check for a valid C variable	type */
int chktype()
	{
	blanks();
	if (chks("int"))
		return T_INT;
	else if	(chks("char"))
		return T_CHAR;
	else if	(chks("unsigned"))
		return T_UNSIGNED;
	else if	(chks("struct"))
		return T_STRUCT;
	else if	(chks("union"))
		return T_UNION;
	else if	(chks("short"))
		return T_SHORT;
	else if	(chks("long")) {
		lngused	= TRUE;
		return T_LONG;
		}
	else if	(chks("float") || chks("double")) {
		fltused	= TRUE;
		return T_FLOAT;
		}
	else
		return T_NONE;
	}
/* Check for a valid C variable	storage	class */
int chksc()
	{
	blanks();
	if (chks("register"))
		return SC_REG;
	else if	(chks("static"))
		return SC_STATIC;
	else if	(chks("extern"))
		return SC_EXTERN;
	else if	(chks("typedef"))
		return SC_TYPE;
	else if	(chks("auto"))
		return SC_AUTO;
	else
		return SC_NONE;
	}
/* Get a valid C variable type if present */
struct typeinfo	*
gettype()
	{
	register type, newtype;
	struct typeinfo	*istypedef(), *decltag(), *p_type;

	type = T_NONE;		/* start with no type */
	for (;;) {		/* process all type specifiers */
		switch (newtype	= chktype()) {
		case T_LONG:
		case T_SHORT:
		case T_FLOAT:
		case T_UNION:
		case T_STRUCT:
		case T_UNSIGNED:
		case T_CHAR:
			if (type == T_NONE)
				type = newtype;
			else
				typerr();
			break;
		case T_INT:
			if (type == T_LONG || type == T_SHORT ||
				 type == T_UNSIGNED) /*	type may be T_INT if we	*/
				break; /* found	long, short, or	unsigned already */
			else if	(type == T_NONE)
				type = T_SHORT;
			else
				typerr();
			break;
		default:
			switch (type) {
			case T_STRUCT:
			case T_UNION:
				return decltag(type);
			case T_UNSIGNED:
				return unsgntype;
			case T_CHAR:
				return chartype;
			case T_SHORT:
				return sinttype;
			case T_LONG:
				return linttype;
			case T_FLOAT:
				return floattype;
			case T_NONE:
				if ((p_type = istypedef()) != NULL)
					skip();	/* skip	typedef	name */
				return p_type;
				}
			}
		skip();			/* skip	the type keyword */
		}
	}
/* See if next token is	a typedef name */
struct typeinfo	*
istypedef()
	{
	register struct	st *sym;
	struct st *findglb(), *findloc();
	char *chksym(),	name[NAMESIZE];

	if (!chksym(name))
		return NULL;
	if (infunc && (sym = findloc(name))
	   || (sym = findglb(name)) && sym->st_info == DECL_GLB) {
		if (sym->st_sc == SC_TYPE)
			return sym->st_type;
		}
	return NULL;
	}
/* Get a valid C variable storage class	if present */
int getsc()
	{
	int sc;
	if ((sc	= chksc()) != 0)
		skip();		/* skip	over the keyword */
	return sc;
	}
/* Parse an abstract declarator	*/
struct typeinfo	*
abstdecl()
	{
	register struct	typeinfo *type;
	struct typeinfo	*gettype(), *loadtype();
	struct parsestack ts;

	if ((type = gettype()) != NULL)	{
		ts.curr_ptr = ts.stack;	/* initialize local stack */
		dodecl(&ts, NULL);	/* analyze the declarator */
		type = loadtype(type, &ts);
		if (type->t_code == T_ARRAY && type->t_size == -1)
			sizerr();
		return type;
		}
	else
		return NULL;
	}
/* Load	a local	type entry into	the type table */
struct typeinfo	*
loadtype(type, ts)
register struct	typeinfo *type;
struct parsestack *ts;
	{
	register struct	typestack *p, *startstack;
	register size;
	struct typeinfo	*addtype(), *ptrto();
	union baseinfo base;

	startstack = ts->stack;
	for (p = ts->curr_ptr; --p >= startstack; ) {
		switch (p->t_code) {
		case T_PTR:
			type = ptrto(type);
			break;
		case T_FUNC:
			if (type->t_code<=T_NONE || type->t_code>T_SIMPLE) {
				/*Error	#436 Function can't return aggregate */
				error(436);
				type = sinttype;
				}
			base.p_type = type;
			type = addtype(T_FUNC, &base, 0);
			break;
		case T_ARRAY:
			if (type->t_code == T_FUNC) {
				/*Error	#437 No	arrays of functions */
				error(437);
				type = ptrto(type);
				}
			if (p->t_size != -1)
				size = p->t_size * type->t_size;
			else {
				if (p == startstack)
					size = -1; /* 1st dimension missing */
				else if	(p == startstack+1
				     &&	(p-1)->t_code == T_PTR)
					size = -1; /* ptr-to-array */
				else {
					sizerr();
					size = type->t_size;
					}
				}
			base.p_type = type;
			type = addtype(T_ARRAY,	&base, size);
			break;
			}
		}
	return type;
	}
/* Process a declarator	*/
void dodecl(ts,	name)
register struct	parsestack *ts;
register char *name;
	{
	if (matchc('*')) {
		dodecl(ts, name);
		pushtype(ts, T_PTR, 1);
		return;
		}
	if (matchc('(')) {
		dodecl(ts, name);
		needpunc(')');
		}
	else if	(name != NULL)	/* no name for abstract	declarator */
		symname(name);
	if (matchc('(')) {
		if (name == NULL /* real declarations will do own chk */
		   || ts->curr_ptr > ts->stack)	/* pointer-to-func */
			needpunc(')');
		pushtype(ts, T_FUNC, 1);
		}
	else while (matchc('['))
		pushtype(ts, T_ARRAY, arraysize());
	}
/* Push	a type onto a local type parsing stack */
void pushtype(ts, code,	itemct)
register struct	parsestack *ts;
int code, itemct;
	{
	register struct	typestack *tsptr;

	tsptr =	ts->curr_ptr;
	if (tsptr < &ts->stack[	sizeof(ts->stack)/sizeof(struct	typestack) ]) {
		tsptr->t_code =	code;
		tsptr->t_size =	itemct;
		++ts->curr_ptr;
		}
	else
		/*Error	#638 This type is too ornate */
		error(638);
	}
/* Test	for legal symbol name */
int symname(name)
char *name;
	{
	char *chksym();
	if (!chksym(name)) {
		skip();		/* pass	by the name */
		illname();	/* illegal symbol name */
		return FALSE;
		}
	else if	(chkeyword()) {
		skip();
		keyerr(name);	/* illegal use of keyword */
		return FALSE;
		}
	else {			/* valid symbol	name */
		skip();
		return TRUE;
		}
	}
/* Check for a valid symbol */
char *chksym(name)
register char *name;
	{
	register i;
	register char *p;
	blanks();
	if (!issymf(ch())) {	/* name	must start with	an alpha */
		*name =	'\0';
		return FALSE;
		}
	i = 0;
	p = lptr;
	while (issym(*p)) {
		if (i++	< NAMEMAX)
			*name++	= *p;
		++p;
		}
	*name =	'\0';
	return p;		/* return location of char after name */
	}
/* Add a statement label to the	local symbol table */
int addlabel(labname, sc)
register char *labname;
register sc;
	{
	register struct	st *p;
	register labno;
	struct st *addloc(), *findloc();
	/* check if labname is already in local	symbol table */
	if ((p = findloc(labname)) != NULL) {
		if (p->st_type->t_code != T_LABEL) {
			multidef(labname); /* previously */
			return 0;	/* defined as a	variable */
			}
		else if	(p->st_sc == SC_AUTO) {
			if (sc == SC_AUTO) /* 2nd definition */
				multidef(labname);
			}
		else /*	labname	previously used	in goto	only */
			p->st_sc = sc;	/* define it now */
		return p->st_info;	/* return label	# assigned */
		}
	addloc(labname,	labeltype, sc, ID_VAR, labno = getlabel());
	return labno;
	}
/* Return next available internal label	number */
int getlabel()
	{
	static label = 1;
	return (++label);
	}
/* Add an entry	to the switch/loop queue */
void addswq(p)
struct swq *p;
	{
	if (pswq >= swq+nswq)
		/*Error	#739 Too many active switches or loops */
		error(739);
	else {
		pswq->loop = p->loop;
		pswq->exit = p->exit;
		++pswq;
		}
	}
/* Delete an entry from	the switch/loop	queue */
void delswq()
	{
	struct swq *getswq();
	if (getswq())
		--pswq;
	else
		/*Error	#840 No	active switches	or loops to delete */
		error(840);
	}
/* Get the most	recent entry in	the switch/loop	queue */
struct swq *
getswq()
	{
	return (pswq ==	swq) ? NULL: pswq - 1;
	}
/* Add an entry	to the switch case table */
void addcase(first_case, new_case)
struct case_table *first_case;
register struct	case_table *new_case;
	{
	register struct	case_table *p;

	for (p = first_case; p < pcasetab; ++p)	{
		if (new_case->value == p->value) {
			/*Error	#441 Duplicate switch case */
			error(441);
			break;
			}
		}
	if (pcasetab >=	casetab+ncases)
		/*Error	#742 Too many switch cases */
		error(742);
	else {
		pcasetab->value	= new_case->value;
		pcasetab->label	= new_case->label;
		++pcasetab;
		}
	}
/* Delete an entry from	the switch case	table */
struct case_table *
delcase()
	{
	if (pcasetab > casetab)
		return (--pcasetab);
	else {
		/*Error	#843 No	entry in case table to delete */
		error(843);
		return NULL;
		}
	}

/* See if name is defined in macro pool	*/
struct mp_hash *
findmac(name)
char *name;
	{
	register struct	mp_hash	*p;

	p = mac_hash + hash(name, mp_size); /* primary hash location for name*/
	while (p != NULL) {
		if (strcmp(p->name, name) == 0)
			return p;	/* found name, return slot ptr */
		p = p->next_mac;	/* follow synonym chain	if any */
		}
	return NULL;
	}
/* Add a macro definition to macro pool	*/
void addmac(name)
register char *name;			/* name	to be added */
	{
	register struct	mp_hash	*mp, *mq;
	register c;
	register char *p;
	char repl_text[LINESIZE];
	struct mp_hash *findmac();

	while (ch() == ' ' || ch() == '\t') /* skip to replacement text	*/
		gch();
	p = repl_text;
	while ((c = gch()) != '\0')	/* get replacement text	*/
		*p++ = c;
	*p = '\0';
	if ((mp	= findmac(name)) != NULL) {	/* already defined? */
		if (strcmp(mp->def, repl_text))	/* redefinition	must */
			multidef(name);		/*	be the same */
		return;
		}
	mp = mac_hash +	hash(name, mp_size);/* primary hash location for name*/
	if (mp->def != NULL) {		/* if primary location already used */
		while (mp->next_mac != NULL)	/* put sym in overflow area */
			mp = mp->next_mac;	/* find	end of synonym chain */
		for (mq	= mac_ovflow; mq->def != NULL; --mq) { /* find empty slot */
			if (mq < mac_hash)
				break;
			}
		if (mq < mac_hash) {
			/*Error	#744 Macro (#define) pool is full */
			error(744);
			return;
			}
		mp->next_mac = mq;	/* add new entry to chain */
		mac_ovflow = mq	- 1;	/* reset next overflow slot pointer */
		mp = mq;
		}
	if ((int)(mp->def = clrsbrk(strlen(repl_text) +	1)) == -1) {
		puts("\nOut of memory");
		exit(1);
		}
	strcpy(mp->name, name);		/* stow	define name and	*/
	strcpy(mp->def,	repl_text);	/* replacement text and	*/
	++n_def;
	}
/* Remove next name from the #define macro pool	*/
void delmac()
	{
	register struct	mp_hash	*p;
	struct mp_hash *isdef();

	if ((p = isdef()) != NULL) {	/* if the name is defined, */
		*p->name = '\0';	/* wipe	it out and */
		p->def = NULL;		/* indicate slot is empty */
		--n_def;		/* and one less	slot used */
		}
	else
		/*Error	#245 Can't #undef - not	defined	*/
		error(245);
	}
/* See if the next token is a #defined name */
struct mp_hash *
isdef()
	{
	struct mp_hash *findmac();
	char name[NAMESIZE];

	if (symname(name))
		return findmac(name);
	else
		return NULL;
	}
/*	Evaluate a constant expression		*/
int constexp(cexp)
register struct	operand	*cexp;
	{
	register saveflag;

	saveflag = genflag;
	genflag	= FALSE;	/* kill	code generation	*/
	heir3(cexp);		/* no comma operators or assignment */
	genflag	= saveflag;
	unreftype(cexp->op_type);
	return (cexp->op_sym) ?	/* if a	symbol was found ... */
		(cexp->op_load & CONSTADDR): /*	chk constant address */
		(cexp->op_load & CONSTANT);  /*	else chk for constant */
	}
/* If next token is a constant build an	operand	with its value */
int constant(op)
register struct	operand	*op;
	{
	register char *p;
	register c, type, sign;
	static int dot,	exp, leading_zero, is_hex, is_char, is_float;
	static char num[21], fnum[21], *save_curr;
	long int val;

	blanks();			/* skip	leading	white space */
	sign = 0;
	is_float = is_char = is_hex = leading_zero = FALSE;
	switch (c = ch()) {

	case '\'':
		is_char	= TRUE;
		gch();			/* eat up leading apostrophe */
		break;
	case '+':
	case '-':
		if (!isdigit(nch()))
			return FALSE;
		sign = gch();
		break;
	case '.':
		if (!isdigit(nch()))
			/*Error	#446 Invalid float constant */
			error(446);
		break;
	case '0':
		if (match("0x")	|| match("0X"))	{
			is_hex = TRUE;
			if (!isxdigit(ch()))
				/*Error	#447 Invalid hex constant */
				error(447);
			}
		else
			leading_zero = TRUE;
		break;
	default:
		if (!isdigit(c))
			return FALSE;
		break;
		}
	if (!is_hex && !is_char) {
		p = num;		/* build a string containing number */
		if (sign)
			*p++ = sign;
		dot = exp = FALSE;
		for (;;) {
			c = toupper(ch());
			if (isdigit(c))
				*p++ = gch();
			else if	(c == '.') {	/* decimal point? */
				if (dot)
					break;	/* already saw one */
				*p++ = gch();
				is_float = dot = TRUE;
				}
			else if	(c == 'E') {	/* exponent? */
				if (exp)
					break;
				*p++ = gch();
				is_float = exp = TRUE;
				c = ch();	/* signed exponent? */
				if (c == '+' ||	c == '-')
					*p++ = gch();
				}
			else if	(c == 'L') {	/* long	suffix?	*/
				if (is_float)
					break;
				*p++ = gch();
				break;
				}
			else
				break;
			}
		*p = '\0';
		}
	if (is_float)			/* floating point? */
		type = T_FLOAT;
	else if	(is_hex)		/* hex?	*/
		type = hex(&val);
	else if	(is_char)
		type = charconst(&val);
	else if	(leading_zero)		/* octal? */
		type = octal(&val, num);
	else
		type = decimal(&val, num);

	op->op_sym = NULL;		/* build operand descriptor */
	op->op_load = (LOADVALUE|CONSTANT);
	if (type != T_FLOAT)
		op->op_val = val;
	else {
		save_curr = lptr;	/* save	current	parsing	location */
		fnum[0]	= '"';		/* load	constant in literal pool */
		strcpy(&fnum[1], num);
		lptr = fnum;
		string(op);
		ld_float(op);		/* do conversion at run	time */
		lptr = save_curr;	/* restore parsing pointer */
		}
	switch (type) {
	case T_INT:
		reftype(op->op_type = sinttype);
		break;
	case T_LONG:
		reftype(op->op_type = linttype);
		lngused	= TRUE;
		break;
	case T_FLOAT:
		reftype(op->op_type = floattype);
		fltused	= TRUE;
		break;
		}
	return TRUE;
	}
/* Check for hexadecimal constant */
int hex(val)
long *val;
	{
	register int d,	long_no;
	long v;

	v = 0;
	long_no	= FALSE;
	for (;;) {
		d = toupper(ch());
		if (isdigit(d))
			d -= '0';
		else if	(isxdigit(d))
			d -= ('A' - 10);
		else if	(d == 'L') {
			long_no	= TRUE;
			gch();
			break;
			}
		else
			break;
		v = (v<<4) | d;
		gch();
		if (v >	65535)		/* long	if > max unsigned integer */
			long_no	= TRUE;
		}
	*val = v;
	return ((long_no) ? T_LONG : T_INT);
	}
/* Check for octal constant */
int octal(val, s)
long *val;
char *s;
	{
	register int c,	long_no;
	long v;

	v = 0;
	long_no	= FALSE;
	while (isdigit(c = *s++)) {
		if (c >	'7')
			/*Error	#448 Invalid octal constant */
			error(448);
		v = (v<<3) | (c	- '0');
		if (v >	65535)		/* long	if > max unsigned integer */
			long_no	= TRUE;
		}
	*val = v;
	return ((long_no || toupper(c) == 'L') ? T_LONG	: T_INT);
	}
/* Check for decimal constant */
int decimal(val, s)
long *val;
char *s;
	{
	int positive;
	register c;
	long d,	v;

	c = *s;
	positive = (c != '-');
	if (c == '+' ||	c == '-')
		++s;			/* eat up the leading sign */
	v = 0;
	while (isdigit(*s)) {		/* build negative to handle largest */
		d = *s++ - '0';		/* negative number in 2's complement */
		v = v *	10 - d;
		}
	if (positive)
		v = -v;
	*val = v;
	if (v >	32767 || v < -32768 || toupper(*s) == 'L')
		return T_LONG;		/* long	if > max signed	integer	*/
	else
		return T_INT;
	}
/* Check for character constant	*/
int charconst(val)
long *val;
	{
	register c, k;
	k = 0;
	while ((c=gch()) != '\'' && c != 0) {
		if (c == '\\')
			c = escape();
		k = (k << 8) + (c & 0xFF);
		}
	*val = k;
	return T_INT;
	}
/* Check for string, build an operand, and add it to the literal pool */
int string(op)
struct operand *op;
	{
	register c;
	static lit_full	= FALSE;

	if (!matchc('"'))
		return FALSE;
	initop(op, &lit_sym, lit_sym.st_type, (LOADADDR|CONSTADDR));
	op->op_val = plitpool-litpool;	/* offset from start of	litpool	*/
	while ((c = getstring()) != -1)	{
		if (plitpool-litpool < litpsize-1)
			*plitpool++ = c;
		else if	(lit_full)
			continue;
		else {
			/*Error	#749 String space full */
			error(749);
			lit_full = TRUE;
			}
		}
	if (plitpool-litpool < litpsize)
		*plitpool++ = '\0';
	return TRUE;
	}
/* Get the next	character in a literal string */
int getstring()
	{
	int c;
	switch (c = gch()) {
	case '"':
	case '\0':
		return -1;
	case '\\':
		return escape();
	default:
		return c;
		}
	}
/* Evaluate an escape character	constant */
char escape()
	{
	register c,count,octal;
	switch (c = gch()) {
	case 'f':
		return '\f';
	case 'r':
		return '\r';
	case 'b':
		return '\b';
	case 't':
		return '\t';
	case 'n':
		return '\n';
		}
	if (isdigit(c))	{	  /* octal sequence */
		count =	0;
		octal =	c - '0';
		while (isdigit(ch()) &&	++count<3)
			octal =	octal *	8 + gch() - '0';
		return octal;
		}
	return c;
	}
/*
 *	Symbol table handling functions
 */
struct st *
findglb(name)
char *name;
	{
	struct st *find_sym();
	return find_sym(name, ID_VAR);
	}
struct st *
findloc(name)
char *name;
	{
	struct st *findvar();
	return findvar(locptr+1, symtab+nsym, name, ID_VAR);
	}
struct st *
findtag(name)
char *name;
	{
	struct st *p, *find_sym(), *findvar();

	if ((p = findvar(locptr+1, symtab+nsym,	name, ID_STRUCT)) != NULL)
		return p;
	else
		return find_sym(name, ID_STRUCT);
	}

struct st *
findvar(start, end, name, idset)
struct st *start;
register struct	st *end;
register char *name;
register int idset;
	{
	register struct	st *p;
	for (p = start;	p < end; ++p){
		if (astreq(name, p->st_name, NAMEMAX)
			&& p->st_idset==idset)
			return p;
		}
	return NULL;
	}

struct st *addglb(name,type,sc,idset,info)
register char *name;
register struct	typeinfo *type;
int sc,	idset, info;
	{
	register struct	typeinfo *def_type, *decl_type;
	struct st *find_sym(), *addvar();
	register struct	st *p;

	if (idset != ID_STRUCT && (p = find_sym(name, ID_VAR)) != NULL)	{
		/* check for consistency (struct info is checked when */
		/* it is declared since	locals must be checked also) */
		if (sc == SC_EXTERN) {		/* is this the declaration? */
			decl_type = type;	/* record declared type	*/
			def_type = p->st_type;	/*	and defined type */
			}
		else {				/* this	is the definition */
			decl_type = p->st_type;
			def_type = type;
			if (p->st_sc ==	SC_GLOBAL || p->st_sc == SC_ST_GLB) {
				multidef(name);	/* already defined */
				return NULL;
				}
			p->st_sc = sc;		/* say it's defined */
			p->st_info = info;
			}
		if (decl_type != def_type) {	/* declare/define types	same?*/
			if (def_type->t_code ==	T_ARRAY	/* array size may */
			&& decl_type->t_code ==	T_ARRAY	/* be missing but */
			&& def_type->t_base.p_type ==	/* base	types must */
			   decl_type->t_base.p_type) {	/* be the same */
				if (sc != SC_EXTERN) {	/* is this the def? */
				    reftype(def_type);
				    unreftype(decl_type);
				    p->st_type = def_type; /*use defined size*/
				    }
				}
			else
				/*Error	#450 Inconsistent declaration */
				errname(450, name);
			}
		}
	else if	((p = addvar(glbptr,name,type,sc,idset,info)) != NULL) {
		++glbptr;
		add_sym(name, p);
		}
	return p;
	}
struct st *
addloc(name,type,sc,idset,info)
char *name;
struct typeinfo	*type;
int sc,	idset, info;
	{
	struct st *p, *addvar();

	if ((p = addvar(locptr,name,type,sc,idset,info)) != NULL)
		--locptr;
	return p;
	}
struct st *
addvar(p,name,type,sc,idset,info)
register struct	st *p;
char *name;
register struct	typeinfo *type;
int sc,	idset, info;
	{
	if (glbptr >= locptr) {
		/*Error	#751 Symbol table full */
		error(751);
		return NULL;
		}
	if (type != NULL) {
		reftype(type);
		if (type->t_code == T_STRUCT &&	idset == ID_VAR
		   && type->t_size == 0)
			/*Error	#452 No	template */
			error(452);
		}
	strcpy(p->st_name, name);
	p->st_type  = type;
	p->st_sc = sc;
	p->st_idset = idset;
	p->st_info = info;
	return p;
	}
/* Look	for a global symbol using hashing index	*/
struct st *
find_sym(name, id)
char *name;
int id;
	{
	register struct	st *sym;
	register struct	st_hash	*p;

	p = sym_hash + hash(name, st_size); /* primary hash location for name*/
/* replaced :  while (p	!= NULL) {   ??? */
	while (p!=NULL && p->sym!=NULL)	{
		sym = p->sym;
		if (strcmp(sym->st_name, name) == 0 && sym->st_idset ==	id)
			return sym;	/* found name */
		p = p->next_sym;	/* follow synonym chain	if any */
		}
	return ((struct	st *) NULL);
	}
/* Add a global	symbol to symbol table hash index */
void add_sym(name, sym)
char *name;				/* name	to be added */
struct st *sym;				/* symbol table	entry */
	{
	register struct	st_hash	*p, *q;

	p = sym_hash + hash(name, st_size); /* primary hash location for name*/
	if (p->sym == NULL)
		p->sym = sym;		/* use primary location	*/
	else {				/* put sym in overflow area */
		while (p->next_sym != NULL)
			p = p->next_sym; /* find end of	synonym	chain */
		for (q = sym_ovflow; q->sym != NULL; --q)
			;		/* find	next empty hash	index slot */
		p->next_sym = q;	/* add new entry to chain */
		q->sym = sym;		/* stow	new entry */
		sym_ovflow = q - 1;	/* reset next overflow slot pointer */
		}
	}
/* Hashing function for	symbol table and macro pool */
int hash(name, size)
char *name;
int size;
	{
	register c, value;

	value =	0;
	while ((c = *name++) !=	'\0')
		value =	(value << 1) ^ c;
	value %= size;
	return (value);
	}
/*
 *	Member table handling functions
 */
/* Add a member	to the member table */
struct memtab *
addmem(sym, type, prevmem)
struct st *sym;
struct typeinfo	*type;
struct memtab *prevmem;
	{
	register struct	memtab *currmem;
	if (freemem == NULL) {
		/*Error	#753 Member table full */
		error(753);
		return NULL;
		}
	else {
		currmem	= freemem;	/* use next free member	slot */
		freemem	= freemem->nextmem; /* update free list	ptr */
		currmem->sym = sym;	/* record symbol table entry */
		currmem->nextmem = NULL; /* end	of current list	*/
		if (prevmem == NULL)	/* first member	in this	struct */
			type->t_base.memlist = currmem;
		else
			prevmem->nextmem = currmem; /* chain members */
		return (currmem);
		}
	}
/* Delete an entire member list	*/
void delmem(memlist)
struct memtab *memlist;
	{
	register struct	memtab *p;
	if (freemem != NULL) {	/* add memlist to front	of free	list */
		for (p = memlist; p->nextmem !=	NULL; p	= p->nextmem)
			;	/* find	last member in list */
		p->nextmem = freemem; /* make it point to old free list	*/
		}
	freemem	= memlist;
	}
/*
 *	Type handling functions
 */
void inittypes()
	{
	register i;
	struct typeinfo	*addtype();
	for (i=0; i<n_types; ++i) {
		typelist[i].t_code = T_NONE;
		typelist[i].t_next = &typelist[i+1];
		}
	typelist[n_types-1].t_next = NULL;
	for (i=0; i<T_MAX; ++i)
		basetypes[i] = NULL;
	basetypes[T_NONE] = typelist;
	/* define some oft-used	types, and mark	them permanent */
	chartype = addtype(T_CHAR, (union baseinfo *) NULL, S_CHAR);
	reftype(chartype);
	sinttype  = addtype(T_SHORT, (union baseinfo *)	NULL, S_SHORT);
	reftype(sinttype);
	unsgntype = addtype(T_UNSIGNED,(union baseinfo *)NULL,S_UNSIGNED);
	reftype(unsgntype);
	linttype  = addtype(T_LONG, (union baseinfo *) NULL, S_LONG);
	reftype(linttype);
	floattype = addtype(T_FLOAT,(union baseinfo *)NULL,S_FLOAT);
	reftype(floattype);
	labeltype = addtype(T_LABEL, (union baseinfo *)	NULL, 0);
	reftype(labeltype);
	ifunctype = addtype(T_FUNC, (union baseinfo *) &sinttype, 0);
	reftype(ifunctype);
	}
/* Add a new type to table */
struct typeinfo	*
addtype(typecode, base,	size)
int typecode;
union baseinfo *base;
int size;
	{
	register struct	typeinfo *p;
	struct typeinfo	*findtype(), *uniquetype();

	if ((p=findtype(typecode,base,size)) ==	NULL)
		p = uniquetype(typecode,base,size);
	return p;
	}
struct typeinfo	*
uniquetype(typecode, base, size)
int typecode;
union baseinfo *base;
int size;
	{
	register struct	typeinfo *p;
	struct typeinfo	*alloctype();

	if ((p=alloctype(typecode,base,size)) == NULL)
		return sinttype;
	else {	p->t_next = basetypes[typecode];
		basetypes[typecode] = p;
		return p;
		}
	}
void reftype(p)
struct typeinfo	*p;
	{
	++p->t_refs;
	}
void unreftype(p)
struct typeinfo	*p;
	{
	if (--p->t_refs	== 0)
		scraptype(p);
	}
/* Convert "type" to "ptr-to-type" */
struct typeinfo	*
ptrto(type)
struct typeinfo	*type;
	{
	struct typeinfo	*addtype();

	return addtype(T_PTR, (union baseinfo *)&type, S_PTR);
	}
/*  Find a type	in the table */
struct typeinfo	*
findtype(typecode,base,size)
int typecode;
union baseinfo *base;
int size;
	{
	register struct	typeinfo *p;

	for (p=basetypes[typecode]; p!=NULL; p=p->t_next) {
		if (p->t_size != size)
			continue;
		if (typecode ==	T_STRUCT) {
			if (p->t_base.memlist == base->memlist)
				break;
			}
		else if	(p->t_base.p_type == base->p_type)
			break;
		}
	return p;
	}
/* Allocate an entry in	the type table */
struct typeinfo	*
alloctype(typecode,base,size)
int typecode;
register union baseinfo	*base;
int size;
	{
	register struct	typeinfo *p, *p_base;

	p_base = (base==NULL ||	typecode==T_STRUCT) ? NULL: base->p_type;
	if (p_base != NULL)
		reftype(p_base);
	if (basetypes[T_NONE] == NULL && !typecollect()) {
		/*Error	#754 Too many different	types in use */
		error(754);
		if (p_base != NULL)
			unreftype(p_base);
		return NULL;
		}
	else {	p = basetypes[T_NONE];
		basetypes[T_NONE] = p->t_next;
		p->t_code = typecode;
		p->t_size = size;
		p->t_refs = 0;
		p->t_next = NULL;
		if (typecode ==	T_STRUCT)
			p->t_base.memlist = NULL; /* no	member list yet	*/
		else
			p->t_base.p_type = p_base;
		return p;
		}
	}
/* Remove a type from the type table */
void scraptype(p)
register struct	typeinfo *p;
{	register struct	typeinfo *t, *prev;

	if (p->t_code != T_STRUCT && p->t_base.p_type != NULL)
		unreftype(p->t_base.p_type);
	prev = NULL;
	for (t=basetypes[(unsigned)p->t_code]; t!=NULL;	t=t->t_next) {
		if (t == p) {
			freetype(prev,p);
			break;
			}
		prev = t;
		}
	}
/* Return a type node to the free list */
void freetype(prev,type)
struct typeinfo	*prev;
register struct	typeinfo *type;
	{
	if (prev == NULL)
		basetypes[(unsigned)type->t_code] = type->t_next;
	else
		prev->t_next = type->t_next;
	type->t_next = basetypes[T_NONE];
	basetypes[T_NONE] = type;
	type->t_code = T_NONE;
	}
int typecollect()
	{
	register code, freed;
	register struct	typeinfo *p, *prev, *next;

	freed =	0;
	for (code=T_NONE+1; code<T_MAX;	++code)	{
	    prev = NULL;
	    for	(p=basetypes[code]; p!=NULL; p = next) {
		next = p->t_next;
		if (p->t_refs>0)
			prev = p;
		else {
		/* Unused node - free it after removing
		 * the reference to its	base type.
		 * Unreftype() isn't used because a node
		 * somewhere in	the chain from here to the
		 * real	base type may have already been	freed,
		 * and that will make unreftype() choke.
		 * Instead, we just decrement the use count,
		 * and let the next collection pick it up.
		 */
			if (code!=T_STRUCT && p->t_base.p_type!=NULL)
				p->t_base.p_type->t_refs--;
			freetype(prev, p);
			++freed;
			}
		}
	    }
	return freed;
	}
/* Strip a layer off of	op's type */
void deref(op)
struct operand *op;
	{
	chtype(op, op->op_type->t_base.p_type);
	}
/* Set type of expression to highest of	either operand */
void settype(lop, rop)
struct operand *lop, *rop;
	{
	register struct	typeinfo *rtype;

	rtype =	rop->op_type;
	if (rtype->t_code > lop->op_type->t_code)
		chtype(lop, rtype);
	unreftype(rtype);
	if (lop->op_type->t_code == T_ARRAY) /*	Change array to	ptr */
		chtype(lop, ptrto(lop->op_type->t_base.p_type));
	}
void chtype(op,new)
register struct	operand	*op;
register struct	typeinfo *new;
	{
	reftype(new);
	unreftype(op->op_type);
	op->op_type = new;
	if (new->t_code	== T_ARRAY)
		unarray(op);
	}
int basesize(op)
struct operand *op;
	{
	register struct	typeinfo *p;

	p = op->op_type->t_base.p_type;
	if (p->t_size == 0)
		/*Error	#455 Size unknown */
		error(455);
	return p->t_size;
	}
/* Check op to see if it is a pointer */
struct typeinfo	*
isptr(op)
struct operand *op;
	{
	register struct	typeinfo *type;
	struct typeinfo	*ptrto();

	type = op->op_type;
	if (type->t_code == T_PTR)
		return type;
	else if	(type->t_code == T_ARRAY)
		return ptrto(type->t_base.p_type);
	else
		return NULL;
	}
int islval(op)
struct operand *op;
	{
	return (op->op_load & LVALUE);
	}
int isconstant(op)
struct operand *op;
	{
	return (op->op_load & CONSTANT);
	}
int isloaded(op)
struct operand *op;
	{
	return (op->op_load & (LOADVALUE|LOADADDR)) == 0;
	}
int isaddr(op)
struct operand *op;
	{
	return (islval(op) && op->op_sym == NULL);
	}
/* determine if	type is	one that is stored as 32 bits */
int is32bit(chktyp)
struct typeinfo	*chktyp;
	{
	register chkcode;
	chkcode	= chktyp->t_code;
	return (chkcode	== T_LONG || chkcode ==	T_FLOAT);
	}
/* determine type to use in computations */
struct typeinfo	*
comptyp(lop, rop)
struct operand *lop, *rop;
	{
	register struct	typeinfo *ltype, *rtype;
	ltype =	lop->op_type;
	rtype =	rop->op_type;
	return ((rtype->t_code > ltype->t_code)	? rtype	: ltype);
	}
/* Determine if	a conversion is	required to go from one	type to	another	*/
int cnvreq(typ1, typ2)
register struct	typeinfo *typ1,	*typ2;
	{
	if (is32bit(typ1))
		return (typ1 !=	typ2);
	else if	(is32bit(typ2))
		return TRUE;
	else
		return FALSE;
	}
/* end of QC5.C	*/

