/********************************************************/
/*							*/
/*		 Q/C Compiler (Part 7)			*/
/*							*/
/*     Copyright (c) 1985 Quality Computer Systems	*/
/*		 All Rights Reserved			*/
/*							*/
/*			07/13/85			*/
/********************************************************/

#include <stdio.h>
#include "qc.h"
#include "qcdecl.c"


/* This	section	contains all code generation routines
 * which generate mnemonics for	a specific assembler.
 */

/* Names of Z80	registers used */
static char *z80reg[] =	{"HL", "DE", "BC", "IX"};

/* Print assembler info	before any code	is generated */
void header()
	{
	outstr(";Compiled by Q/C ");
	outstr(version);
	nl();
	ol(".Z80");
	}
/* Generate code to be done before main() */
void shell(argcnt)
int argcnt;
	{
	static char *func;
	char hex[5];
	char *itob();

	codeseg();
	glblabel("_main", SC_GLOBAL);	/* establish entry point */
	if (romflag) {
		ot("LD\tSP,0");		/* stack addr from -S switch */
		outstr(itob(spaddr, hex, 16));
		outstr("H\n");
		}
	else {
		ol("LD\tSP,(6)");	/* start stack below CP/M BDOS */
		if (redirect)		/* full	redirection requested */
			func = "_rshell";
		else if	(argcnt)	/* parse CP/M command line */
			func = "_shell";
		else {
			ccall("main");
			func = "_exit";
			}
		ccall(func);
		declext(func);
		}
	}
/* Print any assembler stuff needed after all code */
void trailer()
	{
	static int n;
	static struct st *p;
	static struct typeinfo *t;
	char *regname();
	if (libflag)			/* building a library? */
		ol("END");
	else {				/* declare externals */
		for (p = symtab; p < glbptr; ++p) {
			if (p->st_sc ==	SC_EXTERN)
				declext(p->st_name);
			t = p->st_type;
			if (t->t_code == T_STRUCT && t->t_size == 0
			   && p->st_sc == SC_TYPE)
				tagerr(p->st_name);
			}
		for (n = 0; n <	MAXREG;	) /* declare registers */
			declext(regname(++n));
		sextout();
		if ( lngused ||	fltused	)
			lextout();
		if ( fltused )
			fextout();
		ot("END");
		if (mainflag) {
			outbyte('\t');	/* define entry	point */
			outname("_main");
			}
		nl();
		}
	}
/* Output short	integer	run time external declerations.	*/
void sextout()
	{
	/* Short integer run time externals. */
	static char sextrns[][59] = {
	"EXTRN\t?gcf,?gc,?sxt,?gcs,?g,?gf,?gs,?p,?sneg,?smult,?umult",
	"EXTRN\t?sw,?enr,?en,?ensr,?ens,?exr,?exrs,?exs,?sa,?scom",
	"EXTRN\t?sn,?se,?sne,?slt,?ult,?sle,?ule,?sgt,?ugt,?sge,?uge",
	"EXTRN\t?sdiv,?s1div,?udiv,?u1div,?srem,?s1rem,?urem,?u1rem",
	"EXTRN\t?ss,?s1s,?ssr,?s1sr,?usr,?u1sr,?ssl,?s1sl,?so,?sx"
	};
	ol(sextrns[0]);
	ol(sextrns[1]);
	ol(sextrns[2]);
	ol(sextrns[3]);
	ol(sextrns[4]);
	}
/* Output long integer run time	external declerations. */
void lextout()
	{
	/* Long	integer	run time externals. */
	static char lextrns[][59] = {
	"EXTRN\t?stk32,?poptos,?pshtos,?ptots,?ptotu,?ttop,?ttos",
	"EXTRN\t?gl,?g32bit,?g32c,?g32s,?g32u,?gls,?lrem,?l1rem",
	"EXTRN\t?ln,?le,?lne,?llt,?lle,?lgt,?lge,?ldiv,?l1div,?ptotl",
	"EXTRN\t?ls,?l1s,?lsr,?l1sr,?lsl,?l1sl,?lo,?lx,?la,?lcom",
	"EXTRN\t?lmult,?save32,?p32bit,?swptos,?tst32,?ladd,?lneg"
	};
	ol(lextrns[0]);
	ol(lextrns[1]);
	ol(lextrns[2]);
	ol(lextrns[3]);
	ol(lextrns[4]);
	}
/* Output floating point run time external declerations. */
void fextout()
	{
	/* Floating point run time externals. */
	static char fextrns[][50] = {
	"EXTRN\t?ltof,?ftol,?fmult,?fadd,?fs,?f1s,?fneg,?fn",
	"EXTRN\t?fe,?fne,?flt,?fle,?fgt,?fge,?fdiv,?f1div"
	};
	ol(fextrns[0]);
	ol(fextrns[1]);
	}
/* Comment line	for the	assembler */
void comment()
	{
	newline();			/* start new line if necessary */
	outbyte(';');
	}
/* "#asm" preprocessor command */
void doasm()
	{
	register c, oldtype;
	cmode =	FALSE;		/* this	is not C */
	newline();		/* start a new line if necessary */
	for (;;) {
		getline();	/* get next line */
		if (ch() == ';')
			continue; /* discard comment line */
		if (eof)
			break;
		if (ch() == '#') {
			junk();	/* skip	to next	alphanumeric */
			if (amatch("endasm"))
				break;
			else if	(amatch("include")) {
				doinclude();
				continue;
				}
			else if	(amatch("cseg")) {
				oldtype	= segtype;/* let asm command tell */
				codeseg();	/* program curr	seg type */
				segtype	= oldtype;
				continue;
				}
			else if	(amatch("dseg")) {
				oldtype	= segtype;
				dataseg();
				segtype	= oldtype;
				continue;
				}
			}
		while ((c = gch()) != '\0') {
			if (c == ';')
				break; /* strip	comments */
			outbyte(c);
			}
		nl();
		}
	kill_line();		/* clear #endasm line */
	cmode =	TRUE;		/* back	in C mode */
	if (trace.is_on)	/* if TRACE was	requested */
		retvalue = TRUE;/* force compiler to set Z flag	*/
	}
/* Assign rop to lop */
void doasgn(lop, rop, depth, keepflag)
register struct	operand	*lop, *rop;
int depth, keepflag;
	{
	long value;
	register struct	st *sym;
	register offset, sc, size;

	sym = lop->op_sym;
/* replace:  sc	= sym->st_sc; */
	sc = sym==NULL ? SC_NONE : sym->st_sc;
	size = lop->op_type->t_size;
	if (sym	!= NULL) {			/* not loaded? */
		offset = sym->st_info +	(int) lop->op_val;
		if ((sc	== SC_AUTO	/* on stack & */
		     &&	offset >= (128 - size))	/* can't use xreg */
		   || lop->op_load&LOADADDR) {	/* or is address? */
			savepreg();
			getaddr(lop);
			}
		}
	heir2(rop, KEEP);		/* same	level for right-to-left	eval */
	curtyp = lop->op_type;
	if (depth == 1			/* if assignments are not nested */
	   && isconstant(rop)		/* and we're assigning a constant */
	   && curtyp->t_code !=	T_FLOAT) { /* and float	conversion not needed*/
		value =	rop->op_val;
		if (sym	== NULL) { /* address of lop is	in preg	*/
			asgindirect(value, size);
			lop->op_val = value;
			lop->op_load = (ASGCONST|LOADVALUE);
			return;
			}
		else if	(sc == SC_AUTO && offset < (128	- size)) {
			if (size > 2) {
				getaddr(lop);
				asgindirect(value, size);
				}
			else do	{
				ot("LD\t");
				xreg(offset++);
				outbyte(',');
				outdec((int) value & 0xFF);
				nl();
				value >>= 8;
				} while	(--size	> 0);
			lop->op_val = value;
			lop->op_load = (ASGCONST|LOADVALUE);
			return;
			}
		else if	((sc ==	SC_GLOBAL || sc	== SC_ST_GLB
		     ||	sc == SC_EXTERN	|| sc == SC_STATIC) && size == 4) {
			offset = lop->op_val;
			lop->op_val = value;
			lsimmed((int) value, PREG);
			ot("LD\t(");
			prtsymname(sym,	offset);
			outstr("),HL");
			prtname(sym);
			lsimmed((int) (value >>	16), PREG);
			ot("LD\t(");
			prtsymname(sym,	offset+2);
			outstr("),HL\n");
			lop->op_load = (ASGCONST|LOADVALUE);
			return;
			}
		}
	rvalp(rop);			/* normal assign */
	store(lop, keepflag);
	}
/* Assign a constant value to the address in the primary register */
void asgindirect(value,	size)
long value;
int size;
	{
	pregflag = FALSE;		/* preg	no longer in use after this */
	for(;;)	{
		if (value == 0 && size > 1)
			break;
		ldmemi((int) value & 0xFF);
		if (--size <= 0)
			return;
		value >>= 8;
		cinc();
		}
	ol("XOR\tA");			/* generate a zero value */
	for(;;)	{
		ol("LD\t(HL),A");
		if (--size <= 0)
			return;
		cinc();
		}
	}
/* Load	an auto	variable */
void autoload(op)
struct operand *op;
	{
	register struct	st *sym;
	register offset, size;

	size = op->op_type->t_size;
	sym = op->op_sym;
	offset = sym->st_info +	(int) op->op_val;
	if (offset > 0 && offset < 127 && op->op_type->t_code != T_FLOAT
			&& !is32bit(curtyp)) {
		if (size == 1) {
			ot("LD\tA,");
			xreg(offset);
			prtname(sym);	/* print name as comment */
			callib("sxt");
			}
		else {
			ot("LD\tL,");
			xreg(offset);
			prtname(sym);
			ot("LD\tH,");
			xreg(offset+1);
			nl();
			}
		pregflag = TRUE;
		}
	else {
		getaddr(op);
		indirect(op);	/* load	indirect */
		}
	}
/* Load	the address of op in preg */
void getaddr(op)
register struct	operand	*op;
	{
	register struct	st *sym;
	register offset;

	sym = op->op_sym;
	if (sym	== NULL)
		return;			/* address already loaded */
	offset = op->op_val;		/* constant offset from	sym */
	if (is32bit(op->op_type))
		savepreg();
	if (sym->st_sc == SC_AUTO) {
		offset += sym->st_info;	/* offset of sym from stk frame	*/
		if (offset == 0) {
			dumpeep(); /* clear any	pending	pattern	*/
			loadpeep("\tLD\tL,C\n\tLD\tH,B\n");
			peepflag = ADDRSFP;
			if (is32bit(op->op_type))
				dumpeep();
			}
		else {
			lsimmed(offset,	PREG);
			ot("ADD\tHL,BC");
			prtname(sym);
			}
		}
	else {				/* address of global or	static */
		pimmed();
		prtsymname(sym,	offset);
		prtname(sym);
		}
	op->op_sym = NULL;		/* record that preg has	address	*/
	op->op_load &= ~LOADADDR;
	pregflag = TRUE;
	}
/* Load	global/static/register variable	into preg */
void qgetmem(op)
register struct	operand	*op;
	{
	register offset;
	register struct	st *sym;
	if (!is32bit(curtyp) &&	op->op_type->t_code != T_FLOAT)	{
		offset = op->op_val;
		sym = op->op_sym;
		if (op->op_type->t_size	== 1) {	/* load	1 byte */
			ot("LD\tA,(");
			prtsymname(sym,	offset);
			outbyte(')');
			prtname(sym);
			callib("sxt");
			}
		else {		/* either 2 byte variable or low order part
					of long	integer	*/
			ot("LD\tHL,(");
			prtsymname(sym,	offset);
			outbyte(')');
			prtname(sym);
			}
		pregflag = TRUE;
		}
	else {
		getaddr(op);
		indirect(op);
		}
	}
/* Store specified size	from preg to address on	top of stack */
void putstk(op,	keepflag)
register struct	operand	*op;
register int keepflag;
	{
	if (op->op_type->t_code	!= curtyp->t_code)
		cnvrtlt(op, FRM_CUR);
	if (is32bit(op->op_type)) {
		if (pregflag) {		/* 4 byte variable */
			swapstk();
			pregflag = FALSE;
			}
		else
			cpop(PREG);
		put32(keepflag);
		}
	else {
		cpop(SREG);
		if (op->op_type->t_size	== 1) {
			ol("LD\tA,L");	/* 1 byte variable */
			ol("LD\t(DE),A");
			testflag = REGL;
			}
		else {			/* all other types are 2 bytes */
			callib("p");
			testflag = REGH;
			}
		pregflag = keepflag;
		}
	}
/* Store preg in the specified memory location */
void putmem(op,	keepflag)
register struct	operand	*op;
int keepflag;
	{
	register struct	st *sym;
	register offset, size;

	if (op->op_type->t_code	!= curtyp->t_code)
		cnvrtlt(op, FRM_CUR);
	if (is32bit(op->op_type)) {
		getaddr(op);
		put32(keepflag);
		}
	else {
		offset = op->op_val;		/* constant offset from	sym */
		size = op->op_type->t_size;
		sym = op->op_sym;
		if (sym->st_sc == SC_AUTO) {
			ot("LD\t");
			offset += sym->st_info;	/* offset of sym from stk frame	*/
			xreg(offset);
			outstr(",L");
			prtname(sym);
			if (size == 2) {
				ot("LD\t");
				xreg(offset+1);
				outstr(",H\n");
				}
			}
		else {
			if (size == 1) {
				ol("LD\tA,L");
				ot("LD\t(");
				prtsymname(sym,	offset);
				outstr("),A");
				testflag = REGL;
				}
			else {
				ot("LD\t(");
				prtsymname(sym,	offset);
				outstr("),HL");
				}
			prtname(sym);
			}
		pregflag = keepflag;
		peepflag = PREGSYM;	/* remember the	variable in preg */
		strcpy(peepsym,	sym->st_name);
		peepoffset = op->op_val;
		peeplen	= size;
		}
	}
/* Store 32 bit	operand	from tos */
void put32(keepflag)
int keepflag;
	{
	callib(keepflag? "save32": "p32bit");
	tosflag	= keepflag;
	pregflag = FALSE;
	}
/* Print external or internal symbol name */
void prtsymname(sym, offset)
register struct	st *sym;
int offset;
	{
	char *regname();

	switch (sym->st_sc) {
	case SC_GLOBAL:
	case SC_EXTERN:
	case SC_ST_GLB:
		outname(sym->st_name);
		break;
	case SC_REG:
		outname(regname(sym->st_info));
		break;
	case SC_STATIC:
		printlabel(sym->st_info);
		break;
		}
	if (offset) {
		if (offset > 0)
			outbyte('+');
		outdec(offset);
		}
	}
/* Print symbol	name as	comment	*/
void prtname(sym)
struct st *sym;
	{
	if (fullist) {
		ot(";");	/* prt extra .asm info */
		outstr(sym->st_name);
		}
	nl();
	}
/* Print index register	plus offset as address */
void xreg(offset)
int offset;
	{
	outstr("(IX+");
	outdec(offset);
	outstr(")");
	}
/* Return the name of register n */
char *
regname(n)
int n;
	{
	static char name[4] = {'r', ASMRTS, 'n', '\0'};
	name[2]	= n + '0';
	return name;
	}
/* Swap	preg and sreg */
void swap_reg()
	{
	if (is32bit(curtyp))
		callib("swptos");
	else
		ol("EX\tDE,HL");
	}
/* Load	immediate into preg */
void pimmed()
	{ot("LD\tHL,");}
/* Load	immediate into sreg */
void simmed()
	{ot("LD\tDE,");}
/* Load	immediate into memory byte specified by	preg */
void ldmemi(value)
int value;
	{
	ot("LD\t(HL),");
	outdec(value);
	nl();
	}
void set1()
	{
	ol("LD\tHL,1");
	}
/* Push	reg onto stack */
void cpush(reg)
int reg;
	{
	if (peepflag ==	ADDRSFP	&& reg == PREG)	{ /* is	addr stk frame ptr? */
		clearpeep();
		ol("PUSH\tBC");		/* replacement pattern */
		peepflag = ADDRSFP;
		}
	else {
		ot("PUSH");
		ol(z80reg[reg -	PREG]);
		}
	}
/* Pop top of the stack	into reg */
void cpop(reg)
int reg;
	{
	ot("POP");
	ol(z80reg[reg -	PREG]);
	}
/* Restore stack frame at end of function */
void restsfp()
	{
	cpop(SFP);
	cpush(SFP);
	cpop(XREG);
	}
/* Swap	preg and top of	the stack */
void swapstk()
	{
	ol("EX\t(SP),HL");
	}
/* Call	the specified library routine */
void callib(sname)
char *sname;
	{
	static char libname[9] = {ASMRTS};
	struct st *findloc();

	strcpy(libname+1, sname);
	ot("CALL");
	ol(libname);
	if (libflag && !findloc(libname)) /* record this use */
		addloc(libname,	ifunctype, SC_EXTERN, ID_VAR, 0);
	}
/* Call	the specified subroutine name */
void ccall(sname)
char *sname;
	{
	ot("CALL\t");
	outname(sname);
	nl();
	}
/* Perform subroutine call to addres at	stack+offset */
void calladdr(offset)
int offset;
	{
	int label;

	label =	getlabel();
	pimmed();		/* load	address	for return */
	printlabel(label);
	nl();
	cpush(PREG);		/* put it on stack */
	lsimmed(offset,	PREG);	/* load	location of address to call */
	ol("ADD\tHL,SP");
	callib("g");		/* get address to call */
	ol("JP\t(HL)");		/* jump	to call	address	*/
	linelabel(label);	/* where we'll return to */
	}
/* Jump	to specified internal label number */
void jump(label)
int label;
	{
	switch (peepflag) {	/* check for peephole optimization */
	case JUMP:		/* last	line also JMP, do nothing */
		break;
	case JUMPTRUE:		/* 2nd part of	JNZ/JZ	jcondlabel */
	case JUMPFALSE:		/*		JMP	jumplabel */
				/* jcondlabel:			  */
		if (jumplabel != 0)	/* if last line	also a JMP */
			break;		/* skip	this one */
		loadpeep("\tJP\t");	/* save	this part of pattern */
		peeplabel(jumplabel = label);
		break;
	default:
		dumpeep();	/* dump	some other pending pattern */
	case FALSE:
		ot("JP\t");
		printlabel(label);
		nl();
		peepflag = JUMP; /* remember this is start of pattern: */
		break;		/*	JMP	label1		*/
		}		/*	JMP	label2		*/
	}
/* Remember that this is start of a pattern consisting of
 *		JP Z/NZ	jcondlabel
 *		JP	jumplabel
 * jcondlabel:
 */
void jumpcond(type, label)
int type, label;
	{
	dumpeep();		/* dump	any pending pattern */
	if(!testdone)		/* did we just test (e.g. $eq)?	*/
		dotest();	/* if not, generate a test */
	if (type == FALSE) {
		peepflag = JUMPFALSE;
		loadpeep("\tJP\tZ,");
		}
	else {
		peepflag = JUMPTRUE;
		loadpeep("\tJP\tNZ,");
		}
	peeplabel(jcondlabel = label);
	jumplabel = 0;	/* will	be set if 2nd part of pattern found */
	}
/* Generate a test of the preg */
void dotest()
	{
	if (is32bit(curtyp)) {
		savepreg();
		callib("tst32");	/* test	TOS and	places result in PREG and Z flag */
		pregflag = TRUE;
		tosflag	= FALSE;
		curtyp = sinttype;
		}
	else {
		switch (testflag) {	/* see what we need to do test */
			case REGL:
				ol("OR\tH");	/* L already in	A */
				return;
			default:		/* need	full test */
				ol("LD\tA,H");
			case REGH:		/* H already in	A */
				ol("OR\tL");
			case TESTED:		/* zero/non-zero test already done */
				return;
			}
		}
	}
/* Print a line	label */
void linelabel(label)
int label;
	{
	char *jump;
	if ((peepflag == JUMPTRUE || peepflag == JUMPFALSE) /* 1st part? */
	     &&	jumplabel != 0		/* 2nd part of pattern found? */
	     &&	label == jcondlabel) {	/* now check for final part */
		jump = (peepflag == JUMPTRUE) ?	"JP\tZ,": "JP\tNZ,";
		clearpeep();		/* clear pattern buffer	*/
		ot(jump);
		printlabel(jumplabel);
		nl();
		return;
		}
	newline();			/* start new line if necessary */
	printlabel(label);
	outbyte(':');
	}
/* Print a global label	*/
void glblabel(name, sc)
char *name;
int sc;
	{
	if (sc == SC_GLOBAL) {
		ot("PUBLIC\t");
		outname(name);
		nl();
		}
	outname(name);
	outbyte(':');
	}
/* Data	segment	*/
void dataseg()
	{
	if (segtype != DATA) {
		newline();	/* start new line if necessary */
		ol("DSEG");
		segtype	= DATA;
		}
	}
/* Code	segment	*/
void codeseg()
	{
	if (segtype != CODE) {
		newline();
		ol("CSEG");
		segtype	= CODE;
		}
	}
/* Declare external */
void declext(name)
char *name;
	{
	ot("EXTRN\t");
	outname(name);
	nl();
	}
/* Default initialization of a global or static	variable */
int definit(amt)
register amt;
	{
	if (amt	<= 0)
		return 0;
	newline();		/* start new line */
	switch (amt) {
	default:	ot("REPT\t");
		outdec(amt);
		nl();
	case 1:	defbyte();
		goto value;
	case 2:	defword();
	value:	outstr("0\n");
	if (amt	> 2)
		ol("ENDM");
		}
	return amt;
	}
/* pseudo-op to	define a byte */
void defbyte()
	{ot("DEFB\t");}
/* pseudo-op to	define storage */
void defstorage()
	{ot("DEFS\t");}
/* pseudo-op to	define a word */
void defword()
	{ot("DEFW\t");}
/* clear amount	requested off the stack	*/
void clearstk(amt)
int amt;
	{
	switch (amt) {
	case 8:	cpop(WREG);	/* special case	for 8 bytes or less */
	case 6:	cpop(WREG);
	case 4:	cpop(WREG);
	case 2:	cpop(WREG);
	case 0:	return;
	default:
		curtyp = sinttype;
		swap_reg();	/* save	possible return	value */
		lsimmed(amt, PREG);
		ol("ADD\tHL,SP");
		ol("LD\tSP,HL");
		swap_reg();	/* restore return value	*/
		return;
		}
	}
/* Double the preg */
void doublereg()
	{
	ol("ADD\tHL,HL");
	}
/* Add preg and	sreg */
void cadd()
	{
	if (is32bit(curtyp)) {
		if (curtyp->t_code == T_LONG)
			callib("ladd");
		else
			callib("fadd");
		}
	else
		ol("ADD\tHL,DE");
	}
/* Increment preg */
void cinc()
	{
	ol("INC\tHL");
	}
/* Decrement preg */
void cdec()
	{
	ol("DEC\tHL");
	}
/* Return from subroutine */
void cret()
	{
	ol("RET");
	}
/* end of QC7.C	*/

