/********************************************************/
/*							*/
/*		 Q/C Compiler (Part 8)			*/
/*							*/
/*     Copyright (c) 1985 Quality Computer Systems	*/
/*		 All Rights Reserved			*/
/*							*/
/*			09/05/85			*/
/********************************************************/

#include <stdio.h>
#include <ctype.h>
#include "qc.h"
#include "ctp.h"
#include "qcdecl.c"

/* This	section	handles	I/O, preprocessor commands and error messages */

static char *mptr;

static char errstrs[][60] =
  {
    "V4.0a",
    "401 Unbalanced braces",
    "202 Missing #endif",
    "403 Negative array	size illegal",
    "404 Member	can't have storage class",
    "405 Member	has another meaning",
    "406 Can't be a member",
    "407 Can only initialize global and	static variables",
    "508 Can't initialize float	variables",
    "409 Can't initialize unions",
    "410 Only aggregates can be	initialized this way",
    "411 Too many initializers",
    "412 String	is bigger than array",
    "413 Char cannot hold address",
    "414 Expected comma",
    "415 Argument can't	be that	type",
    "416 Name not in argument list",
    "417 Undefined label",
    "418 else not matched with if",
    "419 No active switch statement",
    "420 Missing semicolon",
    "421 Missing while in dowhile",
    "422 No active switches or loops",
    "423 No active loop	statement",
    "424 Switch	expression is not int",
    "425 Multiple default cases",
    "426 Inconsistent use of pointers in conditional expression",
    "427 Not a pointer",
    "428 Illegal address",
    "429 Not a function",
    "430 Not a structure or union member",
    "431 Can't subscript",
    "432 Illegal use of	label",
    "433 Undefined variable",
    "434 Invalid expression",
    "435 Must be lvalue",
    "436 Function can't	return aggregate",
    "437 No arrays of functions",
    "638 This type is too ornate",
    "739 Too many active switches or loops",
    "840 No active switches or loops to	delete",
    "441 Duplicate switch case",
    "742 Too many switch cases",
    "843 No entry in case table	to delete",
    "744 Macro (#define) pool is full",
    "245 Can't #undef -	not defined",
    "446 Invalid float constant",
    "447 Invalid hex constant",
    "448 Invalid octal constant",
    "749 String	space full",
    "450 Inconsistent declaration",
    "751 Symbol	table full",
    "452 No template",
    "753 Member	table full",
    "754 Too many different types in use",
    "455 Size unknown",
    "456 Can't add pointers",
    "457 Can't subtract	pointer	from scalar",
    "458 Can't subtract	unlike pointers",
    "459 Can't pass structures or unions",
    "760 Argument work space full",
    "461 Invalid operation on float",
    "362 Can't close output file",
    "263 Missing delimiter",
    "564 #include nested too deeply",
    "265 Only one #else	allowed",
    "266 #line number must be decimal",
    "267 Unknown #preprocessor command",
    "568 #if nested too	deeply",
    "469 Unclosed comment",
    "470 Missing quote",
    "471 Missing apostrophe",
    "272 Line too long",
    "273 Must match a #if command",
    "474 Invalid storage class",
    "475 Invalid type",
    "476 Illegal symbol	name",
    "477 Already defined",
    "478 Undeclared tag",
    "479 Missing punctuation --	assumed	present",
    "480 Illegal use of	a keyword",
    "481 Need explicit array size",
    "482 Initializer must be constant expression",
    "483 Must be constant expression",
    "184 Out of	memory",
    "385 Can't open input file",
    "386 Can't open output file",
    "387 Can't open include file",
    "188 Error message file not	found",
    "189 Error message file is wrong version",
    "190 Usage:qc infile ... -cdilortv -sxxxx outfile",
    "191 Not enough table space"
  };



/* Get the next	input file */
void getinfil()
	{
	char *strcpy();
	if (incldepth >	0) {
		input =	holdinput[--incldepth];	/* back	to interrupted */
		strcpy(infil, inclfil[incldepth]); /* input file */
		lineno = holdlineno[incldepth];
		if (verbose)
			printf("*** Resume  %s\n", infil);
		}
	else if	(ninfils--) {		/* any more input files? */
		getfilename(*nextfil++,	infil, "c");
		if ((input = fopena(infil,"r"))	== NULL)
			/*Error	#385 Can't open	input file */
			io_err(385, infil);
		lineno = 0;
		if (verbose) printf("*** Read	 %s\n",	infil);
		}
	else
		eof = TRUE;
	}
/* Open	code output file */
void getoutfil()
	{
	if (coutput == stderr)		/* output to console */
		return;
	if ((coutput = fopena(coutfil, "w")) ==	NULL)
		/*Error	#386 Can't open	output file */
		io_err(386, coutfil);
	output = coutput;
	segtype	= CODE;
	lngused	= FALSE;
	fltused	= FALSE;
	header();
	}
/* Close the output files */
void closeout()
	{
	trailer();
	if (output != stderr) {
		if (fclose(coutput) == EOF)
			out_err();
		}
	}
/* Get file name, optionally append extension */
void getfilename(s, name, ext)
char *s, *name,	*ext;
	{
	register i;
	register char *p;
	char *strchr();

	p = name;
	i = 0;
	while (*s && ++i <= 76)	{ /* max. length of filename without ext. */
		*p++ = *s++;
		}
	*p = '\0';
	if (strchr(name, '.') == NULL && ++i < 76) {
		*p++ =	'.';		      /* no extension given */
		while (*ext && ++i <= 79) {	/* use .EXT */
			*p++ = *ext++;
			}
		*p = '\0';
		}
	}
/* Open	an include file	*/
void doinclude()
	{
	register char *p;
	register c, delimiter;
	char *strcpy();

	if (matchc('"'))
		delimiter = '"';
	else if	(matchc('<'))
		delimiter = '>';
	else
		delimiter = 0;
	p = lptr;			/* beginning of	file name */
	if (delimiter) {
		while ((c=ch())!=delimiter && c!='\0' && !isspace(c))
			gch();
		if (c != delimiter)
			needpunc(delimiter);
		*lptr =	'\0';		/* end file name */
		}
	else
		/*Error	#263 Missing delimiter */
		errname(263, "\" or <");
	if (incldepth == MAXINCL)
		/*Error	#564 #include nested too deeply	*/
		error(564);
	else {
		holdinput[incldepth] = input;	/* save	old FILE ptr */
		strcpy(inclfil[incldepth], infil); /* and name */
		holdlineno[incldepth] =	lineno;	/* save	old line num */
		getfilename(p, infil, "C");	/* get include filename	*/
		if ((input=fopena(infil, "r")) == NULL)
			/*Error	#387 Can't open	include	file */
			io_err(387, infil);
		lineno = 0;
		++incldepth;
		if (verbose)
			printf("*** Include %s\n", infil);
		}
	}
/* Get the next	line of	input */
void getline()
	{
	register c;

	chkabort();		/* did user type ^C? */
	for (;;) {		/* find	a non-empty line or EOF	*/
		if (eof)
			return;
		kill_line();		/* clear the input line	*/
		++lineno;
		while ((c = getc(input)) != EOF) {
			if (c == '\n' || lptr-line >= LINEMAX)
				break;
			*lptr++	= c;
			}
		*lptr =	'\0';
		if (lptr != line) {	/* anything on this line? */
			if (fullist && cmode) {
				textline = TRUE;
				comment(); /* print C text */
				outstr(line);
				nl();
				textline = FALSE;
				}
			lptr = line;	/* reset line to beginning */
			return;
			}
		else if	(c == EOF) {
			fclose(input);
			getinfil();	/* get the next	input file */
			}
		}
	}
/* Write a character to	the requested output location */
void outbyte(c)
register c;
	{
	testflag =
	testdone = FALSE;	/* turn	off when anything written */
	if (peepflag &&	!textline) /* is a peephole pattern pending? */
		dumpeep();	/* dump	it out,	this is	not part of it */
	if (genflag == GENCODE || textline) {
		midline	= (c !=	'\n');
		if (putc(c, output) == EOF)
			out_err();
		}
	else if	(genflag == GENARG) {
		midline	= (c !=	'\n');
		loadarg(c);
		}
	}
/* Put a string	out via	outbyte	*/
void outstr(p)
char *p;
	{
	chkabort();		/* did user type ^C? */
	while (*p)
		outbyte(*p++);
	}
/* If a	^C was typed abort the run */
int chkabort()
	{
	static char mesg[] = "\nCompilation aborted";
	int c;
	if ((c = inkey()) == 0x3) {
		outstr(mesg);
		puts(mesg);
		exit(1);
		}
	return c;
	}
/* Start a new line if we're not at the	beginning */
void newline()
	{
	if (midline)
		nl();
	}
/* Print a string with a leading TAB */
void ot(s)
char *s;
	{
	outbyte('\t');
	outstr(s);
	}
/* Print a line	with a leading TAB */
void ol(s)
char *s;
	{
	ot(s);
	nl();
	}
void nl()
	{
	outbyte('\n');
	}
/* Test	parse line to see if lit matches */
/*	the next sequence of characters	*/
/*	and skip over them if it does	*/
int match(lit)
char *lit;
	{
	register k;
	blanks();	/* skip	white space */
	if ((k = streq(lptr, lit)) != 0) {
		lptr +=	k;
		return TRUE;
		}
	return FALSE;
	}
/* Test	parse line to see if c matches */
/*	the next character and skip it */
int matchc(c)
int c;
	{
	blanks();	/* skip	white space */
	if (*lptr == c)	{
		++lptr;
		return TRUE;
		}
	return FALSE;
	}
/* Test	parse line to see if 'c' is the	next char */
int nextc(c)
int c;
	{
	blanks();
	return (*lptr == c);
	}
/* Test	parse line to see if lit matches */
/*	the next token and skip	over it	*/
/*	if a match is found		*/
int amatch(lit)
char *lit;
	{
	blanks();
	if (chks(lit)) {
		skip();
		return TRUE;
		}
	return FALSE;
	}
/* Skip	white space in parse line */
void blanks()
	{
	register c;
	for (;;) {
		while ((c = ch()) == '\0') {	/* at eol? */
			if (stopeol)		/* OK to pass eol? */
				return;		/* no */
			getline();		 /* get	next line */
			if (eof)
				return;
			preprocess();
			}
		if (isspace(c))
			gch();
		else
			return;
		}
	}

/* Throw away garbage in input line */
void junk()
	{
	while (!issym(ch())) {
		gch();			/* flush non-a/n characters */
		if (endst())		/*	up to EOL or ';' */
			break;
		}
	}
/* Test	for end	of statement */
int endst()
	{
	blanks();	/* skip	white space */
	return (ch() ==	';' || ch() == '\0');
	}
/* Look	at current character in	parsing	buffer */
int ch()
	{
	return (int) *lptr;
	}
/* Look	at next	character in parsing buffer */
int nch()
	{
	return (*lptr ?	*(lptr+1) : 0);
	}
/* Get current character in parsing buffer */
int gch()
	{
	return (*lptr ?	*lptr++	: 0);
	}
/* Put a character back	in the parsing buffer */
void putback(c)
int c;
	{
	if (lptr > line)
		*--lptr	= c;
	}
/* Clear current line from parsing buffer */
void kill_line()
	{
	lptr = line;
	*lptr =	'\0';
	}
/* Output an integer in	decimal	*/
void outdec(n)
int n;
	{
	char *itob(), s[7];
	outstr(itob(n, s, -10));
	}
/* Process special comment commmands */
void chkcommand()
	{
	register on_off;
	register char *option;

	if (matchc('$')) {
	    if ((on_off	= matchc('+')) || matchc('-')) {
		switch (toupper(ch())) {
		case 'T':
			if (trace.enabled) {
				trace.is_on = on_off;
				option = "Trace	 ";
				break;
				}
			else
				return;
		default:
			return;
			}
		if (verbose)
			printf("*** %s %s\n",option,on_off?"ON":"OFF");
		}
	    }
	}
/* Check for #preprocessor commands */
void preprocess()
	{
	char name[NAMESIZE];
	struct mp_hash *isdef();
	struct operand cexp;

	if (!matchc('#')) {		/* if it's not a #command... */
		if (condif == PROCESS) {/* check conditional compilation */
			lptr = line;	/* start at beginning of line */
			procline();	/* preprocess the input	line */
			}
		else
			kill_line();	/* skip	this line */
		}
	else {				/* do a	#preprocessor command */
		stopeol	= TRUE;		/* look	only on	this line */
		/* #if,	#else and #endif are only commands recognized */
		/*	when C text is being skipped because of	#if */
		if (amatch("ifdef"))
			dopreif((int) isdef());
		else if	(amatch("ifndef"))
			dopreif(!isdef());
		else if	(amatch("if")) {
			procline();	/* preprocess the const	expr */
			if (constexp(&cexp) != CONSTANT) {
				cexperr();
				cexp.op_val = 0;
				}
			dopreif((int) cexp.op_val);
			}
		else if	(amatch("else")) {
			if (ifdepth == 0)
				matcherr(); /* no matching #if... */
			else if	(ifdepth > MAXIF)
				;	/* nested too deep-can't check */
			else if	(condelse)
				/*Error	#265 Only one #else allowed */
				error(265);
			else {
				condelse = TRUE; /* #else found	*/
				if (condif != IGNORE) /* turn it around	*/
					condif = (condif==SKIP)?
							PROCESS: SKIP;
				}
			}
		else if	(amatch("endif")) {
			if (ifdepth == 0)
				matcherr(); /* no matching #if... */
			else {
				--ifdepth; /* back up one level	*/
				if (ifdepth < MAXIF) { /* get status */
					condif = holdif[ifdepth];
					condelse = holdelse[ifdepth];
					}
				}
			}
		else if	(condif	!= PROCESS) /* chk #if condition */
			;		/* skip	this line */
		else if	(amatch("define")) {
			if (symname(name)) { /*	valid name? */
				procline();  /*	strip comments & apply */
				addmac(name);/*	prev defs recursively */
				}
			else
				kill_line(); /*	kill rest of line */
			}
		else if	(amatch("undef"))
			delmac();
		else if	(amatch("include"))
			doinclude();	/* get include file */
		else if	(amatch("asm"))
			doasm();	/* copy	assembler thru */
		else if	(amatch("line")) {
			procline();	/* strip comments */
			blanks();
			if (!isdigit(ch()))	/* new line number? */
				/*Error	#266 #line number must be decimal */
				error(266);
			else {
				lineno = 0;
				while (isdigit(ch()))
					lineno = 10 * lineno + gch() - '0';
				--lineno;	/* adjust for next line	*/
				if (!endst())	/* optional filename? */
					getfilename(lptr, infil, "C");
				}
			}
		else
			/*Error	#267 Unknown #preprocessor command */
			error(267);
		stopeol	= FALSE;	/* don't stop at end-of-line */
		kill_line();
		}
	}
/* Common code for processing #if preprocessor commands	*/
void dopreif(newstatus)	/* zero	means SKIP; non-zero means PROCESS */
int newstatus;
	{
	if (ifdepth >= MAXIF)
		/*Error	#568 #if nested	too deeply */
		error(568);
	else {
		holdif[ifdepth]	= condif; /* record previous status */
		holdelse[ifdepth] = condelse;
		if (condif == PROCESS)	/* currently processing? */
			condif = (newstatus) ? PROCESS:	SKIP;
		else
			condif = IGNORE;
		condelse = FALSE;	/* no #else yet	*/
		}
	++ifdepth;			/* keep	track even if too deep */
	}
/* Preprocess input lines by removing
 * comments and	excess white space,
 * checking strings and	char constants,
 * & applying #define macros to	the rest
 */
void procline()
	{
	register char *ptr, *savelptr;
	register c;
	register struct	mp_hash	*mp;
	struct mp_hash *findmac();
	char *chksym(),	name[NAMESIZE],	*strcpy();

	mptr = mline;
	savelptr = lptr;	/* remember where we are on this line */
	for (ptr = line; ptr < lptr; ++ptr)
		keepch(*ptr);	/* keep	anything already processed */
	while ((c = ch()) != 0)	{
		if (c == ' ' ||	c == '\t') {	/* eliminate excess */
			keepch(' ');		/* white space */
			while (ch() == ' '|| ch() == '\t')
				gch();
			}
		else if	(chksym(name)) {	/* is it a symbol? */
			if ((mp	= findmac(name)) != NULL) { /* if it's a */
				skip();		/* macro, skip over it and */
				ptr = mp->def;	/* replace it */
				while (*ptr)
					keepch(*ptr++);
				}
			else while (issym(ch()))
				keepch(gch());	/* keep	original text */
			}
		else if	(c == '/' && *(lptr+1) == '*') {
			gch();gch();		/* discard comments */
			stopeol	= TRUE;		/* look	for special */
			chkcommand();		/* command on current */
			stopeol	= FALSE;	/* line	only */
			while (*lptr !=	'*' || *(lptr+1) != '/') {
				if (gch()==0)	/* if EOL get */
					getline(); /* next line	*/
				if (eof) {
					/*Error	#469 Unclosed comment */
					error(469);
					break;
					}
				}
			gch();gch();		/* discard end of comment */
			}
		else if	(c == '"') {
			keepch(gch());		/* pass	strings	through	*/
			while (ch()!='"') {	/* unchanged */
				if (ch()==0) {
					/*Error	#470 Missing quote */
					error(470);
					break;
					}
				if (ch()=='\\')	{ /* pass an escape char */
					if (nch())	/* not EOL */
						keepch(gch());
					else {	/* continue long literal */
						getline();
						continue;
						}
					}
				keepch(gch());
				}
			gch();
			keepch('"');
			}
		else if	(c == '\'') {
			keepch(gch());	/* pass	char constants */
			while (ch()!='\'') {	/* unchanged */
				if (ch() == 0) {
					/*Error	#471 Missing apostrophe	*/
					error(471);
					break;
					}
				if (ch() == '\\') /* pass an escape char */
					keepch(gch());
				keepch(gch());
				}
			gch();
			keepch('\'');
			}
		else keepch(gch());	/* pass	all else unchanged */
		}
	*mptr =	'\0';
	if (mptr-mline>=LINEMAX)
		/*Error	#272 Line too long */
		error(272);
	strcpy(line, mline);		/* put processed line back */
	lptr = savelptr;		/* restore original position */
	}
/* Put a char in macro buffer */
void keepch(c)
int c;
	{
	if ((mptr - mline) < LINEMAX)
		*mptr++	= c;
	}
/* Fatal error - print message and quit	with minimal additional	processing */
void fatal(msg)
char *msg;
	{
	fprintf(stderr,msg);		/* print string	*/
	fprintf(stderr,"\n");
	exit(1);			/* error exit */
	}
/* For I/O errors, print message and quit */
void io_err(errc, filename)
int errc;
char *filename;
	{
	prtmesg(errc);
	printf(": %s\n", filename);
	exit(1);
	}
/* Print an erroneous line with	a pointer (^) to	*/
/*	the location where the error was discovered	*/
/*	and a message describing the error		*/
void error(errc)
int errc;
	{
	prtfile();
	prtmesg(errc);
	printf("\n");
	errptr();
	errpause();
	}
void errname(errc, name)
int errc;
char *name;
	{
	prtfile();
	prtmesg(errc);
	printf(": %s\n", name);
	errptr();
	errpause();
	}
/* Print the name of the file which contains the error */
void prtfile()
	{
	if (incldepth)			/* are we in a #include	file? */
		printf("<%s> @ %d: #include ", inclfil[0], holdlineno[0]);
	printf("<%s> @ %d: ", infil, lineno);
	}
/* Print the error number and message */
void prtmesg(errc)
int errc;
	{
	printf("Error #%s",errstrs[errc%100]);
	}
/* Print the input line	with a pointer (^) under the error */
void errptr()
	{
	register char *p;
	int c;

	p = line;
	if (*p == '\0')
		return;			/* empty line or EOF */
	printf("%s\n", p);	       /* print	erroneous line */
	while (p < lptr) {		/* get to error	location */
		c = (*p++ == '\t') ? '\t': ' ';
		putchar(c);
		}
	puts("^");			/* print error pointer ^ */
	}
/* If screen is	full, pause for	keyboard response */
void errpause()
	{
	++errcnt;
	if ((errcnt % pausecnt == 0) ||	verbose) {
		printf("Press RETURN to	continue or ^C to abort:\n");
		kbclear();		/* clear typeahead and get response */
		while (chkabort() == EOF); /* wait for response	*/
		}
	}

/*	Common error messages		*/
void fltoperr()
	{
	/*Error	#461 Invalid operation on float	*/
	error(461);
	}
void out_err()
	{
	/*Error	#362 Can't close output	file */
	io_err(362, coutfil);
	exit(1);
	}
void matcherr()
	{
	/*Error	#273 Must match	a #if command */
	error(273);
	}
void scerr()
	{
	/*Error	#474 Invalid storage class */
	error(474);
	}
void typerr()
	{
	/*Error	#475 Invalid type */
	error(475);
	}
void illname()
	{
	/*Error	#476 Illegal symbol name */
	error(476);
	junk();
	}
void multidef(name)
char *name;
	{
	/*Error	#477 Already defined */
	errname(477, name);
	}
void tagerr(tag)
char *tag;
	{
	/*Error	#478 Undeclared	tag */
	errname(478, tag);
	}
void needpunc(c)
int c;
	{
	static char s[2];
	if (!matchc(s[0]=c))
		/*Error	#479 Missing punctuation -- assumed present */
		errname(479, s);
	}
void keyerr(keyword)
char *keyword;
	{
	/*Error	#480 Illegal use of a keyword */
	errname(480, keyword);
	}
void sizerr()
	{
	/*Error	#481 Need explicit array size */
	error(481);
	}
void initerr()
	{
	/*Error	#482 Initializer must be constant expression */
	error(482);
	}
void cexperr()
	{
	/*Error	#483 Must be constant expression */
	error(483);
	}
/* Match string	against	next token in parsing buffer */
int chks(s)
register char *s;
  {
    register int count;
    register char *lp;

    count=0;
    lp=lptr;
    while(*s) if(*s++==*lp++) count++; else return(0);
    if(issym(*lp))
      return(0);
    else
      return(count);
  }
/* Test	two strings for	equality up to the end of the second string */
int streq(s1,s2)
register char *s1, *s2;
  {
    register int count;

    count=0;
    while(*s2) if(*s1++==*s2++)	count++; else return(0);
    return(count);
  }
/* Test	two strings for	equality of contents up	to len (<256) */
int astreq(s1,s2,len)
register char *s1, *s2;
register int len;
  {
    register int count;

    count=0;
    while(*s2)
      {
	if(*s1++==*s2++)
	  {
	    count++;
	    if(!--len) break;
	  }
	else return(0);
      }
    if(issym(*s1))
      return(0);
    else
      return(count);
  }
/* Skip	token in input line */
void skip()
  {
    while (issym(*lptr)) lptr++;
  }

/* end of QC8.C	*/

