/********************************************************/
/*							*/
/*	     Q/C Compiler Global Definitions		*/
/*							*/
/*     Copyright (c) 1985 Quality Computer Systems	*/
/*		 All Rights Reserved			*/
/*							*/
/*			08/21/85			*/
/********************************************************/

#include <stdio.h>
#include "qc.h"

/********************************************************/
/*	Global variables referenced by QRESET		*/
/*							*/
/* If you change the compiler and you still want QRESET	*/
/* to work, this code must remain unchanged.		*/
/********************************************************/

char	copyright[] = "	Copyright (c) 1985 Quality Computer Systems";
char	version[] = "V4.0a";

int	nsym = 500,		/* size	of symbol table	*/
	st_size	= 397,		/* size	of symbol hash table (nsym-cellar) */
				/*  should be largest prime < .8 * st_size */
	nmem = 200,		/* size	of structure member table */
	n_types	= 200,		/* size	of type	table */
	nswq = 50,		/* max entries in switch/loop queue */
	ncases = 50,		/* max entries in switch case table */
	argwsize = 2000,	/* size	of function arg	work area */
	litpsize = 4000,	/* size	of literal pool	*/
	n_mac =	300,		/* size	of macro (#define) pool	*/
	mp_size	= 239,		/* size	of macro pool hash area	(nmac-cellar)*/
				/*  should be largest prime < .8 * mp_size */
	inbsize	= 512,		/* size	of input file buffer */
	outbsize = 512,		/* size	of output file buffer */
	pausecnt = 6,		/* pause after this many errors	on screen */
	verbose	= TRUE,		/* TRUE	to chat	during compile */
	initflag= FALSE,	/* TRUE	to initialize arrays>INITSIZE */
	redirect = FALSE;	/* TRUE	if redirection requested */
char	defext[] = "cas";

/********************************************************/
/*	End of global variables	referenced by QRESET	*/
/********************************************************/

int	segtype	= CODE,		/* record whether code or data segment */
	fullist	= FALSE,	/* commented listing? */
	mainflag = FALSE,	/* has main() been defined? */
	libflag	= FALSE,	/* is library being generated? */
	genflag	= GENCODE,	/* FALSE to suppress code generation */
	romflag	= FALSE,	/* generating ROMable code? */
	spaddr,			/* if romflag==TRUE, addr to init SP */
	cmode =	TRUE,		/* parsing C or	passing	asm thru? */
	midline	= FALSE,	/* at beginning	of output line?	*/
	pregflag = FALSE,	/* does	preg contain curr val of expr? */
	tosflag	= FALSE,	/* does	tos contain curr val of	32 bit expr? */
	retvalue = FALSE,	/* is return value in preg? */
	testdone = FALSE,	/* was a test (e.g. ==)	just done? */
	lngused	= FALSE,	/* long	decleration incountered	so include
					long run time externals	*/
	fltused	= FALSE,	/* float decleration incountered so include
					float and long run time	externals */
	eof = FALSE,		/* reached end-of-file?	*/
	stopeol	= FALSE,	/* TRUE	if parser must stop at end of */
				/*	line when skipping white space */
	infunc = FALSE,		/* compiling a function? */
	textline = FALSE;	/* printing a C	text line? */
struct _switch
	trace =	{FALSE,	FALSE};	/* print trace messages? */
/*
 *	Compiler tables	& pointers
 */
struct typeinfo
	*typelist,		/* the type table */
	*basetypes[T_MAX],	/* heads of each base-type list	*/
	*ifunctype,		/* constant 'function returning	short int' */
	*labeltype,		/* constant 'statement label' */
	*chartype,		/* constant 'char'	*/
	*sinttype,		/* constant 'int'	*/
	*unsgntype,		/* constant 'unsigned'	*/
	*linttype,		/* constant 'long' */
	*floattype,		/* constant 'float' */
	*curtyp,		/* current type	*/
	*rettype;		/* type	to be returned by current function */
/*
 *	Symbol table
 */
struct st
	*symtab,	/* beginning of	symbol table */
	*glbptr,	/* next	global entry */
	*locptr;	/* next	local entry */
struct st_hash
	*sym_hash,	/* symbol table	hash index */
	*sym_ovflow;	/* next	entry in hash overflow area */
unsigned
	minsym;		/* record minimum symbol table space */
/*
 *	Member table
 */
struct memtab
	*memtab,	/* beginning of	structure member table */
	*freemem;	/* first free entry in member table */
/*
 *	Literal	pool
 */
struct st
	lit_sym		/* special "symbol" for	literal	pool */
		= {SC_STATIC, NULL, 1, "$LITPOOL", ID_VAR};
unsigned
	minlit;		/* record minimum literal pool space */
char	*litpool,	/* beginning of	literal	pool */
	*plitpool;	/* next	char in	literal	pool */
/*
 *	Macro pool
 */

struct mp_hash
	*mac_hash,	/* beginning of	macro pool hash	index */
	*mac_ovflow;	/* next	entry in hash overflow area */
int
	n_def;		/* record number of macro definitions */

/*
 *	Switch/loop queue
 */
struct swq
	*swq,		/* start of switch/loop	queue */
	*pswq;		/* next	switch/loop entry */
/*
 *	Switch/case table
 */
struct case_table
	*casetab,	/* start of switch case	table */
	*pcasetab;	/* next	switch case entry */

char	*line,		/* parsing buffer */
	*lptr,		/* parsing line	buffer */
	*mline;		/* temp	macro buffer */
/*
 *	Misc variables
 */

char	infil[80],	/* current input file name */
	**nextfil,		/* ptr to name of next input file */
	coutfil[80],		/* code	output file name */
	inclfil[MAXINCL][80];	/* include file	names */

FILE	*input,		/* iob for current input File */
	*output,	/* iob for current output file (if any)	*/
	*coutput,	/* iob for code	output file (if	any) */
	*holdinput[MAXINCL]; /*	hold file iob during #include */

int	ninfils,	/* # of	input files */
	lineno,		/* line	number in current input	file */
	incldepth,	/* current depth of #include nesting */
	holdlineno[MAXINCL], /*	hold input file	line no	during #include	*/
	ifdepth	= 0,	/* current depth of #if	nesting	*/
	holdif[MAXIF],	/* hold	process	status of nested #if */
	holdelse[MAXIF],/* record presence of #else for	nested #if */
	condif=PROCESS,	/* records whether #if says to PROCESS or SKIP */
	condelse=FALSE,	/* records whether matching #else was found */
	locspace,	/* local space needed in current function */
	retlab,		/* label # of return for current function */
	nregvar,	/* # of	register variables in current function */
	ncmp,		/* # open compound statements */
	errcnt;		/* # errors in compilation */
/*
 *	Peephole optimization variables
 */
int	peepflag,	/* type	of pattern (if any) */
	testflag,	/* = TESTED, REGH, or REGL */
	jcondlabel,	/* label on last conditional jump */
	jumplabel;	/* label on following unconditional jump */
char	*peepbuf,	/* buffer for patterns */
	peepsym[NAMESIZE]; /* name of symbol in	preg */
int	peepoffset,	/* offset from symbol in preg */
	peeplen;	/* length of symbol in preg */

/*
 * Work	area to	reverse	function arguments
 */

char
	*argwork,	/* beginning of	work area */
	*parg;		/* next	available space	*/

/* end of QCDEF.C */

