/**************************************************************************\
*									   *
*   SYSIBM.C   -   unido CHiPSy, Systemanpassungen fuer	ATARI ST	   *
*									   *
*		   Compiler: Turbo C					   *
*									   *
*									   *
*					Michael	Roehner, unido ChiPSy	   *
*									   *
\**************************************************************************/





/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>			/* FILE, NULL, EOF, ...		  */
#include <sys/time.h>
#include <unistd.h>



/*					TurboC-spezifische Definitionen	  */
/**************************************************************************/

/* extern char	 *farmalloc();		   "far" malloc()		  */
extern long    farcoreleft();		/* "far" coreleft()		  */





/*						   Modulinterne	Globals	  */
/**************************************************************************/

static char   *memstart;		/* Beginn initialer Freispeicher  */
static long    memfree;			/* # Bytes initialer Freispeicher */
static char   *reststart;		/* Beginn Rest-Freispeicher	  */
static long    restfree;		/* # Bytes Rest-Freispeicher	  */





/**************************************************************************\
*									   *
* action      :	 Allokieren eines zusammenhaengenden Speicherblocks.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 size	-  Anzahl der zu allokierenden Bytes		   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Zeiger	auf den	allokierten Speicherbereich oder NULL,	   *
*		 wenn das Allokieren fehlschlug				   *
*									   *
\**************************************************************************/

/* farmalloc() in TurboC vorhanden */





/**************************************************************************\
*									   *
* action      :	 Groesse des groessten durch farmalloc() zu allokierenden  *
*		 zusammenhaengenden Speicherbereiches.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Anzahl	Bytes des groessten Speicherbereiches (s.o.)	   *
*									   *
\**************************************************************************/

/* farcoreleft() in TurboC vorhanden */





/**************************************************************************\
*									   *
* action      :	 Initialisieren	des eigenen Memory-Allocators clrsbrk().   *
*									   *
*		 Programmabbruch mit Meldung, wenn nicht genug		   *
*		 Freispeicher vom Betriebssystem angefordert werden kann.  *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 memstart   -  Beginn initialer	Freispeicher		   *
*		 memfree    -  Anzahl Bytes Laenge initialer Freispeicher  *
*		 reststart  -  Zeiger auf Anfang Restfreispeicher	   *
*		 restfree   -  Anzahl Bytes noch frei im Restfreispeicher  *
*									   *
* locals      :	 n	    -  Laufvariable Freispeicher Loeschen	   *
*		 p	    -  Zeiger in Freispeicher fuer Loeschen	   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void initsbrk()
  {
    register long    n;
    register char   *p;

    n =	restfree = memfree = 50000L;
    reststart =	memstart = malloc(memfree);
    if (memstart == NULL)
      {
	printf("Fatal: cannot allocate sbrk space\n");
	printf("(request: %ld)\n",memfree);
	exit(1);
      }
    else
      for (p = memstart; n > 0L; --n); *p++=0;
  }





/**************************************************************************\
*									   *
* action      :	 Zeiger	auf ausgenullten Speicherbereich der Laenge	   *
*		 request zurueckgeben, immer gerade Anzahl Bytes	   *
*		 allokieren.						   *
*									   *
*		 Diese Funktion	wird anstatt der Standardfunktion sbrk()   *
*		 benutzt, die beim Atari aufgrund der voll dynamischen	   *
*		 Speicherverwaltung auf	Betriebssystemebene bei	zu vielen  *
*		 Aufrufen Probleme machen kann (nur begrenzte Anzahl	   *
*		 Aufrufe innerhalb eines Programms moeglich).		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 request    -  Anzahl Bytes angeforderter Freispeicher	   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 reststart  -  Zeiger auf Anfang Restfreispeicher	   *
*		 restfree   -  Anzahl Bytes noch frei im Restfreispeicher  *
*									   *
* locals      :	 ret	    -  Zeiger auf angeforderten	Speicher	   *
*									   *
* returns     :	 ret, wenn die Anforderung erfuellt werden kann, -1 sonst  *
*									   *
\**************************************************************************/

char *clrsbrk(request)

unsigned request;

  {
    register char *ret;

    if (request	& 0x0001) ++request;
    if (restfree >= (long)request)
      {
	ret = reststart;
	reststart += request;
	restfree -= request;
	return (ret);
      }
    else
      return ((char *)(-1));
  }





/**************************************************************************\
*									   *
* action      :	 Tastatur abfragen, wenn ein Zeichen eingegeben	wurde,	   *
*		 Rueckgabe dieses Zeichens, sonst Rueckgabe von	EOF.	   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 EOF, wenn keine Eingabe erfolgt ist, das eingegebene	   *
*		 Zeichen sonst						   *
*									   *
\**************************************************************************/

int inkey()
  {
    struct timeval ti;
    int	k = 1;
    char c;

    ti.tv_sec =	0;
    ti.tv_usec = 0;
    select(1, &k, NULL,	NULL, &ti);
    if (k & 1) {
	read(0,&c,1);
	return (int)c;
    } else {
	return EOF;
    }
/*
    if (kbhit())
      return (getch());
    else
      return (EOF);
*/
  }





/**************************************************************************\
*									   *
* action      :	 Tastatur-Zeichenzwischenspeicher leeren.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 -							   *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 -							   *
*									   *
\**************************************************************************/

void kbclear()
  {
    struct timeval ti;
    int	k = 1;
    char c;

    ti.tv_sec =	0;
    ti.tv_usec = 0;
    do {
	select(1, &k, NULL, NULL, &ti);
	if (k &	1)
		read(0,&c,1);
    } while (k & 1);
/*
    while(inkey() != EOF) ;
*/
  }





/**************************************************************************\
*									   *
* action      :	 Oeffnen eines Textfiles mit Weg/Name.			   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 fname	 -  Zeiger auf String: Weg und Name des	Textfiles  *
+		 access	 -  Zeiger auf String: "r" - Oeffnen zum Lesen	   *
*					       "w" - Oeffnen zum Schreiben *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Zeiger	auf FILE-Struktur, wenn	Oeffnen	gelungen,	   *
*		 NULL sonst						   *
*									   *
\**************************************************************************/

FILE *fopena(fname,access)

char *fname;
char access[];

  {
    static char	code[] = "xt";

    code[0]=access[0];
    return (fopen(fname,code));		   /* TurboC fopen("...",".t")	  */
  }





/**************************************************************************\
*									   *
* action      :	 Oeffnen eines Binaerfiles mit Weg/Name.		   *
*									   *
 **************************************************************************
*									   *
* parameter   :	 fname	 -  Zeiger auf String: Weg und Name des	Files	   *
+		 access	 -  Zeiger auf String: "r" - Oeffnen zum Lesen	   *
*					       "w" - Oeffnen zum Schreiben *
*									   *
* r/o globals :	 -							   *
*									   *
* r/w globals :	 -							   *
*									   *
* locals      :	 -							   *
*									   *
* returns     :	 Zeiger	auf FILE-Struktur, wenn	Oeffnen	gelungen,	   *
*		 NULL sonst						   *
*									   *
\**************************************************************************/

FILE *fopenb(fname,access)

char *fname;
char access[];

  {
    static char	code[] = "xb";

    code[0]=access[0];
    return (fopen(fname,code));		   /* TurboC fopen("...",".b")	  */
  }



/* Ende	von SYSIBM.C */
