/**************************************************************************\
*									   *
*									   *
*    *****			*****					   *
*      *****		      *****					   *
*	 *****		    *****					   *
*	   *****	  *****						   *
*	     *****	*****						   *
*	       *****  *****						   *
*	     *****	*****						   *
*	   *****	  *****		 The Firmware.			   *
*	 *****		    *****	 Portable. Compatible.		   *
*      *****		      *****	 Public	Domain.			   *
*    *****			*****	 NORD><LINK.			   *
*									   *
*									   *
*									   *
*    RONNER.C	-   spezieller Komprimierer/Ersetzer fuer Q/C Assembler-   *
*		    Outputs						   *
*									   *
*    angelegt:	    DC4OX						   *
*    modifiziert:   DF2AU, 05-09-89, IBM Version			   *
*									   *
*    ACHTUNG: Option -3	noch nicht getestet (mit QC V3.2)		   *
*									   *
\**************************************************************************/



/*
 *   Aufruf :	ronner <inspec>	[-o <outspec>] [-e34]
 *
 *   <inspec>	  ist die Inputfile-Angabe, default Extension ist .CAS
 *
 *   -o	<outspec> ist die optionale Outputfile-Angabe, default Extension
 *		  ist .MAC
 *
 *   -e		  schaltet Ronner in den Expand-Modus, d.h. es werden
 *		  keine	Makros benutzt,	so dass	AO auch	nach dem Ronner
 *		  lauffaehig ist
 *
 *   -3		  es wird Q/C V3.2 verwendet, es werden	keine Q/C V4.0
 *		  spezifischen Anpassungen/Optimierungen gemacht
 *
 *   -4		  es wird Q/C V4.0 verwendet, es werden	aber keine
 *		  Anpassungen fuer Q/C V3.2 Kompatibilitaet gemacht,
 *		  es wird optimal auf kurzen Code optimiert
 *
 *   -f		  "fast", es werden nur	auch die Geschwindigkeit
 *		  optimierende Ersetzungen gemacht, nur	laengenoptimierende
 *		  Ersetzungen (MLLOAD, ADDHL) werden nicht gemacht,
 *		  einige Ersetzungen werden optimaler gemacht,
 *		  Ausnahme: CALL -> RST	Ersetzungen ohne Parameter werden
 *		  auch ausgefuehrt.
 *
 *
 *   Ronner kopiert das	Inputfile in das Outputfile und	ersetzt	dabei
 *   gewisse Pattern des Compileroutputs durch kuerzere	Aufrufe	aus der
 *   Speziallibrary CRUNTIME und andere	Pattern	durch kuerzere Sequenzen.
 *
 *   ACHTUNG: Durch das	Ersetzen von Pattern ist der Code nicht	absolut
 *	      konsistent fuer alle denkbaren C-Konstrukte.
 *	      Zusammengesetzte boolesche Zuweisungen wie z.B.
 *	      "if (c = (a != 0 || b == 1))" oder "if ((a || b == 1) == 1)"
 *	      muessen unbedingt	unterbleiben !
 *
 *   Beim Checken, ob ein JP/JR	dem eigentlichen Pattern folgt,
 *   werden nicht ggf. vorhandene Label	uebersprungen. Das ist zwar nicht
 *   optimal, weil das Pattern dann nicht ersetzt wird (obwohl man es
 *   ersetzen koennte),	aber es	ist "kompatibel".
 *
 *
 *   Folgende Pattern werden ersetzt :
 *
 *   (die M...-Statements sind RST-Makros, n bei IX+n immer n <	64)
 *
 *
 *     3   CALL	   ?sxt	      ->   1   M?SXT
 *     3   CALL	   ?gf	      ->   1   M?GF
 *     3   CALL	   ?g	      ->   1   M?G
 *     3   CALL	   ?sn	      ->   1   M?SN
 *
 *     1   PUSH	   HL	      ->   1   M?INC
 *     3   CALL	   ?g
 *     1   INC	   HL
 *     1   POP	   DE
 *     3   CALL	   ?p
 *
 *     1   PUSH	   BC	      ->   1   PUSH    BC
 *     3   CALL	   ?gf		   1   POP     HL
 *     1   INC	   HL		   1   M?INC
 *     1   POP	   DE
 *     3   CALL	   ?p
 *
 *     1   PUSH	   HL	      ->   3   CALL    ?dec
 *     3   CALL	   ?g
 *     1   DEC	   HL
 *     1   POP	   DE
 *     3   CALL	   ?p
 *
 *
 *  auch bei  -4 :
 *
 *   ( 1   PUSH	   BC	      ->   1   PUSH    BC     )	    koennte man
 *   ( 3   CALL	   ?gf		   1   POP     HL     )	    auch ersetzen ...
 *   ( 1   DEC	   HL		   3   CALL    ?dec   )	    ...	wird aber
 *   ( 1   POP	   DE				      )	    nicht (Kombati-
 *   ( 3   CALL	   ?p				      )	    bilitaet)
 *
 *     3   LD	   DE,n	      ->   2   MADDHL  n	    (0 < n < 256)
 *     1   ADD	   HL,DE
 *
 *     4   LD	   (IX+n),0   ->   2   MLCLRB  n
 *
 *     4   LD	   (IX+n),0   ->   2   MLCLRW  n	    (m = n + 1)
 *     4   LD	   (IX+m),0
 *
 *     3   LD	   (IX+n),L   ->   2   MLSAVE  n	    (m = n + 1)
 *     3   LD	   (IX+m),H
 *
 *     3   LD	   L,(IX+n)   ->   2   MLLOAD  n	    (m = n + 1)
 *     3   LD	   H,(IX+m)
 *
 *     1   POP	   DE	      ->   1   EX      (SP),HL
 *     1   PUSH	   HL		       CALL    ...
 *	   CALL	   ...
 *
 *     3   CALL	   ?gcs	      ->   1   LD      A,(HL)
 *     3   LD	   HL,0		   1   OR      A
 *     3   CALL	   ?sne		       JP/JR   Z/NZ,
 *	   JP/JR   Z/NZ,
 *
 *     3   CALL	   ?gc	      ->   1   LD      A,(HL)
 *     3   LD	   DE,n		   2   SUB     n
 *     3   CALL	   ?se		   3   CALL    ?Aeq0
 *	   JP/JR   Z/NZ,	       JP/JR   Z/NZ,
 *
 *     3   LD	   DE,0	      ->   1   LD      A,H
 *     3   CALL	   ?sne		   1   OR      L
 *	   JP/JR   Z/NZ,	       JP/JR   Z/NZ,
 *
 *     3   LD	   HL,0	      ->   1   LD      A,D
 *     3   CALL	   ?sne		   1   OR      E
 *	   JP/JR   Z/NZ,	       JP/JR   Z/NZ,
 *
 *     3   CALL	   ?sa	      ->   3   CALL    ?sna
 *     3   CALL	   ?sn
 *
 *     3   CALL	   ?gcs	      ->
 *     3   LD	   A,(var)	   3   LD      A,(var)
 *     3   CALL	   ?sxt		   1   SUB     (HL)
 *     3   CALL	   ?se		   3   CALL    ?Aeq0
 *	   JP/JR   Z/NZ,	       JP/JR   Z/NZ,
 *
 *     3   LD	   DE,255     ->   2   LD      H,0
 *     3   CALL	   ?sa
 *
 *     3   CALL	   ?g	      ->   3   CALL    ?gcsi
 *     3   CALL	   ?gcs
 *
 *   -----------------------------------------------------------------------
 *
 *     um Kompatibilitaet von Q/C V4.0 zu Q/C V3.2 zu erhalten
 *     (nur bei	nicht angewaehlter Option 3 oder 4) :
 *
 *     1   XOR	   A	      ->   2   LD      (HL),0
 *     1   LD	   (HL),A	   1   INC     HL
 *     1   INC	   HL		   2   LD      (HL),0
 *     1   LD	   (HL),A
 *
 *     3   CALL	   ?srem      ->   3   CALL    ?sdiv
 *				   1   EX      DE,HL
 *
 *     3   CALL	   ?urem      ->   3   CALL    ?udiv
 *				   1   EX      DE,HL
 *
 *     3   CALL	   ?s1s	      ->   1   EX      DE,HL
 *				   3   CALL    ?ss
 *
 *     3   LD	   DE,n	      ->   1   EX      DE,HL
 *     3   CALL	   ?s1div	   3   LD      HL,n
 *				   3   CALL    ?sdiv
 *
 *     3   CALL	   ?s1div     ->   1   EX      DE,HL	    (nur wenn nicht
 *				   3   CALL    ?sdiv	    Fall wie vorher)
 *
 *     3   LD	   DE,n	      ->   3   LD      DE,n	    (u... = Fall wie
 *     3   CALL	   ?u1div	   1   EX      DE,HL	     folgt)
 *				   3   CALL    ?udiv
 *
 *     3   CALL	   ?u1div     ->   1   EX      DE,HL	    (nur wenn nicht
 *				   3   CALL    ?udiv	    Fall wie vorher)
 *
 *
 *
 *  auch bei -4	:
 *
 *
 *     3   LD	   HL,s	      ->   3   LD      HL,$+5
 *     1   PUSH	   HL		   1   EX      (SP),HL
 *     3   LD	   HL,n		   1   JP      (HL)
 *     1   ADD	   HL,SP
 *     3   CALL	   ?g	      Der Einfachheit halber wird nur
 *     1   JP	   (HL)	      auf "JP (HL), POP	DE" gematcht,
 *	 s:		      das Label	kann, aber muss	nicht auf
 *     1   POP	   DE	      extra Zeile erscheinen !
 *
 *     Zusaetzlich werden je nach n keine (n = 2), oder	mehrere
 *     Parameterpushungen (n = 4,6,8, ...) fuer	die mit	"JP (HL)"
 *     aufgerufene Funktion mit	"EX (SP),HL" korrigiert.
 *     Dies wird nur durch einfaches Matchen auf "PUSH HL" gemacht -
 *     Aufrufe wie "(*func)(a,b,c)" sind zulaessig, aber Aufrufe
 *     wie "(*func)(a,f(b),c)" werden nicht richtig umgesetzt !
 *
 *
 *
 *     Achtung:	 Die Reihenfolge der Parameter auf dem Stack hat von
 *		 Q/C V3.2 auf Q/C V4.0 gewechselt !
 *
 */

#ifdef NOARGS
#define __ARGS(x) ()
#else
#define __ARGS(x) x
#endif


/*							       Includes	  */
/**************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>



/*							   Festlegungen	  */
/**************************************************************************/

#define	FNSIZE	   80	      /* Zeichenanzahl in Dateiname maximal	  */
#define	INPUTEXT   ".cas"     /* Input Dateiname default Extension	  */
#define	LINELEN	   133	      /* Zeilenlaenge Input maximal		  */
#define	NUMLINES   20	      /* Anzahl	Zeichen	Zeilenbuffer		  */
#define	OUTPUTEXT  ".mac"     /* Output	Dateinname default Extension	  */

#define	VOID void	      /* Funktion gibt nichts zurueck		  */



			 /* moegliche Befehlstoken :			  */
#define	NOTOKEN	    0	 /*   unbekanntes Token				  */
#define	LABEL	    1	 /*   nur Label	in der Zeile			  */
#define	PUSH_HL	    2	 /*   PUSH HL					  */
#define	INC_HL	    3	 /*   INC HL					  */
#define	DEC_HL	    4	 /*   DEC HL					  */
#define	POP_DE	    5	 /*   POP DE					  */
#define	PUSH_BC	    6	 /*   PUSH BC					  */
#define	CALL_p	    7	 /*   CALL ?p					  */
#define	CALL_sxt    8	 /*   CALL ?sxt					  */
#define	CALL_gf	    9	 /*   CALL ?gf					  */
#define	CALL_g	   10	 /*   CALL ?g					  */
#define	CALL_sn	   11	 /*   CALL ?sn					  */
#define	CALL_gc	   12	 /*   CALL ?gc					  */
#define	CALL_gcs   13	 /*   CALL ?gcs					  */
#define	CALL_sne   14	 /*   CALL ?sne					  */
#define	CALL_se	   15	 /*   CALL ?se					  */
#define	CALL_srem  16	 /*   CALL ?srem				  */
#define	CALL_urem  17	 /*   CALL ?urem				  */
#define	CALL_s1s   18	 /*   CALL ?s1s					  */
#define	CALL_sa	   19	 /*   CALL ?sa					  */
#define	CALL_s1div 20	 /*   CALL ?s1div				  */
#define	CALL_u1div 21	 /*   CALL ?u1div				  */
#define	LD_iIX0	   22	 /*   LD (IX+d),0				  */
#define	LD_LiIX	   23	 /*   LD L,(IX+d)				  */
#define	LD_HiIX	   24	 /*   LD H,(IX+d)				  */
#define	LD_iIXL	   25	 /*   LD (IX+d),L				  */
#define	LD_iIXH	   26	 /*   LD (IX+d),H				  */
#define	LD_iHLA	   27	 /*   LD (HL),A					  */
#define	LD_HL0	   28	 /*   LD HL,0					  */
#define	LD_HLn	   29	 /*   LD HL,n					  */
#define	LD_DE0	   30	 /*   LD DE,0					  */
#define	LD_DEn	   31	 /*   LD DE,n					  */
#define	LD_Ain	   32	 /*   LD A,(n)					  */
#define	LD_AiHL	   33	 /*   LD A,(HL)					  */
#define	LD_AiBC	   34	 /*   LD A,(BC)					  */
#define	LD_AiDE	   35	 /*   LD A,(DE)					  */
#define	LD_AiI	   36	 /*   LD A,(IX/IY+n)				  */
#define	LD_DEin	   37	 /*   LD DE,(n)					  */
#define	LD_HLin	   38	 /*   LD HL,(n)					  */
#define	XOR_A	   39	 /*   XOR A					  */
#define	JR_Z	   40	 /*   JR/JP Z,					  */
#define	JR_NZ	   41	 /*   JR/JP NZ,					  */
#define	JP_iHL	   42	 /*   JP (HL)					  */
#define	ADD_HD	   43	 /*   ADD HL,DE					  */



			 /* Zustaende fuer state (s.u.)	:		  */
#define	INITIAL	   0	 /*   Anfangszustand				  */
#define	H	   1	 /*   PUSH HL					  */
#define	HG	   2	 /*   PUSH HL, CALL ?g				  */
#define	HGI	   3	 /*   PUSH HL, CALL ?g,	INC HL			  */
#define	HGID	   4	 /*   PUSH HL, CALL ?g,	INC HL,	POP DE		  */
#define	HGD	   5	 /*   PUSH HL, CALL ?g,	DEC HL			  */
#define	HGDD	   6	 /*   PUSH HL, CALL ?g,	DEC HL,	POP DE		  */
#define	IX0	   7	 /*   LD (IX+d),0				  */
#define	IXL	   8	 /*   LD (IX+d),L				  */
#define	LIX	   9	 /*   LD L,(IX+d)				  */
#define	D	  10	 /*   POP DE					  */
#define	DH	  11	 /*   POP DE, PUSH HL				  */
#define	XA	  12	 /*   XOR A					  */
#define	XAH	  13	 /*   XOR A, LD	(HL),A				  */
#define	XAHI	  14	 /*   XOR A, LD	(HL),A,	INC HL			  */
#define	GCS	  15	 /*   CALL ?gcs					  */
#define	GCS0	  16	 /*   CALL ?gcs, LD HL,0			  */
#define	GCS0SNE	  17	 /*   CALL ?gcs, LD HL,0, CALL ?sne		  */
#define	B	  18	 /*   PUSH BC					  */
#define	BG	  19	 /*   PUSH BC, CALL ?gf				  */
#define	BGI	  20	 /*   PUSH BC, CALL ?gf, INC HL			  */
#define	BGID	  21	 /*   PUSH BC, CALL ?gf, INC HL, POP DE		  */
#define	BGD	  22	 /*   PUSH BC, CALL ?gf, DEC HL			  */
#define	BGDD	  23	 /*   PUSH BC, CALL ?gf, DEC HL, POP DE		  */
#define	H0	  24	 /*   LD HL,0					  */
#define	H0SNE	  25	 /*   LD HL,0, CALL ?SNE			  */
#define	D0	  26	 /*   LD DE,0					  */
#define	D0SNE	  27	 /*   LD DE,0, CALL ?SNE			  */
#define	GC	  28	 /*   CALL ?gc					  */
#define	GCD	  29	 /*   CALL ?gc,	LD DE,n				  */
#define	GCDSE	  30	 /*   CALL ?gc,	LD,DE,n, CALL ?se		  */
#define	SA	  31	 /*   CALL ?sa					  */
#define	JPH	  32	 /*   JP (HL)					  */
#define	JPHL	  33	 /*   JP (HL), Label:				  */
#define	GCSA	  34	 /*   CALL ?gcs, LD A,(n)			  */
#define	GCSAX	  35	 /*   CALL ?gcs, LD A,(n), CALL	?sxt		  */
#define	GCSAXE	  36	 /*   CALL ?gcs, LD A,(n), CALL	?sxt, CALL ?se	  */
#define	DE255	  37	 /*   LD DE,255					  */
#define	DEN	  38	 /*   LD DE,n					  */
#define	G	  39	 /*   CALL ?g					  */





/*								Globals	  */
/**************************************************************************/

char	   infn[FNSIZE+1];		/* Dateiname Inputdatei		  */
FILE	  *infile;			/* Inputdatei			  */
char	   outfn[FNSIZE+1];		/* Dateiname Outputdatei	  */
FILE	  *outfile;			/* Outputdatei			  */
char	   line[NUMLINES][LINELEN];	/* Input-Zeilenbuffer		  */
unsigned   inindex;			/* naechste Zeile in Buffer	  */
unsigned   outindex;			/* letzte Zeile	in Buffer	  */
unsigned   state;			/* s.o.				  */
unsigned   saved;			/* Anzahl Bytes	eingespart	  */
unsigned   ixo;				/* d aus (IX+d)	   setzt token()  */
unsigned   ixo1;			/* d aus (IX+d)	   1. Vorkommen	  */
unsigned   dn;				/* n aus LD DE,n   setzt token()  */
unsigned   hn;				/* n aus LD HL,n   setzt token()  */
unsigned   idx;				/* aktueller Index in line[]	  */
unsigned   opte	= 0;			/* 1 = -e angewaehlt, 0	sonst	  */
unsigned   opt3	= 0;			/* 1 = -3 angewaehlt, 0	sonst	  */
unsigned   opt4	= 0;			/* 1 = -4 angewaehlt, 0	sonst	  */
unsigned   optf	= 0;			/* 1 = -f angewaehlt, 0	sonst	  */
unsigned   call;			/* 1 = Token ist CALL..., 0 sonst */
					/*   gesetzt von token()	  */

VOID usagerr __ARGS((void));
VOID fatal __ARGS((char	*, char	*));
VOID makefn __ARGS((char *,char	*,char *));
VOID putmaks __ARGS((void));
unsigned isnxts __ARGS((char *,	char *));
unsigned isnext __ARGS((char *,	char *));
char *getint __ARGS((char*, int	*));
unsigned token __ARGS((char *));
void new __ARGS((unsigned));
VOID procline __ARGS((void));
VOID previdx __ARGS((void));
char *clrlline __ARGS((void));
VOID replace __ARGS((unsigned, unsigned, char *));
VOID doclr8 __ARGS((void));
VOID doclr16 __ARGS((void));
VOID dosave __ARGS((void));
VOID doload __ARGS((void));
VOID doaddhl __ARGS((void));
VOID dotstahl __ARGS((void));
VOID dosaeq0 __ARGS((void));
VOID doexsdiv __ARGS((void));
VOID dojpihl __ARGS((void));
VOID doinc __ARGS((void));
VOID dodec __ARGS((void));
VOID doexsphl __ARGS((void));
VOID doldaora __ARGS((void));
VOID dobinc __ARGS((void));
VOID dobdec __ARGS((void));
VOID dotsthl __ARGS((void));
VOID dotstde __ARGS((void));
VOID dosna __ARGS((void));
VOID domask __ARGS((void));
VOID dogcsi __ARGS((void));
VOID doclrihl __ARGS((void));
VOID dosrem __ARGS((void));
VOID dourem __ARGS((void));
VOID dos1s __ARGS((void));
VOID dos1div __ARGS((void));
VOID dou1div __ARGS((void));
VOID doexudiv __ARGS((void));



/*					       Hauptprogramm/Funktionen	  */
/**************************************************************************/



/**************************************************************************\
*									   *
* Hauptprogramm	(Aufruf/Funktion siehe oben).				   *
*									   *
\**************************************************************************/

void main(argc,argv)

int argc;
char *argv[];

  {
    char *argptr;			/* zum Scannen der Optionen	  */
    char cs[2];				/* Optionszeichenstring		  */



    /* Kommandozeile abarbeiten, Input-	und Outputdatei	oeffnen	*/

    fprintf(stderr,"NORD><LINK (DC4OX) Ronner Version 1.3a\n"
		   "ported to Linux by dl4sdc\n");
    if ((argc -= 2) < 0) usagerr();		  /* Dateiangabe noetig	  */
    makefn(*++argv,infn,INPUTEXT);		  /* Inputdateiname	  */
    if (*infn == '-') usagerr();
    if (!(infile = fopen(infn,"r")))		  /* Inputdatei	oeffnen	  */
      fatal("Unable to open input file:	",infn);
    strcpy(outfn,infn);				  /* default Outputname	  */
    strcpy(strchr(outfn,'.'),OUTPUTEXT);
    argptr = *++argv;
    cs[1] = '\0';
    while (argc)
      {
	if (*argptr == '-') ++argptr;		  /* '-' uebergehen	  */
	while (*argptr)
	  switch (toupper(cs[0]	= *argptr++))
	    {
	      case 'O' :
		--argc;
		if (argc)
		  makefn(*++argv,outfn,OUTPUTEXT);
		else
		  fatal("-o, must specify file name","");
		if (*outfn == '-') usagerr();
	      break;

	      case 'E' :  opte = 1;  break;

	      case '3' :  opt3 = 1;  break;

	      case '4' :  opt4 = 1;  break;

	      case 'F' :  optf = 1;  break;

	      default  :  fatal("Unknown option: ",cs);	 break;
	   }
	if (argc-- > 1)	argptr = *++argv;
      }
    if (opt3 &&	opt4)				  /* 3 und 4 geht nicht	  */
      fatal("3 and 4 selected","");
    if (!strcmp(infn,outfn))			  /* Input = Output ?	  */
      fatal("Output file name same as input","");
    if (!(outfile = fopen(outfn,"w")))		 /* Output oeffnen	 */
      fatal("Unable to open output file: ",outfn);



    /* spezielle Externals / Makrodefinitionen in Outputdatei eintragen	  */

    putmaks();



    /* Inputdatei zeilenweise abarbeiten */

    for	(inindex = 0; inindex <	NUMLINES; *line[inindex++] = '\0');
    saved = inindex = outindex = 0;
    state = INITIAL;
    while (fgets(line[inindex],LINELEN-1,infile) != NULL)
      {
	procline();
	inindex	= (inindex + 1)	% NUMLINES;
	if (inindex == outindex)
	  {
	    if (*line[outindex])
	      {
		fputs(line[outindex],outfile);
		if (ferror(outfile))
		  {
		    fclose(outfile);
		    fatal("Write error on: ",outfn);
		  }
	      }
	    outindex = (outindex + 1) %	NUMLINES;
	  }
      }



    /* Falls Lesefehler	*/

    if (ferror(infile))
      {
	fclose(infile);
	fclose(outfile);
	fatal("Read error on: ",infn);
      }



    /* Inputdatei schliessen */

    fclose(infile);



    /* Rest des	Buffers	ausgeben auf Outputdatei */

    for	(; outindex != inindex;	outindex = (outindex + 1) % NUMLINES)
      if (*line[outindex])
	{
	  fputs(line[outindex],outfile);
	  if (ferror(outfile))
	    {
	      fclose(outfile);
	      fatal("Write error on: ",outfn);
	    }
	}



    /* Outputdatei schliessen, das war's */

    fclose(outfile);
    fprintf(stderr,"%d bytes saved.\n",saved);
    exit(0);
  }





/**************************************************************************\
*									   *
* Ausgabe einer	Fehlermeldung, wie das Programm	zu benutzen ist, und	   *
* Programm abbrechen.							   *
*									   *
\**************************************************************************/

VOID usagerr()
  {
    fatal("usage is 'ronner <inspec> [-o <outspec>] [-ef34]'","");
  }





/**************************************************************************\
*									   *
* Ausgabe einer	Fehlermeldung und Programm abbrechen.			   *
*									   *
*   s1	-  Zeiger auf String des ersten	Teils der Fehlermeldung		   *
*   s2	-  Zeiger auf String des zweiten Teils der Fehlermeldung	   *
*									   *
* Ausgegeben wird "FATAL ERROR:	", dann	die beiden Strings, dann CR/LF.	   *
*									   *
\**************************************************************************/

VOID fatal(s1,s2)

char *s1;
char *s2;

  {
    fprintf(stderr,"FATAL ERROR: %s%s\n",s1,s2);  /* Meldung ausgeben	  */
    exit(1);					  /* Programm abbrechen	  */
  }





/**************************************************************************\
*									   *
* Filename dst nach src	kopieren, wenn ohne Extension, Extension ext	   *
* anhaengen. Programmabbruch mit Meldung, wenn src zu lang ist oder dst	   *
* zu lang wird.								   *
*									   *
\**************************************************************************/

VOID makefn(src,dst,ext)

char *src;
char *dst;
char *ext;

  {
    if (strlen(src) > FNSIZE)
      fatal("Filename too long:	",src);
    strcpy(dst,src);
    if (!strchr(dst,'.'))
      {
	if(strlen(dst) > (FNSIZE - 4))
	  fatal("With extension	too long: ",dst);
	strcat(dst,ext);
      }
  }





/**************************************************************************\
*									   *
* Ausgabe der bei der Ersetzung	benutzten Makrodefinitionen sowie	   *
* spezielle Externals auf outfile.					   *
*									   *
\**************************************************************************/

VOID putmaks()
  {
    fputs("; NORD><LINK	SPECIAL	  ( Options: ",outfile);
    if (opte) fputs("e ",outfile);
    if (opt3) fputs("3 ",outfile);
    if (opt4) fputs("4 ",outfile);
    if (optf) fputs("f ",outfile);
    fputs(")\n\n extrn ?Aeq0\n extrn ?sna\n",outfile);
    fputs(" extrn ?inc\n extrn ?dec\n extrn ?gcsi\n\n",outfile);
    if (!opte)
      {
	fputs("M?INC\tmacro\n\trst\t08h\n\tendm\n\n"		, outfile);
	fputs("M?SXT\tmacro\n\trst\t10h\n\tendm\n\n"		, outfile);
	fputs("M?GF\tmacro\n\trst\t18h\n\tendm\n\n"		, outfile);
	fputs("M?G\tmacro\n\trst\t20h\n\tendm\n\n"		, outfile);
	fputs("M?SN\tmacro\n\trst\t28h\n\tendm\n\n"		, outfile);
	fputs("MADDHL\tmacro\n\trst\t30h\n\tdefb\t?1"	, outfile);
/*	fputs("MADDHL\tmacro\tofs\n\trst\t30h\n\tdefb\tofs"	, outfile); */
	fputs("\n\tendm\n\n"					, outfile);
	fputs("MLLOAD\tmacro\n\trst\t38h\n\tdefb\t?1"	, outfile);
/*	fputs("MLLOAD\tmacro\tofs\n\trst\t38h\n\tdefb\tofs"	, outfile); */
	fputs("\n\tendm\n\n"					, outfile);
	fputs("MLSAVE\tmacro\n\trst\t38h\n\tdefb\t?1"	, outfile);
/*	fputs("MLSAVE\tmacro\tofs\n\trst\t38h\n\tdefb\tofs"	, outfile); */
	fputs("	| 01000000b\n\tendm\n\n"			, outfile);
	fputs("MLCLRW\tmacro\n\trst\t38h\n\tdefb\t?1"	, outfile);
/*	fputs("MLCLRW\tmacro\tofs\n\trst\t38h\n\tdefb\tofs"	, outfile); */
	fputs("	| 10000000b\n\tendm\n\n"			, outfile);
	fputs("MLCLRB\tmacro\n\trst\t38h\n\tdefb\t?1"	, outfile);
/*	fputs("MLCLRB\tmacro\tofs\n\trst\t38h\n\tdefb\tofs"	, outfile); */
	fputs("	| 11000000b\n\tendm\n\n"			, outfile);
      }
  }





/**************************************************************************\
*									   *
* Vergleichen der beiden Strings ptr und string.			   *
*									   *
*   return   1	 -   string und	ptr stimmen in strlen(string) Zeichen	   *
*		     ueberein, ptr[strlen(string)] muss	isspace	sein	   *
*									   *
*	     0	 -   sonst						   *
*									   *
\**************************************************************************/

unsigned isnxts(ptr,string)

register char *ptr;
register char *string;

  {
    while (*ptr	== *string)
      {
	if (*string == '\0') return (1);
	++ptr;
	++string;
      }
    if (*string	== '\0'	&& (*ptr == '\0' || isspace(*ptr))) return (1);
    return (0);
  }





/**************************************************************************\
*									   *
* Vergleichen der beiden Strings ptr und string.			   *
*									   *
*   return   1	 -   string und	ptr stimmen in strlen(string) Zeichen	   *
*		     ueberein						   *
*									   *
*	     0	 -   sonst						   *
*									   *
\**************************************************************************/

unsigned isnext(ptr,string)

register char *ptr;
register char *string;

  {
    while (*ptr	== *string)
      {
	if (*string == '\0') return (1);
	++ptr;
	++string;
      }
    if (*string	== '\0') return	(1);
    return (0);
  }





/**************************************************************************\
*									   *
* Dezimalzahlstring ab ptr in Dezimalzahl umwandeln und	in ip eintragen,   *
* Rueckgabe eines Zeiger auf hinter die	Zahl im	String.			   *
*									   *
\**************************************************************************/

char *getint(ptr,ip)

register char *ptr;
register int  *ip;

  {
    register int    i;
    unsigned	    minus;

    minus = i =	0;
    if (*ptr ==	'-')
      {
	minus =	1;
	++ptr;
      }
    while (*ptr	>= '0' && *ptr <= '9') i = i * 10 + *ptr++ - '0';
    if (minus)
      *ip = -i;
    else
      *ip = i;
    return (ptr);
  }





/**************************************************************************\
*									   *
* Token	des in der Q/C-Assembler-Outputzeile ptr stehenden Statements	   *
* zurueckgeben.	Falls das Statement unbekannt ist, NOTOKEN zurueckgeben.   *
*									   *
* Hier werden auch folgende Einzeilen-RST-Makroersetzungen vorgenommen :   *
*									   *
*   "CALL    ?sxt"   ->	  "M?SXT"					   *
*   "CALL    ?gf"    ->	  "M?GF"					   *
*   "CALL    ?g"     ->	  "M?G"						   *
*   "CALL    ?sn"    ->	  "M?SN"					   *
*									   *
* saved	wird je	Ersetzung um 2 erhoeht.					   *
*									   *
* Je nach dem Token werden ixo (IX+ixo)	oder dn	(LD DE,dn) gesetzt, es	   *
* wird grundsaetzlich bei jedem	Aufruf ixo1 auf	ixo gesetzt.		   *
*									   *
* call wird auf	1 gesetzt, wenn	Befehl ein CALL	ist, auf 0 sonst.	   *
*									   *
* "JP" wird in den Token wie "JR" behandelt.				   *
*									   *
* Da eine Statemachine zum Patternmatchen benutzt wird,	koennen	die	   *
* Ersetzungen hier gemacht werden, eine	Assemblerzeile wird ja nur ein	   *
* einziges Mal untersucht.						   *
*									   *
\**************************************************************************/

unsigned token(ptr)

register char *ptr;

  {
    char	     num;		/* fuer	LD ..,n			  */
    int		     ixn;		/* fuer	...),0 Feststellung	  */
    char	    *start;		/* fuer	Labelfeststellung	  */



    ixo1 = ixo;				/* fuer	16-Bit lokale Zugriffe	  */
    start = ptr;

    while (*ptr	&& *ptr	!= '\t') ++ptr;	/* Labelfeld ueberspringen	  */
    if (!*ptr++)			/* Leerzeile oder nur Label	  */
      {
	if (strlen(start) > 2) return(LABEL);
	return(NOTOKEN);
      }


    if (isnext(ptr,"CALL"))
      {
	call = 1;

	if (isnxts(ptr,"CALL\t?sxt"))
	  {
/*	    if (!optf)	 */
	      {
		*ptr = '\0';
		if (opte)
		  strcat(ptr,"rst\t10h\t\t; M?SXT\n");
		else
		  strcat(ptr,"M?SXT\n");
		saved += 2;
	      }
	    return (CALL_sxt);
	  }

	if (isnxts(ptr,"CALL\t?gf"))
	  {
/*	    if (!optf)	 */
	      {
		*ptr = '\0';
		if (opte)
		  strcat(ptr,"rst\t18h\t\t; M?GF\n");
		else
		  strcat(ptr,"M?GF\n");
		saved += 2;
	      }
	    return (CALL_gf);
	  }

	if (isnxts(ptr,"CALL\t?g"))
	  {
/*	    if (!optf)	 */
	      {
		*ptr = '\0';
		if (opte)
		  strcat(ptr,"rst\t20h\t\t; M?G\n");
		else
		  strcat(ptr,"M?G\n");
		saved += 2;
	      }
	    return (CALL_g);
	  }

	if (	isnxts(ptr,"CALL\t?sn")
	     ||	(opt3 && isnxts(ptr,"CALL\t?n"))
	   )
	  {
/*	    if (!optf)	 */
	      {
		*ptr = '\0';
		if (opte)
		  strcat(ptr,"rst\t28h\t\t; M?SN\n");
		else
		  strcat(ptr,"M?SN\n");
		saved += 2;
	      }
	    return (CALL_sn);
	  }

	if (isnxts(ptr,"CALL\t?p"    ))	return (CALL_p);
	if (isnxts(ptr,"CALL\t?gc"   ))	return (CALL_gc);
	if (isnxts(ptr,"CALL\t?gcs"  ))	return (CALL_gcs);

	if (opt3)
	  {
	    if (isnxts(ptr,"CALL\t?ne"	 )) return (CALL_sne);
	    if (isnxts(ptr,"CALL\t?e"	 )) return (CALL_se);
	    if (isnxts(ptr,"CALL\t?a"	 )) return (CALL_sa);
	  }
	else
	  {
	    if (isnxts(ptr,"CALL\t?sne"	 )) return (CALL_sne);
	    if (isnxts(ptr,"CALL\t?se"	 )) return (CALL_se);
	    if (isnxts(ptr,"CALL\t?srem" )) return (CALL_srem);
	    if (isnxts(ptr,"CALL\t?urem" )) return (CALL_urem);
	    if (isnxts(ptr,"CALL\t?s1s"	 )) return (CALL_s1s);
	    if (isnxts(ptr,"CALL\t?sa"	 )) return (CALL_sa);
	    if (isnxts(ptr,"CALL\t?s1div")) return (CALL_s1div);
	    if (isnxts(ptr,"CALL\t?u1div")) return (CALL_u1div);
	  }

      }	/* end if (isnext(ptr,"CALL")) */
    else
      call = 0;


    if (isnext(ptr,"LD\t(IX+"))
      {
	ptr = getint(ptr+7,&ixo);
	ptr += 2;
	if (*ptr == 'L') return	(LD_iIXL);
	if (*ptr == 'H') return	(LD_iIXH);
	getint(ptr,&ixn);
	if (ixn) return	(NOTOKEN);
	return (LD_iIX0);
      }

    if (isnext(ptr,"LD\tL,(IX+"))
      {
	getint(ptr + 9,&ixo);
	return (LD_LiIX);
      }

    if (isnext(ptr,"LD\tH,(IX+"))
      {
	getint(ptr + 9,&ixo);
	return (LD_HiIX);
      }

    if (isnext(ptr,"LD\tDE,"))
      {
	if (isnext(ptr,"LD\tDE,(")) return (LD_DEin);
	if ((num = ptr[6]) >= '0' && num <= '9')
	  {
	    getint(ptr + 6,&dn);
	    if (dn) return (LD_DEn);
	    return (LD_DE0);
	  }
	else
	  return(NOTOKEN);
      }

    if (isnext(ptr,"LD\tHL,"))
      {
	if (isnext(ptr,"LD\tHL,(")) return (LD_HLin);
	if ((num = ptr[6]) >= '0' && num <= '9')
	  {
	    getint(ptr + 6,&hn);
	    if (hn) return (LD_HLn);
	    return (LD_HL0);
	  }
	else
	  return(NOTOKEN);
      }

    if (isnext(ptr,"LD\tA,("))
      {
	if (isnxts(ptr,"LD\tA,(HL)")) return (LD_AiHL);
	if (isnxts(ptr,"LD\tA,(BC)")) return (LD_AiBC);
	if (isnxts(ptr,"LD\tA,(DE)")) return (LD_AiDE);
	if (isnext(ptr,"LD\tA,(I"))   return (LD_AiI);
	return (LD_Ain);
      }

    if (isnxts(ptr,"ADD\tHL,DE"	 )) return (ADD_HD);

    if (optf ||	opt4)
      {
	if (isnext(ptr,"JR\tZ,"	 )) return (JR_Z);	/* besser ...	  */
	if (isnext(ptr,"JP\tZ,"	 )) return (JR_Z);
	if (isnext(ptr,"JR\tNZ," )) return (JR_NZ);
	if (isnext(ptr,"JP\tNZ," )) return (JR_NZ);
      }
    else
      {
	if (isnext(start,"\tJR\tZ,"  ))	return (JR_Z);	/* ... kompatibel */
	if (isnext(start,"\tJP\tZ,"  ))	return (JR_Z);
	if (isnext(start,"\tJR\tNZ," ))	return (JR_NZ);
	if (isnext(start,"\tJP\tNZ," ))	return (JR_NZ);
      }

    if (isnxts(ptr,"PUSH\tHL"	 )) return (PUSH_HL);
    if (isnxts(ptr,"PUSH\tBC"	 )) return (PUSH_BC);
    if (isnxts(ptr,"DEC\tHL"	 )) return (DEC_HL);
    if (isnxts(ptr,"POP\tDE"	 )) return (POP_DE);
    if (isnxts(ptr,"INC\tHL"	 )) return (INC_HL);
    if (isnxts(ptr,"XOR\tA"	 )) return (XOR_A);
    if (isnxts(ptr,"LD\t(HL),A"	 )) return (LD_iHLA);
    if (isnxts(ptr,"JP\t(HL)"	 )) return (JP_iHL);

    return (NOTOKEN);
  }





/**************************************************************************\
*									   *
* Die Ersetzungsstatemachine.						   *
*									   *
\**************************************************************************/



void new(t)

unsigned t;

  {
    switch (t)
      {
	case PUSH_HL	:		  state	= H;	   break;
	case PUSH_BC	:		  state	= B;	   break;
	case POP_DE	:		  state	= D;	   break;
	case JP_iHL	:		  state	= JPH;	   break;
	case LD_iIX0	:		  state	= IX0;	   break;
	case LD_iIXL	:		  state	= IXL;	   break;
	case LD_LiIX	:		  state	= LIX;	   break;
	case LD_HL0	:		  state	= H0;	   break;
	case LD_DE0	:		  state	= D0;	   break;
	case LD_DEn	:  if (dn == 255) state	= DE255;
			   else		  state	= DEN;	   break;
	case XOR_A	:		  state	= XA;	   break;
	case CALL_gcs	:		  state	= GCS;	   break;
	case CALL_gc	:		  state	= GC;	   break;
	case CALL_g	:		  state	= G;	   break;
	case CALL_sa	:		  state	= SA;	   break;
	case CALL_srem	:  dosrem();	  state	= INITIAL; break;
	case CALL_urem	:  dourem();	  state	= INITIAL; break;
	case CALL_s1s	:  dos1s();	  state	= INITIAL; break;
	case CALL_s1div	:  dos1div();	  state	= INITIAL; break;
	case CALL_u1div	:  dou1div();	  state	= INITIAL; break;
	default		:		  state	= INITIAL; break;
      }
  }



VOID procline()
  {
    register unsigned t;

    idx	= inindex ? inindex - 1	: NUMLINES - 1;
    t =	token(line[idx]);
    switch (state)
      {

	case INITIAL :	new(t);	 break;

	case H :  if (t	== CALL_g) state = HG; else new(t);  break;

	case HG	:
	  switch (t)
	    {
	      case INC_HL  :  state = HGI;  break;
	      case DEC_HL  :  state = HGD;  break;
	      default	   :  new(t);	    break;
	    }
	break;

	case HGI :  if (t == POP_DE) state = HGID; else	new(t);	 break;

	case HGID :
	  switch (t)
	    {
	      case CALL_p  :  doinc();	   state = INITIAL;  break;
	      case PUSH_HL :  doexsphl();  state = INITIAL;  break;
	      default	   :  new(t);			     break;
	    }
	break;

	case HGD :  if (t == POP_DE) state = HGDD; else	new(t);	 break;

	case HGDD :
	  switch (t)
	    {
	      case CALL_p  :  dodec();	   state = INITIAL;  break;
	      case PUSH_HL :  doexsphl();  state = INITIAL;  break;
	      default	   :  new(t);			     break;
	    }
	break;

	case IX0 :
	  if (t	== LD_iIX0)
	    if (ixo == ixo1 + 1) { doclr16(); state = INITIAL; }
	    else doclr8();
	  else { doclr8(); new(t); }
	break;

	case IXL :
	  if (t	== LD_iIXH)
	    {
	      if (ixo == ixo1 +	1) dosave();
	      state = INITIAL;
	    }
	  else new(t);
	break;

	case LIX :
	  if (t	== LD_HiIX)
	    {
	      if (ixo == ixo1 +	1) doload();
	      state = INITIAL;
	    }
	  else new(t);
	break;

	case D :
	  switch (t)
	    {
	      case PUSH_HL :  doexsphl();  state = INITIAL;  break;
	      default	   :  new(t);
	    }
	break;

	case DH	:
	  if (call) { doexsphl(); state	= INITIAL; }
	  else new(t);
	break;

	case XA	:  if (t == LD_iHLA) state = XAH; else new(t);	break;

	case XAH :  if (t == INC_HL) state = XAHI; else	new(t);	 break;

	case XAHI :
	  if (t	== LD_iHLA) { doclrihl(); state	= INITIAL; }
	  else new(t);
	break;

	case GCS :
	  switch (t)
	    {
	      case LD_HL0 :  state = GCS0;  break;
	      case LD_Ain :  state = GCSA;  break;
	      default	  :  new(t);	    break;
	    }
	break;

	case GCS0 :  if	(t == CALL_sne)	state =	GCS0SNE; else new(t); break;

	case GCS0SNE :
	  if (t	== JR_Z	|| t ==	JR_NZ) { doldaora(); state = INITIAL; }
	  else new(t);
	break;

	case B :  if (t	== CALL_gf) state = BG;	else new(t);  break;

	case BG	:
	  switch (t)
	    {
	      case INC_HL  :  state = BGI;  break;
	      case DEC_HL  :  state = BGD;  break;
	      default	   :  new(t);	    break;
	    }
	break;

	case BGI :  if (t == POP_DE) state = BGID; else	new(t);	 break;

	case BGID :
	  switch (t)
	    {
	      case CALL_p  :  dobinc();	   state = INITIAL;  break;
	      case PUSH_HL :  doexsphl();  state = INITIAL;  break;
	      default	   :  new(t);			     break;
	    }
	break;

	case BGD :  if (t == POP_DE) state = BGDD; else	new(t);	 break;

	case BGDD :
	  switch (t)
	    {
	      case CALL_p  :  dobdec();	   state = INITIAL;  break;
	      case PUSH_HL :  doexsphl();  state = INITIAL;  break;
	      default	   :  new(t);			     break;
	    }
	break;

	case H0	:  if (t == CALL_sne) state = H0SNE; else new(t); break;

	case H0SNE :
	  if (t	== JR_Z	|| t ==	JR_NZ) { dotstde(); state = INITIAL; }
	  else new(t);
	break;

	case D0	:  if (t == CALL_sne) state = D0SNE; else new(t); break;

	case D0SNE :
	  if (t	== JR_Z	|| t ==	JR_NZ) { dotsthl(); state = INITIAL; }
	  else new(t);
	break;

	case GC	:  if (t == LD_DEn) state = GCD; else new(t);  break;

	case GCD :
	  switch (t)
	    {
	      case CALL_se    :			  state	= GCDSE;    break;
	      case ADD_HD     :	 doaddhl();	  state	= INITIAL;  break;
	      case CALL_s1div :	 doexsdiv();	  state	= INITIAL;  break;
	      case CALL_u1div :	 doexudiv();	  state	= INITIAL;  break;
	      case CALL_sa    :	 if (dn	== 255)
				   {
				     domask();
				     state = INITIAL;
				   }
				 else new(t);			    break;
	      default	      :	 new(t);			    break;
	    }
	break;

	case GCDSE :
	  if (t	== JR_Z	|| t ==	JR_NZ) { dotstahl(); state = INITIAL; }
	  else new(t);
	break;

	case SA	:
	  if (t	== CALL_sn) { dosna(); state = INITIAL;	}
	  else new(t);
	break;

	case JPH :
	  switch (t)
	    {
	      case POP_DE :  dojpihl();	 state = INITIAL;  break;
	      case LABEL  :		 state = JPHL;	   break;
	      default	  :  new(t);			   break;
	    }
	break;

	case JPHL :
	  if (t	== POP_DE) { dojpihl();	state =	INITIAL; }
	  else new(t);
	break;

	case GCSA :  if	(t == CALL_sxt)	state =	GCSAX; else new(t);  break;

	case GCSAX :  if (t == CALL_se)	state =	GCSAXE;	else new(t);  break;

	case GCSAXE :
	  if (t	== JR_Z	|| t ==	JR_NZ) { dosaeq0(); state = INITIAL; }
	  else new(t);
	break;

	case DE255 :
	  switch (t)
	    {
	      case CALL_sa    :	 domask();    state = INITIAL;	break;
	      case ADD_HD     :	 doaddhl();   state = INITIAL;	break;
	      case CALL_s1div :	 doexsdiv();  state = INITIAL;	break;
	      case CALL_u1div :	 doexudiv();  state = INITIAL;	break;
	      default	      :	 new(t);			break;
	    }
	break;

	case DEN :
	  switch (t)
	    {
	      case ADD_HD     :	 doaddhl();   state = INITIAL;	break;
	      case CALL_s1div :	 doexsdiv();  state = INITIAL;	break;
	      case CALL_u1div :	 doexudiv();  state = INITIAL;	break;
	      default	      :	 new(t);			break;
	    }
	break;

	case G :
	  if (t	== CALL_gcs) { dogcsi(); state = INITIAL; }
	  else new(t);
	break;

      }	/* end switch(state) */
  }





/**************************************************************************\
*									   *
* idx um eine Position zurueck positionieren.				   *
*									   *
\**************************************************************************/

VOID previdx()
  {
    idx	= idx ?	idx - 1	: NUMLINES - 1;
  }





/**************************************************************************\
*									   *
* Textzeile line[idx] loeschen,	Label bleibt erhalten.			   *
*									   *
* Rueckgabe :  Zeiger auf das Zeilenende ('\0').			   *
*									   *
\**************************************************************************/

char *clrlline()
  {
    register char *ptr;

    ptr	= line[idx];
    while (*ptr	!= '\t') ++ptr;
    *++ptr = '\0';
    return(ptr);
  }





/**************************************************************************\
*									   *
* nl Zeilen in line ab idx rueckwaerts durch s ersetzen, Label		   *
* in der ersten	zu ersetzenden Zeile bleibt erhalten.			   *
*									   *
* idx wird nl-mal zurueckgezaehlt, sp wird auf saved addiert.		   *
*									   *
\**************************************************************************/

VOID replace(nl,sp,s)

register unsigned    nl;
unsigned	     sp;
char		    *s;

  {
    register unsigned	 n;

    --nl;
    for	(n = 0;	n < nl;	++n)
      {
	*line[idx] = '\0';
	previdx();
      }
    strcat(clrlline(),s);
    previdx();
    saved += sp;
  }





/**************************************************************************\
*									   *
* >>> 4	  LD	  (IX+ixo1),0	->   2	 MLCLRB	 ixo1			   *
*									   *
* Die Ersetzung	wird nur ausgefuehrt, wenn ixo1	< 64 ist.		   *
*									   *
\**************************************************************************/

VOID doclr8()
  {
    if (ixo1 < 64 && !optf)
      {
	previdx();
	if (opte)
	  sprintf(clrlline(),"rst\t38h\t\t; MLCLRB %d\n\tdefb\t%d\n",
		  ixo1,ixo1 | 0xC0);
	else
	  sprintf(clrlline(),"MLCLRB\t%d\n",ixo1);
	saved += 2;
      }
  }





/**************************************************************************\
*									   *
*     4	  LD	  (IX+ixo1),0	->   2	 MLCLRW	 ixo1	   (m =	ixo1 + 1)  *
* >>> 4	  LD	  (IX+m),0						   *
*									   *
* Die Ersetzung	wird nur ausgefuehrt, wenn ixo1	< 64 ist.		   *
*									   *
\**************************************************************************/

VOID doclr16()
  {
    if (ixo1 < 64 && !optf)
      {
	*line[idx] = '\0';
	previdx();
	if (opte)
	  sprintf(clrlline(),"rst\t38h\t\t; MLCLRW %d\n\tdefb\t%d\n",
		  ixo1,ixo1 | 0x80);
	else
	  sprintf(clrlline(),"MLCLRW\t%d\n",ixo1);
	saved += 6;
      }
  }





/**************************************************************************\
*									   *
*     3	  LD	  (IX+ixo1),L	->   2	 MLSAVE	 ixo1	   (m =	ixo1 + 1)  *
* >>> 3	  LD	  (IX+m),H						   *
*									   *
* Die Ersetzung	wird nur ausgefuehrt, wenn ixo1	< 64 ist.		   *
*									   *
\**************************************************************************/

VOID dosave()
  {
    if (ixo1 < 64 && !optf)
      {
	*line[idx] = '\0';
	previdx();
	if (opte)
	  sprintf(clrlline(),"rst\t38h\t\t; MLSAVE %d\n\tdefb\t%d\n",
		  ixo1,ixo1 | 0x40);
	else
	  sprintf(clrlline(),"MLSAVE\t%d\n",ixo1);
	saved += 4;
      }
  }





/**************************************************************************\
*									   *
*     3	  LD	  L,(IX+ixo1)	->   2	 MLLOAD	 ixo1	   (m =	ixo1 + 1)  *
* >>> 3	  LD	  H,(IX+m)						   *
*									   *
\**************************************************************************/

VOID doload()
  {
    if (ixo1 < 64 && !optf)
      {
	*line[idx] = '\0';
	previdx();
	if (opte)
	  sprintf(clrlline(),"rst\t38h\t\t; MLLOAD %d\n\tdefb\t%d\n",
		  ixo1,ixo1);
	else
	  sprintf(clrlline(),"MLLOAD\t%d\n",ixo1);
	saved += 4;
      }
  }





/**************************************************************************\
*									   *
*     3	  LD	  DE,dn	     ->	  2   MADDHL  dn	   (0 <	n < 256)   *
* >>> 1	  ADD	  HL,DE							   *
*									   *
\**************************************************************************/

VOID doaddhl()
  {
    if (dn < 256 && !optf)
      {
	*line[idx] = '\0';
	previdx();
	if (opte)
	  sprintf(clrlline(),"rst\t30h\t\t; MADDHL %d\n\tdefb\t%d\n",
		  dn,dn);
	else
	  sprintf(clrlline(),"MADDHL\t%d\n",dn);
	saved += 2;
      }
  }





/**************************************************************************\
*									   *
*     3	  CALL	  ?gc	     ->	  1   LD      A,(HL)			   *
*     3	  LD	  DE,dn		  2   SUB     dn			   *
*     3	  CALL	  ?se		  3   CALL    ?Aeq0			   *
* >>>	  JP/JR	  NZ,		      JP/JR   NZ,			   *
*									   *
\**************************************************************************/

VOID dotstahl()
  {
    previdx();
    strcat(clrlline(),"call\t?Aeq0\n");
    previdx();
    sprintf(clrlline(),"sub\t%d\n",dn &	0xFF);
    previdx();
    strcat(clrlline(),"ld\tA,(HL)\n");
    saved += 3;
  }





/**************************************************************************\
*									   *
*     3	  CALL	  ?gcs	     ->						   *
*     3	  LD	  A,(var)	  3   LD      A,(var)			   *
*     3	  CALL	  ?sxt		  1   SUB     (HL)			   *
*     3	  CALL	  ?se		  3   CALL    ?Aeq0			   *
* >>>	  JP/JR	  NZ,		      JP/JR   NZ,			   *
*									   *
\**************************************************************************/

VOID dosaeq0()
  {
    previdx();
    replace(2, /* optf ? 5 : */	3,"sub\t(HL)\n\tcall\t?Aeq0\n");
    previdx();
    clrlline();
    if (strlen(line[idx]) <= 2)	*line[idx] = '\0';
  }





/**************************************************************************\
*									   *
*     3	  LD	  DE,n	     ->	  1   EX      DE,HL			   *
* >>> 3	  CALL	  ?s1div	  3   LD      HL,n			   *
*				  3   CALL    ?sdiv			   *
*									   *
\**************************************************************************/

VOID doexsdiv()
  {
    if (!opt4 && !opt3)
      {
	sprintf(clrlline(),"ld\tHL,%d\n\tcall\t?sdiv\n",dn);
	previdx();
	replace(1,-1,"ex\tDE,HL\n");
      }
  }





/**************************************************************************\
*									   *
*     3	  LD	  HL,s	     ->	  3   LD      HL,$+5			   *
*     1	  PUSH	  HL		  1   EX      (SP),HL			   *
*     3	  LD	  HL,hn		  1   JP      (HL)			   *
*     1	  ADD	  HL,SP							   *
*     3	  CALL	  ?g	     Der Einfachheit halber wird nur		   *
*     1	  JP	  (HL)	     auf "JP (HL), POP DE" gematcht,		   *
*	s:		     das Label kann, aber muss nicht auf	   *
* >>> 1	  POP	  DE	     extra Zeile erscheinen :			   *
*									   *
*									   *
*     3	  LD	  HL,s	     ->	  3   LD      HL,$+5			   *
*     1	  PUSH	  HL		  1   EX      (SP),HL			   *
*     3	  LD	  HL,hn		  1   JP      (HL)			   *
*     1	  ADD	  HL,SP							   *
*     3	  CALL	  ?g							   *
*     1	  JP	  (HL)							   *
* >s: 1	  POP	  DE							   *
*									   *
*     Zusaetzlich werden je nach hn keine (hn =	2), oder mehrere	   *
*     Parameterpushungen (hn = 4,6,8, ...) fuer	die mit	"JP (HL)"	   *
*     aufgerufene Funktion mit "EX (SP),HL" korrigiert.			   *
*     Dies wird	nur durch einfaches Matchen auf	"PUSH HL" gemacht -	   *
*     Aufrufe wie "(*func)(a,b,c)" sind	zulaessig, aber	Aufrufe		   *
*     wie "(*func)(a,f(b),c)" werden nicht richtig umgesetzt !		   *
*									   *
\**************************************************************************/

VOID dojpihl()
  {
    unsigned n;
    unsigned safety;

    if (opt4 ||	!opt3)
      {
	if (state == JPHL)
	  n = 8;
	else
	  n = 7;
	replace(n, /* optf ? 8 : */ 6,
		"ld\tHL,$+5\n\tex\t(SP),HL\n\tjp\t(HL)\n");
	for (n = 2; n <	hn; n+=2)
	  {
	    safety = 5;
	    while (token(line[idx]) != PUSH_HL)
	      {
		previdx();
		if (!--safety) break;
	      }
	    if (safety)	replace(1,-1,"ex\t(SP),HL\n\tpush\tHL\n");
	  }
      }
  }





/**************************************************************************\
*									   *
* Einfache Ersetzungen.							   *
*									   *
\**************************************************************************/

VOID doinc()	 {
/*		   if (optf)
		     replace(5,6,"call\t?inc\n");
		   else
*/
		     if	(opte)
		       replace(5,6,"rst\t08h\t\t; M?INC\n"); /*	M?G > s-2 */
		     else
		       replace(5,6,"M?INC\n");		     /*	M?G > s-2 */
		 }

VOID dodec()	 { replace(5, /* optf ?	6 : */ 4,"call\t?dec\n"); }

VOID doexsphl()	 { replace(2,1,"ex\t(SP),HL\n"); }

VOID doldaora()	 { previdx(); replace(3,7,"ld\tA,(HL)\n\tor\tA\n"); }

VOID dobinc()	 {
/*		   if (optf)
		     replace(4,4,"pop\tHL\n\tcall\t?inc\n");
		   else
*/
		     if	(opte)
		       replace(4,4,"pop\tHL\n\trst\t08h\t\t; M?INC\n");
		     else
		       replace(4,4,"pop\tHL\n\tM?INC\n");    /*	M?G > s-2 */
		 }

VOID dobdec()	 { if (opt4 || optf) replace(4,4,"pop\tHL\n\tcall\t?dec\n"); }

VOID dotsthl()	 { previdx(); replace(2,4,"ld\tA,H\n\tor\tL\n"); }

VOID dotstde()	 { previdx(); replace(2,4,"ld\tA,D\n\tor\tE\n"); }

VOID dosna()	 { replace(2, /* optf ?	3 : */ 1,"call\t?sna\n"); }

VOID domask()	 { replace(2,4,"ld\tH,0\n"); }

VOID dogcsi()	 { replace(2, /* optf ?	3 : */ 1,"call\t?gcsi\n"); }

VOID doclrihl()	 {
		   if (!opt4 &&	!opt3)
		     replace(4,-1,"ld\t(HL),0\n\tinc\tHL\n\tld\t(HL),0\n");
		 }

VOID dosrem()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"call\t?sdiv\n\tex\tDE,HL\n");
		 }

VOID dourem()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"call\t?udiv\n\tex\tDE,HL\n");
		 }

VOID dos1s()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"ex\tDE,HL\n\tcall\t?ss\n");
		 }

VOID dos1div()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"ex\tDE,HL\n\tcall\t?sdiv\n");
		 }

VOID dou1div()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"ex\tDE,HL\n\tcall\t?udiv\n");
		 }

VOID doexudiv()	 {
		   if (!opt4 &&	!opt3)
		     replace(1,-1,"ex\tDE,HL\n\tcall\t?udiv\n");
		 }



/* Ende	von RONNER.C */
