/*
**	File: LOPSYS.C
**	QCX Library
**	Z80 Assembly Language (non-runtime) Library Functions which:
**		- use QCX calling conventions
**		- are operating system or hardware dependent
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	03/21/84
**	modified 4apr85 rj
**		21apr85 jc
*/

/*
**	Z80 port I/O functions contributed by Kirk Bailey
**	modified (slightly) by Jim Colvin
*/

/*: in -- Read a byte from specified input port */
in()
	{
#asm
;	in(port_address) -- Returns the value of the specified input
;		port.
;
	PUSH	BC		;Save frame pointer
	LD	HL,4		;Compute address of "port_address"
	ADD	HL,SP
	LD	C,(HL)		;Get "port_address"
	IN	A,(C)		;Read port
	LD	L,A
	LD	H,0		;Zero extend byte read
	OR	H		;Set zero flag per result
	POP	BC		;Restore frame pointer
#endasm
	}

/*: out -- Write a byte to specified output port */
out()
	{
#asm
;	out(data,port_address) -- Writes "data" to the specified output
;		port.  Returns a copy of the data being written out.
;
	PUSH	BC		;Save frame pointer
	LD	HL,4		;Compute address of "data"
	ADD	HL,SP
	LD	A,(HL)		;Get "data"
	INC	HL		;Compute address of "port_address"
	INC	HL
	LD	C,(HL)		;Get "port_address"
	OUT	(C),A		;Write "data" to port
	LD	L,A
	LD	H,0		;Zero extend byte read
	OR	H		;Set zero flag per result
	POP	BC		;Restore frame pointer
#endasm
	}

/*: setjmp -- UNIX style non-local goto */
setjmp()
	{
#asm
;setjmp(env)
;jmp_buf env;
	PUSH	BC	;save frame ptr
	LD	HL,8	;get value of SP to restore
	ADD	HL,SP
	LD	C,L	;save it
	LD	B,H
	DEC	HL
	LD	D,(HL)	;get contents of cell above SP
	DEC	HL
	LD	E,(HL)
	DEC	HL
	LD	A,(HL)	;get addr of env
	DEC	HL
	LD	L,(HL)
	LD	H,A
	LD	(HL),C	;save SP value to restore
	INC	HL
	LD	(HL),B
	INC	HL
	LD	(HL),E	;save cell above SP
	INC	HL
	LD	(HL),D
	POP	BC	;restore frame ptr
	POP	DE	;get RET addr
	PUSH	DE
	INC	HL
	LD	(HL),E	;save RET addr
	INC	HL
	LD	(HL),D
	INC	HL
	LD	(HL),C	;save frame pointer
	INC	HL
	LD	(HL),B
	XOR	A	;set Z flag
	LD	L,A	;load return value
	LD	H,A
#endasm
	}

/*: longjmp -- non-local return */
longjmp()
	{
#asm
;longjmp(env, retval)
;jmp_buf env;
	POP	BC	;throw away RET address
	POP	HL	;get address of env
	POP	BC	; and return value
	LD	E,(HL)	;retrieve value of SP from env
	INC	HL
	LD	D,(HL)
	EX	DE,HL
	LD	SP,HL	;reset stack pointer
	EX	DE,HL
	INC	HL
	LD	E,(HL)	;reset cell above SP
	INC	HL
	LD	D,(HL)
	PUSH	DE
	PUSH	DE	;fake the arg to setjmp which caller will remove
	INC	HL
	LD	E,(HL)	;restore RET address
	INC	HL
	LD	D,(HL)
	PUSH	DE
	PUSH	BC	;save return value for later
	INC	HL	;retrieve frame pointer
	LD	C,(HL)
	INC	HL
	LD	B,(HL)
	PUSH	BC
	POP	IX	;restore IX
	POP	HL	;retrieve return value
	LD	A,H	;set Z flag
	OR	L
#endasm
	}

/*: makfcb -- Build a CP/M file control block (fcb) */
makfcb()
	{
#asm
;makfcb(filename, fcb)
;char *filename, *fcb;
	PUSH	BC	;save frame pointer
	LD	HL,4	;get args
	ADD	HL,SP
	LD	E,(HL)	;filename
	INC	HL
	LD	D,(HL)
	INC	HL
	LD	A,(HL)	;fcb
	INC	HL
	LD	H,(HL)
	LD	L,A
	INC	DE
	LD	A,(DE)	;2nd char of filename
	DEC	DE
	CP	58	;check for drive specifier (:)
	LD	A,0	;preset logged drive
	JR	NZ,?FCB0
	LD	A,(DE)	;get drive letter
	INC	DE	;move past colon
	INC	DE
	CALL	?FCB9	;convert to upper case
	SUB	64	;convert to drive number (A=1,B=2,etc)
	CP	1
	JR	C,?FCB99 ;<'A' no good
	CP	17
	JR	NC,?FCB99 ;>'P' no good
?FCB0:	LD	(HL),A	;store drive number
	INC	HL
	LD	C,46	;end character can be '.'
	LD	B,8	;max 8 characters for file name
	CALL	?FCB2	;put file name if fcb
	JR	C,?FCB99  ;found error
	LD	C,0	;only '\0' is valid end character
	LD	B,3	;max 3 characters for file type
	CALL	?FCB2
	JR	C,?FCB99
	XOR	A	;fill rest of fcb with zeros
	LD	B,24
?FCB1:	LD	(HL),A
	INC	HL
	DJNZ	?FCB1
	JR	?FCB98	;normal end
;Put file name or file type in fcb
?FCB2:	CALL	?FCB4	;end of string?
	JR	NC,?FCB7
	CP	42	;check for wildcard character '*'
	JR	NZ,?FCB3
	INC	DE	;next character must be end (0 or .)
	CALL	?FCB4
	RET	C	;error
	LD	C,63	;set pad character to '?'
	JR	?FCB8
?FCB3:	CP	32	;check for illegal characters (<' ')
	RET	C
	CP	128	;'~'+1
	JR	NC,?FCB5
	CALL	?FCB9	;convert to upper case
	LD	(HL),A
	INC	HL
	INC	DE
	DJNZ	?FCB2	;any more allowed?
;Check for end character and return carry set if not found
?FCB4:	LD	A,(DE)
	OR	A	;end of string?
	RET	Z
	CP	C	;valid end character?
	JR	NZ,?FCB5
	INC	DE
	RET
?FCB5:	SCF
	RET
;Fill rest with pad character (' ' or '?')
?FCB7:	LD	C,32
?FCB8:	LD	(HL),C
	INC	HL
	DJNZ	?FCB8
	RET
;Convert lower case to upper
?FCB9:	CP	97	;'a'
	RET	C
	CP	123	;'z'+1
	RET	NC
	SUB	32
	RET
;Good return
?FCB98: XOR	A
	LD	L,A	;return value must not be -1
	LD	H,A
	POP	BC	;restore frame pointer
	RET
;Error return
?FCB99: LD	HL,-1	;return value is -1
	OR	L
	POP	BC
#endasm
	}

/*: _shell -- Parse command line but do not provide redirection */
_shell()
	{
#asm
WBOOT	EQU	0
TAB	EQU	9
BLANK	EQU	32
QUOTE	EQU	34
INARG	EQU	1
INSTR	EQU	2
	EXTRN	main?,?ult,r?1?
	LD	HL,-128 ;reserve space for argv strings
	ADD	HL,SP
	LD	SP,HL
	LD	(r?1?),HL ;remember where
	EX	DE,HL	;DE points to argv strings
	LD	BC,1	;initialize argc
	PUSH	BC
	XOR	A
	LD	(DE),A	;argv[0] points to null string
	INC	DE
	LD	C,A	;zero inarg and instring flags
	LD	HL,80H	;CP/M command line
	LD	B,(HL)	;size of command line
?@sh0:	INC	HL	;next command char
	DEC	B
	JP	M,?@sh6 ;end of command line
	LD	A,C
	CP	INSTR	;in a string?
	JR	NZ,?@sh1
	LD	A,(HL)
	CP	QUOTE	;end of string?
	JR	?@sh2
?@sh1:	CP	INARG	;in an arg?
	JR	NZ,?@sh3
	CALL	?@sh9	;white space?
?@sh2:	JR	NZ,?@sh5
	XOR	A	;argv string terminator
	LD	C,A	;turn off flag
	JR	?@sh5
?@sh3:	CALL	?@sh9	;white space?
	JR	Z,?@sh0 ;ignore
	EX	(SP),HL ;retrieve argc
	INC	HL	;count arg
	EX	(SP),HL
	CP	QUOTE
	JR	NZ,?@sh4
	LD	C,INSTR
	JR	?@sh0
?@sh4:	LD	C,INARG
?@sh5:	LD	(DE),A
	INC	DE
	JR	?@sh0
?@sh6:	XOR	A
	LD	(DE),A	;terminate last arg
	POP	BC	;retrieve argc
?@sh7:	DEC	DE
	PUSH	DE
	LD	HL,(r?1?) ;beginning of args
	CALL	?ult	;end of argv area?
	POP	DE
	JR	NZ,?@sh8
	LD	A,(DE)
	CP	0
	JR	NZ,?@sh7
	INC	DE
	PUSH	DE	;record in argv
	DEC	DE
	JR	?@sh7
?@sh8:	INC	DE
	PUSH	DE	;argv[0]
	LD	HL,0
	ADD	HL,SP	;address of argv
	PUSH	HL	;argv
	PUSH	BC	;argc
	CALL	main?
	JP	WBOOT	;reboot CP/M
?@sh9:	LD	A,(HL)	;chk for white space
	CP	TAB
	RET	Z
	CP	BLANK
#endasm
	}

/*: bdos -- Do CP/M system call and return what CP/M puts in HL */
bdos()
	{
#asm
;bdos(c, de)
;int c; (c -> C)
;int de; (de -> DE)
	PUSH	BC	;save frame ptr
	LD	HL,4
	ADD	HL,SP
	LD	C,(HL)	;load function number (c)
	INC	HL
	INC	HL
	LD	E,(HL)	;load de
	INC	HL
	LD	D,(HL)
	CALL	5H
	POP	BC	;restore frame ptr
	PUSH	BC	;restore IX
	POP	IX
	LD	A,L	;don't assume A is set to L
	OR	H	;set Z flag per result
#endasm
	}

/*: bdos1 -- CP/M system call like bdos but return A zero-extended into HL */
bdos1()
	{
#asm
;bdos1(c, de)
	PUSH	BC
	LD	HL,4
	ADD	HL,SP
	LD	C,(HL)
	INC	HL
	INC	HL
	LD	E,(HL)
	INC	HL
	LD	D,(HL)
	CALL	5H
	POP	BC
	PUSH	BC
	POP	IX
	LD	L,A
	LD	H,0
	OR	H
#endasm
	}

/*: mpm -- Make an MP/M system call and return HL and A */
mpm()
	{
#asm
;mpm(c, de, a)
;int c; (c -> C)
;int de; (de -> DE)
;int *a; (*a <- A zero-extended)
	PUSH	BC	;save calling routine's stack frame ptr
	LD	HL,9
	ADD	HL,SP
	LD	D,(HL)	;get address of a
	DEC	HL
	LD	E,(HL)
	PUSH	DE	;save for later
	DEC	HL
	LD	D,(HL)	;load de
	DEC	HL
	LD	E,(HL)
	DEC	HL
	DEC	HL
	LD	C,(HL)	;load function number (c)
	CALL	5H
	POP	DE	;retrieve address of a
	EX	DE,HL	;save return value
	LD	(HL),A
	INC	HL
	LD	(HL),0
	EX	DE,HL
	POP	BC	;restore frame ptr
	PUSH	BC	;restore IX
	POP	IX
	LD	A,L
	OR	H	;set Z flag per result
#endasm
	}
/* end file LOPSYS.C */
