/*
**	File: LQCX.C
**
**	QCX Library
**
**	Non-standard routines used by the QCX compiler
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4Apr85 rj
**		 23Jul85 jc
*/

#define TRUE	1
#define FALSE	0

#include "stdio.h"

/*: getkey -- If a key was pressed, get the character */
getkey()
	{
	return (bdos1(11) & 01) ? bdos1(1): EOF;
	}

/*: qprintf -- short version of printf to conserve space on a Z80 */
qprintf(fmt, args)
register char *fmt;
char *args;
	{
	register c, base;
	register char **arg;
	char s[7], *itob();

	arg = &args;
	while (c = *fmt++) {
		if (c != '%') {
			putchar(c);
			continue;
			}
		switch (c = *fmt++) {
		case 'c':
			putchar( *(char *) arg++);
			continue;
		case 'd':
			base = -10;
			goto prt;
		case 'o':
			base = 8;
			goto prt;
		case 'u':
			base = 10;
			goto prt;
		case 'x':
			base = 16;
		prt:
			fputs(itob( *(int *) arg++, s, base), stdout);
			continue;
		case 's':
			fputs(*arg++, stdout);
			continue;
		default:
			putchar(c);
			continue;
			}
		}
/* end qprintf */ }

/*: cantopen -- report that file can't be opened */
cantopen(filename)
char *filename;
	{
	fputs("Can't open: ", stderr);
	fputs(filename, stderr);
	exit(1);
	}

/*: itob -- convert an integer to a string in any base (2-36) */
char *itob(n, s, base)
char *s;
	{
	register unsigned int u;
	register char *p, *q;
	register negative, c;

	if (n < 0 && base == -10) {
		negative = TRUE;
		u = -n;
		}
	else {
		negative = FALSE;
		u = n;
		}
	if (base == -10)		/* signals signed conversion */
		base = 10;
	p = q = s;
	do {				/* generate digits in reverse order */
		if ((*p = u % base + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		} while ((u /= base) > 0);
	if (negative)
		*p++ = '-';
	*p = '\0';			/* terminate the string */
	while (q < --p) {		/* reverse the digits */
		c = *q;
		*q++ = *p;
		*p = c;
		}
	return s;
/* end itob */ }

/*: imin -- find the minimum of two integers */
imin(i, j)
	{
	return (i < j) ? i : j;
	}
/*: imax -- find the maximum of two integers */
imax(i, j)
	{
	return (i > j) ? i : j;
	}

/* end of LQCX.C */
