/*
**	File: romfloat.c
**
**	Q/C Library
**
**	This file splits out the floating point support for printf and
**	scanf to reduce the size of programs which don't need it.  It
**	is compiled into a separate library called QCFLOAT.REL which
**	linked ahead of CRUNLIB.REL when floating point support is needed.
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 10 sep 85 jc
**		  17 Apr 93 ACZ - reduced to functions used for
**				  ROM-programs in EMUFs
*/

#include "romio.h"

#define	TRUE	1
#define	FALSE	0
#define MAXFLD	128	/* maximum field width for printf formats */
#define MAXSCAN 128	/* maximum size of string in scanf %[] format */
#define MAX_FLT_EXP 100	/* maximum floating point exponent in scanf */

/* some constants for faster processing */

float one, onehalf, tenth, ten, two;
float sqrof2, sqrof_5, frp8, frp48, frp384;

/* this function MUST! be called before doing any output of floats */
floatinit()
{
    one = 1;
    onehalf = 1 / (float)2;
    tenth = 1 / (float)10;
    ten = 10;
    two = 2;
	/* the following is for sqrt */
    sqrof2 = 1.414213;
    sqrof_5 = 0.707107;
    frp8 = 1 / (float)8;
    frp48 = 1 / (float)48;
    frp384 = 1 / (float)384;
}

/*: sprintf -- format a string */
sprintf(s, format, args)
char *s, *format, *args;
	{
	int _cat();			/* function _fmt will use as output */

	*s = '\0';			/* initialize the string */
	_fmt(_cat, &s, format, &args);
	}

/*: _fmt -- Internal routine for printf/fprintf/sprintf */
_fmt(func, dest, format, args)
register (*func)();
register char *dest, **args;
char *format;
{
/*
 * WARNING: the static variables used in this function for compactness
 *	    prevent it from being used in an interrupt environment
 *	    where this routine can be overlaid and then reloaded.
 *	    Change all static declarations to auto or register to
 *	    eliminate this problem.
 */
	static char *sptr;
	register char *wptr;
	register int c;
	static base, pf, ljf, padchr, width, precision;
	int longflag;
	char wbuf[MAXFLD], *itob(), *ltob(), *_ftof(), *_ftoe();

	while (c = *format++) {
	    if (c == '%') {
		wptr = wbuf;
		precision = 6;
		ljf = pf = longflag = FALSE;

		if (*format == '-') {
			++format;
			ljf = TRUE;
			}
		padchr = (*format == '0') ? '0': ' ';
		width = (isdigit(*format)) ? _atoi(&format) : 0;
		if ((c = *format++) == '.') {
			precision = _atoi(&format);
			pf = TRUE;
			c = *format++;
			}
		if (toupper(c) == 'L') {
			longflag = TRUE;
			c = *format++;
			}
		switch(toupper(c)) {

		case 'D':
			base = -10; goto val;
		case 'U':
			base = 10; goto val;
		case 'X':
			base = 16; goto val;
		case 'O':
			base = 8;
		val:
			if (longflag) {
				ltob( *(long *)args, wptr, base);
				args += 2;
				}
			else
				itob( *(int *)args++, wptr, base);
			width -= strlen(wptr);
			goto pad2;
		case 'F':
			_ftof( *(float *)args, wptr, precision);
			args += 2;
			width -= strlen(wptr);
			goto pad2;
		case 'E':
			_ftoe( *(float *)args, wptr, precision);
			args += 2;
			width -= strlen(wptr);
			goto pad2;
		case 'C':
			*wptr++ = *args++;
			--width;
			goto pad;
		case 'S':
			if (!pf) precision = MAXFLD;
			sptr = *args++;
			while (*sptr && precision) {
				*wptr++ = *sptr++;
				--precision;
				--width;
			}
		pad:	*wptr = '\0';
		pad2:	wptr = wbuf;
			if (!ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			while (*wptr)
				(*func)(*wptr++,dest);
			if (ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			break;

		default:
			(*func)(c,dest);
		}
	    }
	    else (*func)(c,dest);
	}
/* end _fmt */ }

/*: _ftof -- convert a float to a string in f format */
char *_ftof(n, s, precision)
float n;
int precision;
char *s;
	{
	float u, round;
	register char *p;
	register int count, c;
/*	float fhalf, ten; */

	p = s;
	if ( _nan_test( n ) ) {
		*p++ = '*';
		*p++ = '.';
		for ( count=precision; count; --count )
			*p++ = '*';
		*p = '\0';
		return s;
		}
/*	ten = 10; 
	fhalf = 1 / (float) 2; */
	if (n < 0) {
		*p++ = '-';
		u = -n;
	}
	else
		u = n;
	for (count=0, round=onehalf; count < precision; count++)
		round *= tenth;
	u += round;	/* half round to correct position */
	for (count=0; (u >= ten && count < 9); count++)
		u *= tenth;
	while (count-- >= 0) {		/* generate digits left of decimal point */
		*p++ = (c = (char) u) + '0';
		u = (u - c) * ten;
		}
	if (precision > 0)
		*p++ = '.';
	while (precision-- > 0) {
		*p++ = (c = (char) u) + '0';
		u = (u - c) * ten;
		}
	*p = '\0';
	return s;
/* end _ftof */ }

/*: _ftoe -- convert a float to a string in e format */
char *_ftoe(n, s, precision)
float n;
int precision;
char *s;
	{
	float u, round;
	register char *p;
	register int count, exponent, c;
/*	float fhalf, ten, one; */

	p = s;
	if ( _nan_test( n ) ) {
		*p++ = '*';
		*p++ = '.';
		for ( count=precision; count; --count )
			*p++ = '*';
		*p++ = 'e';
		*p++ = ' ';
		*p++ = '*';
		*p++ = '*';
		*p = '\0';
		return s;
		}
/*	one = 1;
	ten = 10;
	fhalf = one / 2; */
	if (n < 0) {
		*p++ = '-';
		u = -n;
		}
	else
		u = n;
	for (exponent=0; u >= ten; ++exponent)
		u *= tenth;
	if (u > 0)
		while (u < one) {
			u *= ten;
			--exponent;
			}
	round = onehalf;
	for ( count=0; count < precision; ++count)
		round *= tenth;
	u += round;	/* half round to correct position */
	if (u >= ten) {
		u *= tenth;
		++exponent;
		}
	*p++ = (c = (char) u) + '0';	/* generate digit to left of decimal point */
	u = (u - c) * ten;
	if (precision > 0)
		*p++ = '.';
	while (precision-- > 0) {
		*p++ = (c = (char) u) + '0';
		u = (u - c) * ten;
		}
	*p++ = 'e';
	if (exponent < 0) {
		*p++ = '-';
		exponent = -exponent;
		}
	else
		*p++ = ' ';
	*p++ = exponent / 10 + '0';
	*p++ = exponent % 10 + '0';
	*p = '\0';
	return s;
/* end _ftoe */ }

/*: _nan_test -- Check floating point number to see if it is Not-A-Number */
_nan_test( n )
	float n;
	{
	union mix {
		float f;
		short i[2];
		} u;
	
	u.f = n;
	if ( ( u.i[1] & 0x7f80 ) == 0x7f80 )
		return 1;
	else
		return 0;
/* end _nan_test */ }

/*: sscanf -- Get formatted input from a string */
sscanf(s, fmt, args)
char *s, *fmt, *args;
	{
	char ss[MAXSCAN+1], *strchr();
	static char *p, **pargs;
	static count, c, f, include, present, quit, suppress, width,
		base, d, maxd, longflag, floatflag, sign;
	static long value;
	static float /* ten, */ tenpwr, fvalue;

	pargs = &args;
	count = 0;
	quit = FALSE;
	c = *s++;
	while (f = *fmt++) {
		if (isspace(f)) 	/* skip white space in format */
			continue;
		if (f != '%') { 	/* must match next input string */
			while (isspace(c))
				c = *s++;
			if (f == c) {
				c = *s++; /* matches so get next char */
				continue;
				}
			else
				break;
			}
		if (*fmt == '*') {
			suppress = TRUE;
			++fmt;
			}
		else
			suppress = FALSE;
		width = (isdigit(*fmt))? _atoi(&fmt): -1;
		if (*fmt == 'h') {	/* short int modifier */
			longflag = FALSE;
			++fmt;
			}
		else if (*fmt == 'l') {	/* long integer modifier */
			longflag = TRUE;
			++fmt;
			}
		else
			longflag = FALSE;
		floatflag = FALSE;
		switch (*fmt++) {
		case '%':
			while (isspace(c))
				c = *s++;
			if (c == '%') {
				c = *s++;
				continue;
				}
			else {
				quit = TRUE;
				break;
				}
		case '[':
			if (*fmt != '^')
				include = TRUE; /* include chars in [] */
			else {
				include = FALSE;
				++fmt;
				}
			for (d = 0; d < MAXSCAN; ++d) {
				f = *fmt++;	/* get char to incl/excl */
				if (f == ']')
					break;
				else if (f == '\0') {
					quit = TRUE;
					break;
					}
				else
					ss[d] = f;
				}
			ss[d] = '\0';
			if (quit)
				break;		/* format didn't end with ] */
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				s = *pargs++;
				}
			while (width--) {
				present = (strchr(ss, c) != NULL);
				if ((include && !present)
				   || (!include && present))
					break;	/* doesn't meet %[s] */
				if (!suppress)
					*s++ = c;
				if ((c = *s++) == '\0')
					break;	/* end of input string */
				}
			if (!suppress)
				*s = '\0';
			break;
		case 'X':
			longflag = TRUE;
		case 'x':
			base = 16;
			maxd = 'F';
			sign = 1;
			goto getnum;
		case 'O':
			longflag = TRUE;
		case 'o':
			base = 8;
			maxd = '7';
			sign = 1;
			goto getnum;
		case 'F':
		case 'E':
		case 'f':
		case 'e':
			floatflag = TRUE;
			fvalue = 0;
			/* ten = 10; */
			while (isspace(c))
				c = *s++;
			if (c == '-') {
				--width;
				sign = -1;
				c = *s++;
				}
			else if (c == '+') {
				--width;
				sign = 1;
				c = *s++;
				}
			else
				sign = 1;
			if (isdigit(c) || c == '.') {
				if (!suppress)
					++count;
				}
			else {
				quit = TRUE;
				break;
				}
			while (isdigit(c) && width--) {
				fvalue = ten * fvalue + (c - '0');
				c = *s++;
				}
			if (c == '.' && width) {
				tenpwr = ten;
				c = *s++;
				--width;
				while (isdigit(c) && width--) {
					fvalue += (c - '0') / tenpwr;
					tenpwr *= ten;
					c = *s++;
					}
				}
			fvalue *= sign;	/* apply correct sign to number */
			if (tolower(c) == 'e' && width) {
				c = *s++;
				--width;
				base = 10;
				maxd = '9';
				if (width)
					goto getexp;
				else {
					--count;
					quit = TRUE;
					break;
					}
				}
			if (!suppress)
				*(float *)*pargs++ = fvalue;
			break;
		case 'D':
			longflag = TRUE;
		case 'd':
			base = 10;
			maxd = '9';
			while (isspace(c))
				c = *s++;
		getexp:
			if (c == '-') {
				sign = -1;
				c = *s++;
				--width;
				}
			else if (c == '+') {
				sign = 1;
				c = *s++;
				--width;
				}
			else
				sign = 1;
			goto getnum1;
		getnum:
			while (isspace(c))
				c = *s++;
		getnum1:
			d = toupper(c);
			if (isalnum(d) && d <= maxd && width) {
				if (!suppress && !floatflag)
					++count;
				}
			else {
				if (floatflag)
					--count;
				quit = TRUE;
				break;
				}
			value = 0;
			while (width--) {
				d = toupper(c);
				if (isdigit(d) && d <= maxd)
					d -= '0';
				else if (isalpha(d) && d <= maxd)
					d -= ('A' - 10);
				else
					break;
				/* build negative to handle max negative int */
				value = base * value - d;
				c = *s++;
				}
			if (!suppress)
				if (sign > 0)
					value = -value; /* to handle max neg */
				if (floatflag) {
					if (value > MAX_FLT_EXP)
						value = MAX_FLT_EXP;
					else if (value < -MAX_FLT_EXP)
						value = -MAX_FLT_EXP;
					if (sign > 0)
						while (value--)
							fvalue *= ten;
					else
						while (value++)
							fvalue /= ten;
					*(float *)*pargs++ = fvalue;
					}
				else if (longflag)
					*(long *)*pargs++ = value;
				else
					*(int *)*pargs++ = value;
			break;
		case 'C':
		case 'c':
			if (c == '\0')
				break;
			if (suppress) {
				if (width == -1) {
					c = *s++; /* get next character */
					continue;
					}
				while (width--) {
					if ((c=*s++) == '\0')
						break;
					}
				break;
				}
			++count;
			if (width == -1) { /* get single char */
				**pargs++ = c;
				c = *s++;
				continue;
				}
			else {		/* get string */
				p = *pargs++;
				while (width--) {
					*p++ = c;
					if ((c=*s++) == '\0')
						break;
					}
				*p = '\0';
				}
			break;
		case 'S':
		case 's':
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				p = *pargs++;
				}
			while (width--) {
				if (!suppress)
					*p++ = c;
				c = *s++;
				if (isspace(c) || c == '\0')
					break;
				}
			if (!suppress)
				*p = '\0';
			break;
		default:		/* can't figure this out */
			quit = TRUE;
			break;
			}
		if (c == '\0'  || quit == TRUE)
			break;
		}
	return ((count == 0 && c == '\0')? EOF: count);
/* end sscanf */	}

/* ---------------- extenions by dl4sdc : sqr, sqrt --------------------- */

float sqr(arg)
float arg;
{
    return (arg*arg);
}

float sqrt(arg)
float arg;
{
    float fac,xv,rsum;

    fac = one;
    if (arg <= 0) return ((float) -1);
    while (arg > two) {
	arg *= onehalf;
	fac *= sqrof2;
    }
    while (arg < one) {
	arg *= two;
	fac *= sqrof_5;
    }
    arg = arg - one;
    xv = arg;
    rsum = 1 + xv * onehalf;
    xv *= arg;
    rsum -= xv * frp8;
    xv *= arg;
    rsum += xv * frp48;
    xv *= arg;
    rsum -= xv * frp384;
    rsum *= fac;
    return (rsum);
}

/* end file LFLOAT.C */
