/*
**	File: romio.c
**
**
**	Notes:
**		   Includes routines which implement common
**		   UNIX(TM) system calls.
**
**	Z80 port I/O functions contributed by Kirk Bailey
**	modified (slightly) by Jim Colvin
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4 Apr 85 rj
**		 11 Sep 85 jc
**		 16 Apr 93 ACZ - reduced to the funcions needed for
**				 ROM-programs in EMUFs
*/

#include "romio.h"

/* Library parameters */

#define	TRUE	1
#define	FALSE	0
#define MAXFLD	128	/* maximum field width for printf formats */
#define MAXSCAN 128	/* maximum size of string in _scanf %[] format */

/*: ltob -- convert long to string in any base (2-36) */
char *ltob(n, s, base)
long n;
int base;
char *s;
	{
	long u;
	register char *p, *q;
	register negative;
	int c;
	long lundiv();

	if (n < 0 && base < 0) {
		negative = TRUE;
		u = -n;
		}
	else {
		negative = FALSE;
		u = n;
		}
	if (base < 0)	/* signals signed conversion */
		base = -base;
	p = q = s;
	if (u < 0) {
		u = lundiv( u, base, &c);
		if ((*p = c + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		}	/* provides for correct unsigned conversion */
	do {			/* generate digits in reverse order */
		if ((*p = u % base + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		} while ((u /= base) > 0);
	if (negative)
		*p++ = '-';
	*p = '\0';		/* terminate the string */
	while (q < --p) {	/* reverse the digits */
		c = *q;
		*q++ = *p;
		*p = c;
		}
	return s;
/* end ltob */	}
	
/*: lundiv -- divide long int by unsigned int treating long int as unsigned */
long lundiv( n, base, rem)
	long n;
	register unsigned base;
	unsigned *rem;
	{
	register unsigned numerator;
	long result;
	register int count;
	
	numerator = 0;
	result = 0;
	for( count=24; count >= 0; count -= 8) {
		numerator = (numerator<<8) + ((unsigned) ((n>>count) & 255));
		result = (result<<8) + numerator / base;
		numerator %= base;
		}
	*rem = numerator;
	return result;
/* end lundiv */ }

/*: _cat -- add character to end of string (internal routine for sprintf) */
_cat(c, s)
char **s;
	{
	*(*s)++ = c;		/* add c at end and advance pointer*/
	*(*s) = '\0';		/* new end of string */
	}


/*: in -- Read a byte from specified input port */
in()
	{
#asm
;	in(port_address) -- Returns the value of the specified input
;		port.
;
	PUSH	BC		;Save frame pointer
	LD	HL,4		;Compute address of "port_address"
	ADD	HL,SP
	LD	C,(HL)		;Get "port_address"
	IN	A,(C)		;Read port
	LD	L,A
	LD	H,0		;Zero extend byte read
	OR	H		;Set zero flag per result
	POP	BC		;Restore frame pointer
#endasm
	}

/*: out -- Write a byte to specified output port */
out()
	{
#asm
;	out(data,port_address) -- Writes "data" to the specified output
;		port.  Returns a copy of the data being written out.
;
	PUSH	BC		;Save frame pointer
	LD	HL,4		;Compute address of "data"
	ADD	HL,SP
	LD	A,(HL)		;Get "data"
	INC	HL		;Compute address of "port_address"
	INC	HL
	LD	C,(HL)		;Get "port_address"
	OUT	(C),A		;Write "data" to port
	LD	L,A
	LD	H,0		;Zero extend byte read
	OR	H		;Set zero flag per result
	POP	BC		;Restore frame pointer
#endasm
	}

/*: setjmp -- UNIX style non-local goto */
setjmp()
	{
#asm
;setjmp(env)
;jmp_buf env;
	PUSH	BC	;save frame ptr
	LD	HL,8	;get value of SP to restore
	ADD	HL,SP
	LD	C,L	;save it
	LD	B,H
	DEC	HL
	LD	D,(HL)	;get contents of cell above SP
	DEC	HL
	LD	E,(HL)
	DEC	HL
	LD	A,(HL)	;get addr of env
	DEC	HL
	LD	L,(HL)
	LD	H,A
	LD	(HL),C	;save SP value to restore
	INC	HL
	LD	(HL),B
	INC	HL
	LD	(HL),E	;save cell above SP
	INC	HL
	LD	(HL),D
	POP	BC	;restore frame ptr
	POP	DE	;get RET addr
	PUSH	DE
	INC	HL
	LD	(HL),E	;save RET addr
	INC	HL
	LD	(HL),D
	INC	HL
	LD	(HL),C	;save frame pointer
	INC	HL
	LD	(HL),B
	XOR	A	;set Z flag
	LD	L,A	;load return value
	LD	H,A
#endasm
	}

/*: longjmp -- non-local return */
longjmp()
	{
#asm
;longjmp(env, retval)
;jmp_buf env;
	POP	BC	;throw away RET address
	POP	HL	;get address of env
	POP	BC	; and return value
	LD	E,(HL)	;retrieve value of SP from env
	INC	HL
	LD	D,(HL)
	EX	DE,HL
	LD	SP,HL	;reset stack pointer
	EX	DE,HL
	INC	HL
	LD	E,(HL)	;reset cell above SP
	INC	HL
	LD	D,(HL)
	PUSH	DE
	PUSH	DE	;fake the arg to setjmp which caller will remove
	INC	HL
	LD	E,(HL)	;restore RET address
	INC	HL
	LD	D,(HL)
	PUSH	DE
	PUSH	BC	;save return value for later
	INC	HL	;retrieve frame pointer
	LD	C,(HL)
	INC	HL
	LD	B,(HL)
	PUSH	BC
	POP	IX	;restore IX
	POP	HL	;retrieve return value
	LD	A,H	;set Z flag
	OR	L
#endasm
	}


/*
**	File: LQCX.C
**
**	QCX Library
**
**	Non-standard routines used by the QCX compiler
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4Apr85 rj
**		 23Jul85 jc
*/


/*: itob -- convert an integer to a string in any base (2-36) */
char *itob(n, s, base)
char *s;
	{
	register unsigned int u;
	register char *p, *q;
	register negative, c;

	if (n < 0 && base == -10) {
		negative = TRUE;
		u = -n;
		}
	else {
		negative = FALSE;
		u = n;
		}
	if (base == -10)		/* signals signed conversion */
		base = 10;
	p = q = s;
	do {				/* generate digits in reverse order */
		if ((*p = u % base + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		} while ((u /= base) > 0);
	if (negative)
		*p++ = '-';
	*p = '\0';			/* terminate the string */
	while (q < --p) {		/* reverse the digits */
		c = *q;
		*q++ = *p;
		*p = c;
		}
	return s;
/* end itob */ }

/*: imin -- find the minimum of two integers */
imin(i, j)
	{
	return (i < j) ? i : j;
	}
/*: imax -- find the maximum of two integers */
imax(i, j)
	{
	return (i > j) ? i : j;
	}

/* Library parameters */

/*: index -- find the first occurrence of a given character */
/*		Note: UNIX V7 name calling the new C standard name */
char *index(s, c)
char *s;
	{
	char *strchr();

	return strchr(s, c);
	}

/*: rindex -- find the last occurrence of a given character */
/*		Note: UNIX V7 name calling the new C standard name */
char *rindex(s, c)
char *s;
	{
	char *strrchr();

	return strrchr(s, c);
	}

/*: strmov -- copy second string into first string & point to end */
char *strmov(s1, s2)
char *s1, *s2;
	{
	register char *p;

	p = s1;
	while (*p = *s2++)	/* copy 2nd string into 1st */
		++p;
	return p;
	}

/*: peek -- get the byte at address */
peek(address)
char *address;
	{
	return (*address & 0xFF);
	}

/*: poke -- put value in the byte at address */
poke(address, value)
char *address, value;
	{
	int oldval;

	oldval = (*address & 0xFF);
	*address = value;
	return oldval;
	}

/*: wpeek -- get the word at address */
wpeek(address)
int *address;
	{
	return *address;
	}

/*: wpoke -- put value in the word at address */
wpoke(address, value)
int *address;
	{
	int oldval;

	oldval = *address;
	*address = value;
	return oldval;
	}

/*
**	D.10 -- General Utilities --
**		D.10.1 string conversion
**		D.10.2 (not yet implemented) random numbers
**		D.10.3 memory management
**		D.10.4 communication with environment
*/

/*
** Structure for the memory allocator
*/
struct _header {
	struct _header	*_next; /* next free block */
	unsigned	_bsize; /* size of block */
	};
typedef struct _header HEADER;

extern char	*_free; 	/* free space address */
				/* Note: the Microsoft L80 linker puts */
				/* the address of the next available memory */
				/* in the variable $MEMRY. (RMAC uses ?MEMRY) */
				/* the compiler external variable _free */
				/* is at the same location, and so the linker */
				/* does the needed magic */

unsigned	_moat = 1000;	/* moat between heap and stack */
HEADER		_base,		/* empty free space (heap) list */
		*_allocp;	/* last allocated block */

/*: atof -- convert a float string constant to its float value */
float atof(s)
char *s;
	{
	register char *p;
	register c, exp, sign;
	float tenpwr, fvalue;

/* NOTE: do NOT use any floating point constants in this function
 *	 because QCX calls this routine to convert floating point
 *	 constants at run time when it is used as a cross-compiler.
 */
	p = s;
	fvalue = 0;	
	while (isspace(c = *p++))	/* skip leading white space */
		;
	if (c == '-') {
		sign = -1;
		c = *p++;
		}
	else if (c == '+') {
		sign = 1;
		c = *p++;
		}
	else
		sign = 1;
	while (isdigit(c)) {
		fvalue = 10 * fvalue + (c - '0');
		c = *p++;
		}
	if (c == '.') {
		tenpwr = 10;
		c = *p++;
		while (isdigit(c)) {
			fvalue += (c - '0') / tenpwr;
			tenpwr *= 10;
			c = *p++;
			}
		}
	fvalue *= sign;			/* apply correct sign to number */
	if (toupper(c) == 'E') {
		exp = 0;
		c = *p++;
		if (c == '-') {
			sign = -1;
			c = *p++;
			}
		else if (c == '+') {
			sign = 1;
			c = *p++;
			}
		else
			sign = 1;
		while (isdigit(c)) {
			exp = 10 * exp + (c - '0');
			c = *p++;
			if (exp > 100) {	/* artificial limit for */
				exp = 100;	/* while loops below */
				break;
				}
			}
		if (sign > 0)
			while (exp--)
				fvalue *= 10;
		else
			while (exp--)
				fvalue /= 10;
		}
	return fvalue;
/* end atof */	}

/*: atoi -- convert an ASCII decimal string to integer */
atoi(s)
register char *s;
	{
	register sign;
	char *p;

	while (isspace(*s))
		++s;
	sign = 1;
	switch (*s) {
	case '-':
		sign = -1;
	case '+':
		++s;
		break;
		}
	p = s;
	return (sign * _atoi(&p));
	}

/*: _atoi -- internal routine for atoi and _fmt to perform ascii-
 *	     to-decimal conversion and update an associated pointer
 */
_atoi(sptr)
char **sptr;
	{
	register n;
	register char *p;

	n = 0;
	p = *sptr;
	while (isdigit(*p))
		n = 10 * n + *p++ - '0';
	*sptr = p;
	return n;
	}

/*: malloc -- allocate memory */
char *malloc(nbytes)
unsigned nbytes;
	{
	register HEADER *p, *q;
	register unsigned nunits;
	char *sbrk();

	nunits = 1 + (nbytes + (sizeof(HEADER) - 1)) / sizeof(HEADER);
	if ((q = _allocp) == NULL)	/* no free list */
		_base._next = _allocp = q = &_base;
	for (p = q->_next; ; q = p, p = p->_next) {
		if (p->_bsize >= nunits) {
			if (p->_bsize == nunits)
				q->_next = p->_next;
			else {
				p->_bsize -= nunits;
				p += p->_bsize;
				p->_bsize = nunits;
				}
			_allocp = q;
			return ((char *)(p+1));
			}
		if (p == _allocp) {	/* wrapped around */
			p = (HEADER *) sbrk(nunits * sizeof(HEADER));
			if ((int) p == -1)
				return (char *) NULL;
			p->_bsize = nunits;
			free((char *)(p+1)); /* add to free list */
			p = _allocp;
			}
		}
	}

/*: calloc -- allocate n elements of a given size of memory */
char *calloc(nelements, size)
unsigned nelements, size;
	{
	register char *p, *q;
	char *malloc();

	size *= nelements;
	if ((q = p = malloc(size)) != NULL) {
		while (size--)
			*q++ = 0;
		}
	return p;
	}

/*: free -- free memory allocated with malloc */
free(pblk)
char *pblk;
	{
	register HEADER *p, *q;

	p = (HEADER *)pblk - 1; 	/* point to block header */
	for (q=_allocp; !(p>q && p<q->_next); q=q->_next)
		if (q >= q->_next && (p>q || p<q->_next))
			break;
	if (p+p->_bsize == q->_next) {	/* join to upper neighbor */
		p->_bsize += q->_next->_bsize;
		p->_next = q->_next->_next;
		}
	else
		p->_next = q->_next;
	if (q+q->_bsize == p) { 	/* join to lower neighbor */
		q->_bsize += p->_bsize;
		q->_next = p->_next;
		}
	else
		q->_next = p;
	_allocp = q;
	}

/*: sbrk -- set the memory breakpoint */
char *sbrk(size)
register unsigned size;
	{
	unsigned maxsbrk();

	if (maxsbrk() < size)
		return -1;	/* can't go into moat */
	_free += size;		/* move free space ptr */
	return (_free - size);
	}

/*: moat -- Change size of moat between free space and stack */
unsigned moat(size)
unsigned size;
	{
	unsigned oldsize;

	oldsize = _moat;
	_moat = size;
	return oldsize;
	}

/*: maxsbrk -- Compute memory left to be allocated */
unsigned maxsbrk()
	{
	auto char sp[1];	/* sp is on the stack , so sp = SP */
	return ((sp>_free+_moat) ? sp-(_free+_moat): 0);
	}

/* end ROMIO.C */
